Package: huge
Type: Package
Title: High-Dimensional Undirected Graph Estimation
Version: 2.0.1
Authors@R: c(
    person("Haoming", "Jiang", role = "aut"),
    person("Xinyu", "Fei", role = "aut"),
    person("Han", "Liu", role = "aut"),
    person("Kathryn", "Roeder", role = "aut"),
    person("John", "Lafferty", role = "aut"),
    person("Larry", "Wasserman", role = "aut"),
    person("Xingguo", "Li", role = "aut"),
    person("Tuo", "Zhao", role = c("aut", "cre"), email = "tourzhao@gatech.edu")
  )
Maintainer: Tuo Zhao <tourzhao@gatech.edu>
Depends: R (>= 3.4.0)
Imports: Matrix, igraph, MASS, grDevices, graphics, methods, parallel,
        stats, utils, Rcpp
Suggests: testthat (>= 3.1.7)
LinkingTo: Rcpp
Description: Provides a general framework for
        high-dimensional undirected graph estimation. It integrates
        data preprocessing, neighborhood screening, graph estimation,
        and model selection techniques into a pipeline. In
        preprocessing stage, the nonparanormal(npn) transformation is
        applied to help relax the normality assumption. In the graph
        estimation stage, the graph structure is estimated by
        Meinshausen-Buhlmann graph estimation, the graphical lasso,
        or the TIGER (tuning-insensitive graph estimation and
        regression) method, and the first two can be further
        accelerated by the lossy screening rule preselecting the
        neighborhood of each variable by correlation thresholding. We
        target on high-dimensional data analysis usually d >> n, and
        the computation is memory-optimized using the sparse matrix
        output. We also provide a computationally efficient approach,
        correlation thresholding graph estimation. Three
        regularization/thresholding parameter selection methods are
        included in this package: (1)stability approach for
        regularization selection (2) rotation information criterion (3)
        extended Bayesian information criterion which is only available
        for the graphical lasso.
License: GPL-2
URL: https://github.com/Gatech-Flash/huge
BugReports: https://github.com/Gatech-Flash/huge/issues
Repository: CRAN
NeedsCompilation: yes
SystemRequirements: C++17
RoxygenNote: 7.3.3
Encoding: UTF-8
Packaged: 2026-08-01 00:51:38 UTC; tourzhao
Author: Haoming Jiang [aut],
  Xinyu Fei [aut],
  Han Liu [aut],
  Kathryn Roeder [aut],
  John Lafferty [aut],
  Larry Wasserman [aut],
  Xingguo Li [aut],
  Tuo Zhao [aut, cre]
Date/Publication: 2026-08-04 10:30:34 UTC
Built: R 4.5.2; aarch64-apple-darwin20; 2026-08-04 13:04:28 UTC; unix
Archs: huge.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method(plot,huge)
S3method(plot,roc)
S3method(plot,select)
S3method(plot,sim)
S3method(print,huge)
S3method(print,roc)
S3method(print,select)
S3method(print,sim)
export(huge)
export(huge.ct)
export(huge.generator)
export(huge.glasso)
export(huge.inference)
export(huge.mb)
export(huge.npn)
export(huge.plot)
export(huge.roc)
export(huge.select)
export(huge.tiger)
importFrom(MASS,mvrnorm)
importFrom(Matrix,Matrix)
importFrom(Rcpp,evalCpp)
importFrom(Rcpp,sourceCpp)
importFrom(grDevices,dev.off)
importFrom(grDevices,gray.colors)
importFrom(grDevices,postscript)
importFrom(graphics,image)
importFrom(graphics,lines)
importFrom(graphics,par)
importFrom(graphics,plot)
importFrom(igraph,graph_from_adjacency_matrix)
importFrom(igraph,layout_with_fr)
importFrom(methods,new)
importFrom(parallel,mclapply)
importFrom(stats,cor)
importFrom(stats,cov2cor)
importFrom(stats,pnorm)
importFrom(stats,qnorm)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(utils,flush.console)
useDynLib(huge, .registration = TRUE)

#-----------------------------------------------------------------------#
# Package: High-dimensional Undirected Graph Estimation                 #
# huge.plot(): graph visualization                                      #
#-----------------------------------------------------------------------#

#' Graph visualization
#'
#' Implements the graph visualization using adjacency matrix. It can automatic organize 2D embedding layout.
#'
#' The user can change \code{cur.num} to plot several figures and select the best one. The implementation is based on the popular package "igraph".
#'
#' @param G The adjacency matrix corresponding to the graph.
#' @param epsflag If \code{epsflag = TRUE}, save the plot as an eps file in the target directory. The default value is \code{FALSE}.
#' @param graph.name The name of the output eps files. The default value is "default".
#' @param cur.num The number of plots saved as eps files. Only applicale when \code{epsflag = TRUE}. The default value is 1.
#' @param location Target directory. The default value is the current working directory.
#' @seealso \code{\link{huge}} and \code{\link{huge-package}}.
#' @examples
#' ## visualize the hub graph
#' L = huge.generator(graph = "hub")
#' huge.plot(L$theta)
#'
#' ## visualize the band graph
#' L = huge.generator(graph = "band",g=5)
#' huge.plot(L$theta)
#'
#' ## visualize the cluster graph
#' L = huge.generator(graph = "cluster")
#' huge.plot(L$theta)
#'
#' ## plot 5 graphs and save the plots as eps files in the tempdir()
#' huge.plot(L$theta, epsflag = TRUE, cur.num = 5, location = tempdir())
#' @export
huge.plot = function(G, epsflag = FALSE, graph.name = "default", cur.num = 1, location=NULL){
  if(missing(location))  location = tempdir()
  g = graph_from_adjacency_matrix(as.matrix(G!=0), mode="undirected", diag=FALSE)
  layout.grid = layout_with_fr(g)

  if(epsflag == TRUE){
    caller.device = grDevices::dev.cur()
    postscript(file.path(location, paste(paste(graph.name, cur.num, sep=""), "eps", sep=".")), width = 8.0, height = 8.0)
    plot.device = grDevices::dev.cur()
    on.exit({
      devices = grDevices::dev.list()
      if(!is.null(devices) && plot.device %in% devices)
        grDevices::dev.off(plot.device)
      devices = grDevices::dev.list()
      if(caller.device != 1L && !is.null(devices) &&
         caller.device %in% devices)
        grDevices::dev.set(caller.device)
    }, add = TRUE)
  } else {
    old.par = .huge_graphics_state()
    on.exit(.huge_restore_graphics_state(old.par), add = TRUE)
  }
  par(mfrow = c(1,1))
  plot(g, layout=layout.grid, edge.color='gray50',vertex.color="red", vertex.size=2, vertex.label=NA)
}

stress_graphics_state <- function() {
  graphics::par(
    mfrow = c(2, 3), pty = "m",
    omi = rep(0.05, 4), mai = rep(0.2, 4)
  )
  graphics::par(
    mfg = c(1, 2), cex = 1.37, mex = 0.91,
    usr = c(-2, 3, -4, 5),
    xaxp = c(-2, 3, 5), yaxp = c(-4, 5, 9),
    las = 2, xaxs = "i", yaxs = "i", xpd = NA
  )
  graphics::par(new = FALSE)
}

custom_graphics_state <- function(kind) {
  graphics::par(
    mfrow = c(1, 1), pty = "m", oma = rep(0, 4),
    mar = c(5.1, 4.1, 4.1, 2.1), cex = 1, mex = 1
  )
  graphics::plot.new()
  graphics::plot.window(c(-2, 3), c(-4, 5))
  switch(
    kind,
    omi = graphics::par(omi = c(0.11, 0.22, 0.13, 0.17)),
    fig = graphics::par(fig = c(0.12, 0.83, 0.18, 0.88)),
    fin = graphics::par(fin = c(4.2, 3.8)),
    plt = graphics::par(plt = c(0.17, 0.82, 0.21, 0.79)),
    pin = graphics::par(pin = c(3.7, 3.2)),
    mai = graphics::par(mai = c(0.31, 0.44, 0.27, 0.19)),
    mar = graphics::par(mar = c(2.2, 3.1, 1.8, 0.7)),
    omd = graphics::par(omd = c(0.08, 0.91, 0.12, 0.94)),
    oma = graphics::par(
      oma = c(1.2, 2.1, 0.8, 1.3),
      mar = c(2.2, 3.1, 1.8, 0.7)
    )
  )
  graphics::plot.new()
  graphics::plot.window(c(-2, 3), c(-4, 5))
  graphics::par(xaxp = c(-2, 3, 5), yaxp = c(-4, 5, 9))
}

expect_graphics_state_unchanged <- function(
    draw, configure = stress_graphics_state) {
  plot.file <- tempfile(fileext = ".pdf")
  grDevices::pdf(plot.file, width = 6, height = 6)
  plot.device <- grDevices::dev.cur()
  on.exit({
    devices <- grDevices::dev.list()
    if (!is.null(devices) && plot.device %in% devices) {
      grDevices::dev.off(plot.device)
    }
    unlink(plot.file)
  }, add = TRUE)

  configure()
  before <- graphics::par(no.readonly = TRUE)
  set.seed(231)
  value <- force(draw)

  expect_identical(grDevices::dev.cur(), plot.device)
  expect_identical(graphics::par(no.readonly = TRUE), before)
  invisible(value)
}

test_that("plot.huge and huge.plot restore caller graphics state", {
  edge <- Matrix::Matrix(matrix(c(0, 1, 1, 0), 2, 2), sparse = TRUE)
  empty <- Matrix::Matrix(matrix(0, 2, 2), sparse = TRUE)
  fit <- structure(
    list(
      lambda = c(1, 0.5), sparsity = c(0, 1),
      path = list(empty, edge)
    ),
    class = "huge"
  )

  expect_graphics_state_unchanged(plot(fit, align = FALSE))
  expect_graphics_state_unchanged(
    invisible(capture.output(plot(fit, align = TRUE)))
  )
  expect_graphics_state_unchanged(huge.plot(edge, epsflag = FALSE))
})

test_that("generator, sim, select, and ROC plots restore graphics state", {
  edge <- Matrix::Matrix(matrix(c(0, 1, 1, 0), 2, 2), sparse = TRUE)
  sim <- structure(
    list(theta = edge, sigma = diag(2), sigmahat = diag(2)),
    class = "sim"
  )
  selected <- structure(
    list(
      cov.input = FALSE, refit = edge,
      lambda = c(1, 0.5), sparsity = c(0, 1),
      opt.lambda = 0.5, opt.sparsity = 1
    ),
    class = "select"
  )
  roc <- structure(
    list(fp = c(0, 0.25, 1), tp = c(0, 0.75, 1)),
    class = "roc"
  )
  truth <- matrix(0, 3, 3)
  truth[1, 2] <- truth[2, 1] <- 1
  complete <- 1 - diag(3)
  path <- list(matrix(0, 3, 3), complete)

  expect_graphics_state_unchanged(huge.generator(
    n = 8, d = 2, graph = "band",
    vis = TRUE, verbose = FALSE
  ))
  expect_graphics_state_unchanged(plot(sim))
  expect_graphics_state_unchanged(plot(selected))
  expect_graphics_state_unchanged(plot(roc))
  expect_graphics_state_unchanged(
    huge.roc(path, truth, verbose = FALSE)
  )
})

test_that("plotting restores custom figure, plot, margin, and outer regions", {
  roc <- structure(
    list(fp = c(0, 0.25, 1), tp = c(0, 0.75, 1)),
    class = "roc"
  )

  for (kind in c(
    "fig", "fin", "plt", "pin", "mai", "mar", "oma", "omi", "omd"
  )) {
    expect_graphics_state_unchanged(
      plot(roc),
      configure = function() custom_graphics_state(kind)
    )
  }
})

test_that("restoration does not mask an error before the first plot", {
  bad <- structure(
    list(
      theta = matrix("bad", 2, 2),
      sigma = diag(2), sigmahat = diag(2)
    ),
    class = "sim"
  )
  old.warn <- getOption("warn")
  on.exit(options(warn = old.warn), add = TRUE)
  options(warn = 2)

  expect_graphics_state_unchanged(
    expect_error(plot(bad), "numeric or logical")
  )
})

test_that("huge.plot EPS output preserves and returns to the caller device", {
  edge <- Matrix::Matrix(matrix(c(0, 1, 1, 0), 2, 2), sparse = TRUE)
  background.file <- tempfile(fileext = ".pdf")
  plot.file <- tempfile(fileext = ".pdf")
  output.dir <- tempfile(pattern = "huge-eps-")
  dir.create(output.dir)
  grDevices::pdf(background.file, width = 6, height = 6)
  background.device <- grDevices::dev.cur()
  grDevices::pdf(plot.file, width = 6, height = 6)
  caller.device <- grDevices::dev.cur()
  on.exit({
    devices <- grDevices::dev.list()
    if (!is.null(devices) && caller.device %in% devices) {
      grDevices::dev.off(caller.device)
    }
    devices <- grDevices::dev.list()
    if (!is.null(devices) && background.device %in% devices) {
      grDevices::dev.off(background.device)
    }
    unlink(background.file)
    unlink(plot.file)
    unlink(output.dir, recursive = TRUE)
  }, add = TRUE)

  stress_graphics_state()
  before <- graphics::par(no.readonly = TRUE)
  before.devices <- grDevices::dev.list()
  huge.plot(
    edge, epsflag = TRUE, graph.name = "state", cur.num = 1,
    location = output.dir
  )
  eps.file <- file.path(output.dir, "state1.eps")

  expect_identical(grDevices::dev.cur(), caller.device)
  expect_identical(grDevices::dev.list(), before.devices)
  expect_identical(graphics::par(no.readonly = TRUE), before)
  expect_true(file.exists(eps.file))
  expect_gt(file.info(eps.file)$size, 0)

  testthat::local_mocked_bindings(
    plot = function(...) stop("forced plot failure"),
    .package = "huge"
  )
  expect_error(
    huge.plot(
      edge, epsflag = TRUE, graph.name = "state", cur.num = 2,
      location = output.dir
    ),
    "forced plot failure"
  )
  expect_identical(grDevices::dev.cur(), caller.device)
  expect_identical(grDevices::dev.list(), before.devices)
  expect_identical(graphics::par(no.readonly = TRUE), before)
})

#-----------------------------------------------------------------------#
# Package: High-dimensional Undirected Graph Estimation                 #
# huge.npn(): nonparanormal(npn) transformation                         #
#-----------------------------------------------------------------------#

#' Nonparanormal(npn) transformation
#'
#' Implements the Gaussianization to help relax the assumption of normality.
#'
#' The nonparanormal extends Gaussian graphical models to semiparametric Gaussian copula models.Motivated by sparse additive models, the nonparanormal method estimates the Gaussian copula by marginally transforming the variables using smooth functions.Computationally, the estimation of a nonparanormal transformation is very efficient and only requires one pass of the data matrix.
#'
#' @param x The \code{n} by \code{d} data matrix representing \code{n} observations in \code{d} dimensions
#' @param npn.func The transformation function used in the npn transformation. If \code{npn.func = "truncation"}, the truncated ECDF is applied. If \code{npn.func = "shrinkage"}, the shrunken ECDF is applied. The default is \code{"shrinkage"}. If \code{npn.func = "skeptic"}, the nonparanormal skeptic is applied.
#' @param npn.thresh The truncation threshold used in nonparanormal transformation, ONLY applicable when \code{npn.func = "truncation"}. The default value is \code{1/(4*(n^0.25)*} \code{sqrt(pi*log(n)))}.
#' @param verbose If \code{verbose = FALSE}, tracing information printing is disabled. The default value is \code{TRUE}.
#' @param na.last for controlling the treatment of NAs. If TRUE, missing values in the data are put last; if FALSE, they are put first; if NA, they are removed; if "keep" they are kept with rank NA. See also \code{\link{rank}}.
#' @return
#' \item{data}{
#' A \code{d} by \code{d} nonparanormal correlation matrix if \code{npn.func = "skeptic"}, and A \code{n} by \code{d} data matrix representing \code{n} observations in \code{d} transformed dimensions otherwise.
#' }
#' @seealso \code{\link{huge}} and \code{\link{huge-package}}.
#' @examples
#' # generate nonparanormal data
#' L = huge.generator(graph = "cluster", g = 5)
#' L$data = L$data^5
#'
#' # transform the data using the shrunken ECDF
#' Q = huge.npn(L$data)
#'
#' # transform the non-Gaussian data using the truncated ECDF
#' Q = huge.npn(L$data, npn.func = "truncation")
#'
#' # estimate the nonparanormal correlation matrix using the skeptic estimator
#' Q = huge.npn(L$data, npn.func = "skeptic")
#' @export
huge.npn = function(x, npn.func = "shrinkage", npn.thresh = NULL, verbose = TRUE, na.last = "keep"){
  n = nrow(x)
  x.col = colnames(x)
  x.row = rownames(x)

  if(!npn.func %in% c("shrinkage", "truncation", "skeptic"))
    stop("npn.func must be one of \"shrinkage\", \"truncation\", \"skeptic\".")

  normalize.columns = function(z){
    z.sd = apply(z, 2, sd, na.rm = TRUE)
    z.sd[!is.finite(z.sd) | z.sd == 0] = 1
    sweep(z, 2, z.sd, "/")
  }

  if(verbose) cat(sprintf("Conducting the nonparanormal (npn) transformation via %s....",
    switch(npn.func, shrinkage = "shrunken ECDF", truncation = "truncated ECDF", skeptic = "skeptic")))

  if(npn.func == "shrinkage"){
    # Shrunken ECDF: rank/(n+1) keeps the transform inside qnorm's open support
    x = normalize.columns(qnorm(apply(x,2,rank,na.last=na.last)/(n+1)))
    dimnames(x) = list(x.row, x.col)
  }

  if(npn.func == "truncation"){
    if(is.null(npn.thresh)) npn.thresh = 1/(4*(n^0.25)*sqrt(pi*log(n)))
    x = normalize.columns(qnorm(pmin(pmax(apply(x,2,rank,na.last=na.last)/n, npn.thresh), 1-npn.thresh)))
    dimnames(x) = list(x.row, x.col)
  }

  if(npn.func == "skeptic"){
    # Spearman correlation = Pearson correlation of column ranks; computing it
    # as a BLAS crossprod of standardized ranks is ~5x faster than
    # cor(method = "spearman") at d = 2000. Fall back for NA input.
    if(anyNA(x)){
      rho = cor(x, method = "spearman", use = "pairwise.complete.obs")
    } else {
      rk = scale(apply(x, 2, rank))
      rho = crossprod(rk) / (n - 1)
      diag(rho) = 1
    }
    x = 2*sin(pi/6*rho)
    # d x d correlation matrix: variable names on both dimensions
    dimnames(x) = list(x.col, x.col)
  }

  if(verbose) cat("done.\n")
  return(x)
}

test_that("huge.npn shrinkage transformation works", {
  set.seed(60)
  L <- huge.generator(n = 80, d = 20, graph = "hub", verbose = FALSE)
  Q <- huge.npn(L$data, npn.func = "shrinkage", verbose = FALSE)

  expect_equal(dim(Q), dim(L$data))
  expect_true(all(is.finite(Q)))
  # columns should have unit variance (approximately)
  sds <- apply(Q, 2, sd)
  expect_true(all(abs(sds - 1) < 0.1))
})

test_that("huge.npn truncation transformation works", {
  set.seed(61)
  L <- huge.generator(n = 80, d = 20, graph = "hub", verbose = FALSE)
  Q <- huge.npn(L$data, npn.func = "truncation", verbose = FALSE)

  expect_equal(dim(Q), dim(L$data))
  expect_true(all(is.finite(Q)))
})

test_that("huge.npn skeptic returns a correlation matrix", {
  set.seed(62)
  L <- huge.generator(n = 80, d = 20, graph = "hub", verbose = FALSE)
  Q <- huge.npn(L$data, npn.func = "skeptic", verbose = FALSE)

  expect_equal(dim(Q), c(20, 20))
  expect_equal(Q, t(Q), tolerance = 1e-10)
  expect_true(all(abs(diag(Q) - 1) < 1e-10))
})

test_that("npn-transformed data works with estimation methods", {
  set.seed(63)
  L <- huge.generator(n = 80, d = 20, graph = "hub", verbose = FALSE)
  Q <- huge.npn(L$data^3, verbose = FALSE)  # nonlinear distortion
  fit <- huge(Q, method = "mb", verbose = FALSE)

  expect_s3_class(fit, "huge")
  expect_true(all(diff(fit$sparsity) >= -1e-10))
})

# Regression tests pinning bugs fixed in 2.0.0. Each test names the defect it
# guards; if one fails, the corresponding fix has been reverted or broken.

test_that("tiger icov is exactly symmetric (2.0.0: in-place symmetrization bug)", {
  set.seed(42)
  L = huge.generator(n = 60, d = 30, graph = "hub", verbose = FALSE)
  fit = huge(L$data, method = "tiger", nlambda = 5, verbose = FALSE)
  for (ic in fit$icov)
    expect_identical(max(abs(as.matrix(ic) - t(as.matrix(ic)))), 0)
})

test_that("glasso precision, path, df, sparsity, and loglik agree", {
  S = matrix(c(
    1, .517180, -.517349, .326007, .263074, -.434894, -.594540,
    .517180, 1, -.007732, .807446, -.171208, -.164722, .110568,
    -.517349, -.007732, 1, .108058, -.413103, .652356, .356858,
    .326007, .807446, .108058, 1, .131565, -.283979, -.107923,
    .263074, -.171208, -.413103, .131565, 1, -.473447, -.695847,
    -.434894, -.164722, .652356, -.283979, -.473447, 1, .390406,
    -.594540, .110568, .356858, -.107923, -.695847, .390406, 1
  ), 7, 7, byrow = TRUE)
  fit = huge(S, method = "glasso", nlambda = 8,
             lambda.min.ratio = .005, cov.output = TRUE,
             verbose = FALSE)
  denom = nrow(S) * (nrow(S) - 1)

  for (i in seq_along(fit$path)) {
    path = as.matrix(fit$path[[i]])
    precision = as.matrix(fit$icov[[i]])
    expect_equal(precision, t(precision), tolerance = 0, info = paste("lambda", i))
    expect_equal(path, t(path), tolerance = 0, info = paste("lambda", i))
    support = precision != 0
    diag(support) = FALSE
    expect_identical(path != 0, support, info = paste("lambda", i))
    edge.count = sum(path[upper.tri(path)] != 0)
    expect_identical(fit$df[i], as.integer(edge.count))
    expect_equal(fit$sparsity[i], 2 * edge.count / denom, tolerance = 0)
    determinant.info = determinant(precision, logarithm = TRUE)
    expect_identical(determinant.info$sign, 1L)
    expect_lte(max(rowSums(abs(fit$cov[[i]] %*% precision - diag(7)))),
               1e-2)
    direct = as.numeric(determinant.info$modulus) -
      sum(diag(S %*% precision))
    expect_equal(fit$loglik[i], direct, tolerance = 1e-10)
  }
})

test_that("RIC selection is scale-invariant (2.0.0: unstandardized data bug)", {
  set.seed(9)
  L = huge.generator(n = 100, d = 40, graph = "hub", verbose = FALSE)
  fit1 = huge(L$data, method = "mb", nlambda = 8, verbose = FALSE)
  fit2 = huge(L$data * 100, method = "mb", nlambda = 8, verbose = FALSE)
  set.seed(101); s1 = huge.select(fit1, criterion = "ric", verbose = FALSE)
  set.seed(101); s2 = huge.select(fit2, criterion = "ric", verbose = FALSE)
  expect_equal(s1$opt.lambda, s2$opt.lambda, tolerance = 1e-12)
  expect_equal(s1$opt.sparsity, s2$opt.sparsity)
})

test_that("R raw-data estimators preserve correlation at extreme finite scales", {
  x = cbind(c(-1, 0, 1), c(-1, 1, 0))
  minimum = .Machine$double.xmin * .Machine$double.eps
  inputs = list(
    x * 1e308,
    x * 1e-200,
    x * minimum,
    sweep(x, 2, c(1e308, 1e-200), "*")
  )

  for (method in c("ct", "mb", "glasso")) {
    explicit.reference = huge(
      x, method = method, lambda = .4, verbose = FALSE
    )
    automatic.reference = huge(
      x, method = method, nlambda = 3,
      lambda.min.ratio = .5, verbose = FALSE
    )

    for (current in inputs) {
      expect_no_warning({
        explicit = huge(
          current, method = method, lambda = .4, verbose = FALSE
        )
        automatic = huge(
          current, method = method, nlambda = 3,
          lambda.min.ratio = .5, verbose = FALSE
        )
      })
      expect_false(explicit$cov.input)
      expect_equal(explicit$lambda, explicit.reference$lambda, tolerance = 0)
      expect_equal(explicit$sparsity, explicit.reference$sparsity, tolerance = 0)
      expect_equal(automatic$lambda, automatic.reference$lambda, tolerance = 1e-14)
      expect_equal(
        automatic$sparsity, automatic.reference$sparsity, tolerance = 0
      )
      for (i in seq_along(explicit$path))
        expect_identical(
          as.matrix(explicit$path[[i]]),
          as.matrix(explicit.reference$path[[i]])
        )
      for (i in seq_along(automatic$path))
        expect_identical(
          as.matrix(automatic$path[[i]]),
          as.matrix(automatic.reference$path[[i]])
        )
    }
  }
})

test_that("R fast correlation preserves adjacent maximum-scale ULPs", {
  maximum = .Machine$double.xmax
  ulp = 2^971
  indices = cbind(
    first = 0:6,
    second = c(0, 1, 3, 6, 2, 5, 4)
  )
  x = maximum - ulp * indices

  expect_equal(huge:::.huge_fast_cor(x), cor(indices), tolerance = 1e-14)

  set.seed(3)
  n = sample(3:200, 1)
  values = rnorm(n)
  repeated = cbind(positive = values, duplicate = values, negative = -values)
  correlation = huge:::.huge_fast_cor(repeated)
  expect_identical(dimnames(correlation), list(colnames(repeated), colnames(repeated)))
  expect_lte(max(abs(correlation)), 1)
  thresholded = huge.ct(repeated, lambda = 1, verbose = FALSE)
  expect_identical(thresholded$sparsity, 0)
})

test_that("npn output columns have unit sd and names survive (2.0.0 fixes)", {
  set.seed(3)
  x = matrix(rnorm(80 * 12), 80, 12,
             dimnames = list(paste0("r", 1:80), paste0("c", 1:12)))
  for (f in c("shrinkage", "truncation")) {
    z = huge.npn(x, npn.func = f, verbose = FALSE)
    expect_equal(unname(apply(z, 2, sd)), rep(1, 12), tolerance = 1e-12)
    expect_identical(dimnames(z), dimnames(x))
  }
  # skeptic on named input errored before 2.0.0 (n-length rownames on d x d)
  k = huge.npn(x, npn.func = "skeptic", verbose = FALSE)
  expect_identical(dim(k), c(12L, 12L))
  expect_identical(rownames(k), colnames(x))
  expect_identical(colnames(k), colnames(x))
  # unknown npn.func silently returned x unchanged before 2.0.0
  expect_error(huge.npn(x, npn.func = "bogus", verbose = FALSE))
})

test_that("generator sigma is a correlation matrix with symmetric omega", {
  set.seed(7)
  L = huge.generator(n = 50, d = 25, graph = "hub", verbose = FALSE)
  expect_equal(unname(diag(L$sigma)), rep(1, 25))
  expect_identical(max(abs(L$omega - t(L$omega))), 0)
  expect_lt(max(abs(L$sigma %*% L$omega - diag(25))), 1e-10)
})

test_that("select sets opt.cov only for glasso with cov (2.0.0: partial-match bug)", {
  set.seed(3)
  x = matrix(rnorm(120 * 30), 120, 30)
  for (m in c("mb", "ct", "tiger")) {
    fit = huge(x, method = m, nlambda = 5, verbose = FALSE)
    set.seed(9); s = huge.select(fit, criterion = "ric", verbose = FALSE)
    expect_null(s[["opt.cov"]])
  }
  fit = huge(x, method = "glasso", nlambda = 5, cov.output = TRUE, verbose = FALSE)
  set.seed(9); s = huge.select(fit, criterion = "ric", verbose = FALSE)
  expect_false(is.null(s[["opt.cov"]]))
})

test_that("huge() rejects unknown method (2.0.0: returned half-empty object)", {
  x = matrix(rnorm(200), 50, 4)
  expect_error(huge(x, method = "bogus", verbose = FALSE))
})

test_that("huge.select preserves its historical positional verbose argument", {
  expect_identical(
    tail(names(formals(huge.select)), 3),
    c("rep.num", "verbose", "num.cores")
  )

  set.seed(3)
  x = matrix(rnorm(60 * 10), 60, 10)
  fit = huge(x, method = "mb", nlambda = 4, verbose = FALSE)
  set.seed(19)
  named = huge.select(fit, criterion = "ric", rep.num = 2, verbose = FALSE)
  set.seed(19)
  positional = huge.select(fit, "ric", 0.5, 0.1, NULL, 2, FALSE)
  expect_identical(positional, named)
})

test_that("stars num.cores > 1 reproduces serial results exactly", {
  skip_on_os("windows")
  with.check.core.limit = function(code) {
    previous = Sys.getenv("_R_CHECK_LIMIT_CORES_", unset = NA_character_)
    on.exit({
      if(is.na(previous))
        Sys.unsetenv("_R_CHECK_LIMIT_CORES_")
      else
        do.call(
          Sys.setenv,
          setNames(list(previous), "_R_CHECK_LIMIT_CORES_")
        )
    }, add = TRUE)
    do.call(
      Sys.setenv,
      setNames(list("TRUE"), "_R_CHECK_LIMIT_CORES_")
    )
    force(code)
  }

  set.seed(3)
  x = matrix(rnorm(120 * 30), 120, 30)
  fit = huge(x, method = "mb", nlambda = 5, verbose = FALSE)
  set.seed(42); s1 = huge.select(fit, criterion = "stars", rep.num = 4,
                                 num.cores = 1, verbose = FALSE)
  set.seed(42); expect_warning(
    s2 <- huge.select(fit, criterion = "stars", rep.num = 4,
                      num.cores = 2, verbose = FALSE),
    "OpenMP.*BLAS"
  )
  s3 = with.check.core.limit({
    set.seed(42)
    expect_warning(
      huge.select(fit, criterion = "stars", rep.num = 4,
                  num.cores = 20, verbose = FALSE),
      "OpenMP.*BLAS"
    )
  })
  expect_identical(s1$opt.index, s2$opt.index)
  expect_equal(s1$variability, s2$variability, tolerance = 0)
  expect_identical(s1$opt.index, s3$opt.index)
  expect_equal(s1$variability, s3$variability, tolerance = 0)
})

test_that("stars serial accumulation matches the batch reference exactly", {
  set.seed(90)
  x = matrix(rnorm(50 * 8), 50, 8)
  fit = huge(x, method = "ct", nlambda = 4, verbose = FALSE)
  rep.num = 4
  ratio = 0.75
  subsample.size = floor(nrow(x) * ratio)

  # Reproduce the historical implementation: retain every subsample path,
  # then merge replication-by-replication within each lambda value.
  set.seed(901)
  ind.list = lapply(seq_len(rep.num), function(i)
    sample(seq_len(nrow(x)), subsample.size, replace = FALSE))
  paths = lapply(ind.list, function(ind)
    huge.ct(x[ind, ], lambda = fit$lambda, verbose = FALSE)$path)
  expected.seed = .Random.seed

  expected.merge = lapply(seq_along(fit$lambda), function(lambda.idx) {
    count = Matrix::Matrix(0, ncol(x), ncol(x))
    for(rep.idx in seq_len(rep.num))
      count = count + paths[[rep.idx]][[lambda.idx]]
    count / rep.num
  })
  expected.variability = vapply(expected.merge, function(m) {
    4 * (sum(m) - sum(m^2)) / (ncol(x) * (ncol(x) - 1))
  }, numeric(1))
  expect_true(any(expected.variability > 0))
  stars.cross = which(expected.variability >= 0.1)
  expected.index = if(length(stars.cross) == 0) {
    length(fit$lambda)
  } else {
    max(stars.cross[1] - 1, 1)
  }

  set.seed(901)
  selected = huge.select(
    fit, criterion = "stars", stars.subsample.ratio = ratio,
    rep.num = rep.num, num.cores = 1, verbose = FALSE
  )
  selected.seed = .Random.seed

  for(lambda.idx in seq_along(expected.merge))
    expect_equal(as.matrix(selected$merge[[lambda.idx]]),
                 as.matrix(expected.merge[[lambda.idx]]),
                 tolerance = 0, info = paste("lambda", lambda.idx))
  expect_true(all(vapply(selected$merge, inherits, logical(1),
                         what = "sparseMatrix")))
  expect_equal(selected$variability, expected.variability, tolerance = 0)
  expect_identical(selected$opt.index, expected.index)
  expect_equal(selected$refit, fit$path[[expected.index]], tolerance = 0)
  expect_identical(selected.seed, expected.seed)
})

test_that("mb/tiger sparse paths have sorted dgCMatrix indices (core collect_sorted)", {
  set.seed(5)
  x = matrix(rnorm(100 * 30), 100, 30)
  for (m in c("mb", "tiger")) {
    fit = huge(x, method = m, nlambda = 5, verbose = FALSE)
    for (p in fit$path) expect_true(validObject(p, complete = TRUE))
  }
})

test_that("solvers are silent normally and glasso rejects an uncertified limit case", {
  set.seed(1)
  x = matrix(rnorm(100 * 30), 100, 30)
  for (m in c("mb", "glasso", "tiger"))
    expect_no_warning(huge(x, method = m, nlambda = 5, verbose = FALSE))

  # Pathological: n < d with 5 latent factors and an absurdly small lambda.
  # The old iteration-limit result is not a usable inverse pair, so glasso
  # must reject it instead of returning it with only a warning.
  set.seed(50)
  n = 40; d = 60
  base = matrix(rnorm(n * 5), n, 5)
  xp = base[, sample(5, d, replace = TRUE)] + matrix(rnorm(n * d, sd = 0.05), n, d)
  expect_error(
    huge(scale(xp), method = "glasso", lambda = 0.001, verbose = FALSE),
    "inconsistent precision and covariance"
  )
})

test_that("ROC rejects truth matrices with only one edge class", {
  path = list(matrix(0, 3, 3))
  expect_error(huge.roc(path, matrix(0, 3, 3), verbose = FALSE), "ROC/AUC")

  complete = matrix(1, 3, 3)
  diag(complete) = 0
  expect_error(huge.roc(path, complete, verbose = FALSE), "ROC/AUC")
})

test_that("ROC AUC is invariant to equal-FPR order and duplicates", {
  graph = function(edges) {
    value = matrix(0, 4, 4)
    for(index in seq_len(nrow(edges))) {
      left = edges[index, 1]
      right = edges[index, 2]
      value[left, right] = 1
      value[right, left] = 1
    }
    value
  }

  truth = graph(matrix(c(1, 2, 1, 3), ncol = 2, byrow = TRUE))
  low = graph(matrix(c(3, 4), ncol = 2))
  high = graph(matrix(
    c(1, 2, 1, 3, 3, 4), ncol = 2, byrow = TRUE
  ))
  complete = matrix(1, 4, 4)
  diag(complete) = 0

  paths = list(
    list(low, high, complete),
    list(high, low, complete),
    list(low, high, high, complete)
  )
  auc = vapply(
    paths,
    function(path) huge.roc(path, truth, verbose = FALSE)$AUC,
    numeric(1)
  )

  expect_equal(auc, rep(.75, 3))
})

#-----------------------------------------------------------------------#
# Package: High-dimensional Undirected Graph Estimation                 #
# huge(): Draw ROC Curve for a solution path                            #
#         The ground truth is required                                  #
#-----------------------------------------------------------------------#

#' Draw ROC Curve for a graph path
#'
#' Draws ROC curve for a graph path according to the true graph structure.
#'
#' To avoid the horizontal oscillation, false positive rates is automatically sorted in the ascent order and true positive rates also follow the same order.
#'
#' @param path A graph path.
#' @param theta The true graph structure, containing at least one edge and one absent off-diagonal edge.
#' @param verbose If \code{verbose = FALSE}, tracing information printing is disabled. The default value is \code{TRUE}.
#' @note ROC/AUC is undefined when \code{theta} contains only edges or only non-edges, so those one-class truth matrices are rejected. For a lasso regression, the number of nonzero coefficients is at most \code{n-1}. If \code{d>>n}, even when regularization parameter is very small, the estimated graph may still be sparse. In this case, the AUC may not be a good choice to evaluate the performance.
#' @return
#' An object with S3 class "roc" is returned:
#'   \item{F1}{
#'     The F1 scores along the graph path.
#'   }
#' \item{tp}{
#'   The true positive rates along the graph path
#' }
#' \item{fp}{
#'   The false positive rates along the graph paths
#' }
#' \item{AUC}{
#'   Area under the ROC curve
#' }
#' @seealso \code{\link{huge}} and \code{\link{huge-package}}.
#' @examples
#' #generate data
#' L = huge.generator(d = 200, graph = "cluster", prob = 0.3)
#' out1 = huge(L$data)
#'
#' #draw ROC curve
#' Z1 = huge.roc(out1$path,L$theta)
#'
#' #Maximum F1 score
#' max(Z1$F1)
#' @export
huge.roc = function(path, theta, verbose = TRUE){

  ROC = list()

  theta = as.matrix(theta)
  d = ncol(theta)
  # Off-diagonal true/null edge masks, computed once; the per-lambda work is
  # then two logical-AND sums instead of dense double products + diag resets.
  offdiag = !diag(TRUE, d)
  pos.mask = (theta != 0) & offdiag
  neg.mask = (theta == 0) & offdiag
  pos.total = sum(pos.mask)
  neg.total = sum(neg.mask)
  if(pos.total == 0 || neg.total == 0)
    stop(paste(
      "theta must contain at least one edge and at least one absent",
      "off-diagonal edge; ROC/AUC is undefined for a one-class truth."
    ))

  if(verbose) cat("Computing F1 scores, false positive rates and true positive rates....")
  ROC$tp = rep(0,length(path))
     ROC$fp = rep(0,length(path))
     ROC$F1 = rep(0,length(path))
     for (r in 1:length(path)){
       est = as.matrix(path[[r]]) != 0
    tp.count = sum(est & pos.mask)
    ROC$tp[r] <- tp.count/pos.total
    fp.count = sum(est & neg.mask)
    ROC$fp[r] <- fp.count/neg.total

    pred.count = tp.count + fp.count
    precision = if(pred.count > 0) tp.count / pred.count else 0
    recall = ROC$tp[r]
    ROC$F1[r] = if(precision + recall > 0) 2*precision*recall/(precision+recall) else 0
  }
  if(verbose) cat("done.\n")

  ord.fp = order(ROC$fp, ROC$tp)

  tmp1 = ROC$fp[ord.fp]
  tmp2 = ROC$tp[ord.fp]
  old.par = .huge_graphics_state()
  on.exit(.huge_restore_graphics_state(old.par), add = TRUE)
  par(mfrow = c(1,1))
  plot(tmp1,tmp2,type="b",main = "ROC Curve", xlab = "False Positive Rate", ylab = "True Positive Rate",ylim = c(0,1))
  ROC$AUC = sum(diff(tmp1)*(tmp2[-1]+tmp2[-length(tmp2)]))/2

  class(ROC) = "roc"
  return(ROC)
}

#' Print function for S3 class "roc"
#'
#' Print the information about true positive rates, false positive rates, the area under curve and maximum F1 score.
#'
#' @param x An object with S3 class \code{"roc"}.
#' @param \dots System reserved (No specific usage)
#' @seealso \code{\link{huge.roc}}
#' @export
print.roc = function(x, ...){
  cat("True Positive Rate: from",min(x$tp),"to",max(x$tp),"\n")
  cat("False Positive Rate: from",min(x$fp),"to",max(x$fp),"\n")
  cat("Area under Curve:",x$AUC,"\n")
  cat("Maximum F1 Score:",max(x$F1),"\n")
}

#' Plot function for S3 class "roc"
#'
#' Plot the ROC curve for an object with S3 class \code{"roc"}.
#'
#' @param x An object with S3 class \code{"roc"}
#' @param \dots System reserved (No specific usage)
#' @seealso \code{\link{huge.roc}}
#' @export
plot.roc = function(x, ...){
  ord.fp = order(x$fp, x$tp)
  old.par = .huge_graphics_state()
  on.exit(.huge_restore_graphics_state(old.par), add = TRUE)
  par(mfrow = c(1,1))
  plot(x$fp[ord.fp],x$tp[ord.fp],type="b",main = "ROC Curve", xlab = "False Positive Rate", ylab = "True Positive Rate",ylim = c(0,1))
}

#-----------------------------------------------------------------------#
# Package: High-dimensional Undirected Graph Estimation                 #
# huge.tiger(): Tuning-insensitive graph estimation                     #
#-----------------------------------------------------------------------#

#' Tuning-insensitive graph estimation
#'
#' See more details in \code{\link{huge}}
#' @details Raw observations are centered and normalized, while covariance
#'   input is converted to a correlation matrix, inside the shared C++ core.
#'   When \code{lambda = NULL}, the same native correlation matrix determines
#'   the returned default lambda path. A user-supplied path must be finite,
#'   strictly positive, and non-increasing; tied values are allowed and are
#'   used unchanged. If smaller generated values cannot be certified to the solver's
#'   KKT tolerance, the function warns and returns the longest certified path
#'   prefix. A user-supplied value that cannot be certified raises an error.
#' @param x There are 2 options: (1) \code{x} is an \code{n} by \code{d} data matrix (2) a \code{d} by \code{d} sample covariance matrix. The program automatically identifies the input matrix by checking the symmetry. (\code{n} is the sample size and \code{d} is the dimension).
#' @param lambda A numeric scalar or non-empty one-dimensional numeric input
#'   defining a finite, strictly positive, non-increasing regularization path.
#'   Tied values are allowed. Leave \code{lambda = NULL} to generate
#'   the path from \code{nlambda} and \code{lambda.min.ratio} in C++.
#' @param nlambda The number of regularization/thresholding parameters. The default value is \code{20} for \code{method = "ct"} and \code{10} for \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}.
#' @param lambda.min.ratio If \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}, it is the smallest value for \code{lambda}, as a fraction of the upperbound (\code{MAX}) of the regularization/thresholding parameter which makes all estimates equal to \code{0}. The program can automatically generate \code{lambda} as a sequence of length = \code{nlambda} starting from \code{MAX} to \code{lambda.min.ratio*MAX} in log scale. If \code{method = "ct"}, it is the largest sparsity level for estimated graphs. The program can automatically generate \code{lambda} as a sequence of length = \code{nlambda}, which makes the sparsity level of the graph path increases from \code{0} to \code{lambda.min.ratio} evenly.The default value is \code{0.1} when \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}, and 0.05 when \code{method = "ct"}.
#' @param sym Symmetrize the output graphs. If \code{sym = "and"}, the edge between node \code{i} and node \code{j} is selected ONLY when both node \code{i} and node \code{j} are selected as neighbors for each other. If \code{sym = "or"}, the edge is selected when either node \code{i} or node \code{j} is selected as the neighbor for each other. The default value is \code{"or"}. ONLY applicable when \code{method = "mb"} or \code{"tiger"}.
#' @param verbose If \code{verbose = FALSE}, tracing information printing is disabled. The default value is \code{TRUE}.
#' @param input.type How to interpret \code{x}: \code{"auto"} preserves
#'   symmetry-based detection, \code{"data"} forces an observation matrix,
#'   and \code{"covariance"} requires a square covariance or correlation
#'   matrix. Correlation construction, covariance validation, and automatic
#'   lambda selection then occur together in C++.
#' @seealso \code{\link{huge}}, and \code{\link{huge-package}}.
#' @export
huge.tiger = function(x, lambda = NULL, nlambda = NULL, lambda.min.ratio = NULL, sym = "or", verbose = TRUE, input.type = "auto")
{
		sym = .huge_validate_sym(sym)
		inp = .huge_validate_estimation_input(
			x, input.type = input.type, prepare.covariance = FALSE
		)
		d = inp$d
		cov.input = inp$cov.input
		if(cov.input && verbose)
			cat("The input is identified as the covariance matrix.\n")
		fit = list()
		fit$cov.input = cov.input

		if(!is.null(lambda))
		{
			lambda = .huge_validate_lambda(lambda)
			if(length(lambda) > 1L &&
			   any(lambda[-1L] > lambda[-length(lambda)]))
				stop("lambda must be non-increasing for method = \"tiger\"; tied values are allowed.")
			nlambda = length(lambda)
			# The native fixed-lambda branch does not use this argument.
			lambda.min.ratio = 0.1
		}
		else
		{
			if(is.null(nlambda))
				nlambda = 10
			else
				nlambda = .huge_validate_positive_integer(nlambda, "nlambda")
			if(is.null(lambda.min.ratio))
				lambda.min.ratio = 0.1
			else
				lambda.min.ratio = .huge_validate_ratio(lambda.min.ratio)
		}

	if(verbose)
	{
	  cat("Conducting graph estimation through a tuning-insensitive approach (tiger)....")
	  flush.console()
		}
		fit$idx_mat = NULL
		fit$scr = FALSE
		requested.nlambda = nlambda
		out = .Call("_huge_SPMBgraphsqrtFit", inp$input, lambda, as.integer(nlambda), d,
		            cov.input, lambda.min.ratio, PACKAGE = "huge")
		lambda = out$lambda
		nlambda = length(lambda)

	if (isTRUE(out$path_truncated))
	  warning(sprintf(
	    "tiger returned the %d-value certified prefix of the %d-value native lambda path; smaller lambda values did not converge or were numerically degenerate.",
	    nlambda, requested.nlambda))
	else if (isTRUE(out$hit_max_iter))
	  warning("tiger solver reached its iteration limit; estimates may not be fully converged.")

	# The core emits each column's row indices in ascending order (see
	# collect_sorted in huge_core.cpp), as dgCMatrix requires.
		nnz = out$col_cnz[d + 1]
		x_values = if(nnz > 0) out$x[seq_len(nnz)] else numeric(0)
		i_values = if(nnz > 0) out$row_idx[seq_len(nnz)] else integer(0)
	  G = new("dgCMatrix", Dim = as.integer(c(d*nlambda,d)),
	          x = as.vector(x_values), p = as.integer(out$col_cnz),
	          i = as.integer(i_values))

	fit$beta = list()
	fit$path = list()
	fit$df = matrix(0,d,nlambda)
	fit$sparsity = rep(0,nlambda)
	for(i in 1:nlambda)
	{
		fit$beta[[i]] = G[((i-1)*d+1):(i*d),,drop = FALSE]
		fit$path[[i]] = abs(fit$beta[[i]])
		fit$df[,i] = Matrix::colSums(fit$path[[i]] != 0)

		# Matrix::t on the sparse matrix (not as.matrix) keeps the update sparse
		if(sym == "or")
			fit$path[[i]] = sign(fit$path[[i]] + Matrix::t(fit$path[[i]]))
		if(sym == "and")
			fit$path[[i]] = sign(fit$path[[i]] * Matrix::t(fit$path[[i]]))
		fit$sparsity[i] = if(d <= 1) 0 else sum(fit$path[[i]])/d/(d-1)
	}
	fit$icov = out$icov
 	fit$lambda = lambda

	if(verbose)
 	{
 		cat("done\n")
      flush.console()
  }

 	return(fit)
}

test_that("huge.select with ebic works for glasso", {
  set.seed(50)
  L <- huge.generator(n = 80, d = 20, graph = "hub", verbose = FALSE)
  fit <- huge(L$data, method = "glasso", verbose = FALSE)
  sel <- huge.select(fit, criterion = "ebic", verbose = FALSE)

  expect_s3_class(sel, "select")
  expect_true(!is.null(sel$refit))
  expect_true(!is.null(sel$opt.lambda))
  expect_true(!is.null(sel$opt.index))
  expect_true(sel$opt.lambda > 0)
  expect_true(sel$opt.index >= 1 && sel$opt.index <= length(sel$lambda))
  expect_equal(dim(sel$refit), c(20, 20))
})

test_that("huge.select with ric works for mb", {
  set.seed(51)
  L <- huge.generator(n = 80, d = 20, graph = "hub", verbose = FALSE)
  fit <- huge(L$data, method = "mb", verbose = FALSE)
  sel <- huge.select(fit, criterion = "ric", verbose = FALSE)

  expect_s3_class(sel, "select")
  expect_true(!is.null(sel$refit))
  expect_true(sel$opt.lambda > 0)
})

test_that("RIC is invariant to extreme finite uniform data scales", {
  x = cbind(
    c(-2, -1, 0, 1, 2, 3),
    c(3, -1, 2, -2, 1, 0),
    c(0, 2, -1, 3, -2, 1)
  )
  selected = lapply(c(1, 1e-300, 1e300), function(factor) {
    fit = huge(
      x * factor, method = "ct", lambda = c(.8, .2), verbose = FALSE
    )
    huge.select(
      fit, criterion = "ric", rep.num = nrow(x), verbose = FALSE
    )
  })

  expect_true(all(vapply(
    selected, function(value) is.finite(value$opt.lambda), logical(1)
  )))
  expect_equal(
    vapply(selected, function(value) value$opt.lambda, numeric(1)),
    rep(selected[[1]]$opt.lambda, 3),
    tolerance = 1e-14
  )
})

test_that("RIC handles a zero lower endpoint without weakening solvers", {
  x = cbind(
    c(-1, -1, 1, 1),
    c(-1, -1, 1, 1)
  )

  for(method in c("ct", "mb", "glasso", "tiger")) {
    fit = huge(
      x, method = method, lambda = 1, verbose = FALSE
    )
    if(method %in% c("glasso", "tiger")) {
      selected = expect_warning(
        huge.select(
          fit, criterion = "ric", rep.num = nrow(x), verbose = FALSE
        ),
        "RIC selected lambda = 0.*original fitted path",
        info = method
      )
      nearest = which.min(abs(fit$lambda))
      expect_identical(
        as.matrix(selected$refit), as.matrix(fit$path[[nearest]]),
        info = method
      )
    } else {
      selected = expect_no_warning(
        huge.select(
          fit, criterion = "ric", rep.num = nrow(x), verbose = FALSE
        )
      )
      expect_equal(sum(selected$refit), 2, info = method)
    }
    expect_identical(selected$opt.lambda, 0)
    expect_true(is.finite(selected$opt.sparsity), info = method)
  }
})

test_that("RIC zero does not erase a representable weak correlation", {
  first = c(-1, -1, 1, 1)
  orthogonal = c(-1, 1, -1, 1)
  # Below the former fixed 8-epsilon cutoff, but safely above the
  # pair-specific dot-product roundoff bound.
  x = cbind(first, orthogonal + 1.5e-15 * first)
  fit = huge(
    x, method = "ct", lambda = .5, verbose = FALSE
  )
  selected = huge.select(
    fit, criterion = "ric", rep.num = nrow(x), verbose = FALSE
  )

  expect_identical(selected$opt.lambda, 0)
  expect_equal(sum(selected$refit), 2)
})

test_that("huge.select with stars works for mb", {
  set.seed(52)
  L <- huge.generator(n = 80, d = 20, graph = "hub", verbose = FALSE)
  fit <- huge(L$data, method = "mb", verbose = FALSE)
  sel <- huge.select(fit, criterion = "stars", rep.num = 5, verbose = FALSE)

  expect_s3_class(sel, "select")
  expect_true(!is.null(sel$refit))
  expect_true(!is.null(sel$variability))
  expect_equal(length(sel$variability), length(sel$lambda))
  expect_true(all(sel$variability >= 0))
})

test_that("huge.select keeps TIGER RIC and rejects unsafe StARS grids", {
  set.seed(321)
  x = matrix(rnorm(120 * 20), 120, 20)
  fit = huge(x, method = "tiger", nlambda = 4,
             lambda.min.ratio = 0.4, verbose = FALSE)

  expect_error(
    huge.select(fit, criterion = "stars", rep.num = 3,
                verbose = FALSE),
    "TIGER.*StARS|common certified prefix"
  )

  set.seed(7)
  ric = huge.select(fit, criterion = "ric", rep.num = 10,
                    verbose = FALSE)
  expect_true(is.finite(ric$opt.lambda) && ric$opt.lambda > 0)
  expect_equal(dim(ric$refit), c(20, 20))
})

test_that("huge.select uses method-specific default criteria", {
  set.seed(53)
  x <- matrix(rnorm(100 * 10), nrow = 100)
  expected <- c(mb = "ric", ct = "stars", glasso = "ebic", tiger = "ric")

  for (method in names(expected)) {
    fit <- huge(x, method = method, nlambda = 3,
                lambda.min.ratio = .5, verbose = FALSE)
    set.seed(531)
    selected <- huge.select(
      fit, rep.num = 3, num.cores = 1, verbose = FALSE
    )
    expect_identical(selected$criterion, unname(expected[[method]]),
                     info = method)
    expect_false(is.null(selected$refit), info = method)
  }
})

test_that("huge.select rejects malformed estimates and criteria", {
  set.seed(54)
  x <- matrix(rnorm(80 * 8), nrow = 80)
  fit <- huge(x, method = "mb", nlambda = 3, verbose = FALSE)

  expect_error(huge.select(NULL, verbose = FALSE), "est")
  expect_error(huge.select(list(), verbose = FALSE), "est")

  incomplete <- fit
  incomplete$path <- NULL
  expect_error(huge.select(incomplete, verbose = FALSE), "est|path")

  for (criterion in list("bogus", c("ric", "stars"), 1, NA_character_)) {
    expect_error(
      huge.select(fit, criterion = criterion, verbose = FALSE),
      "criterion"
    )
  }
  expect_error(
    huge.select(fit, criterion = "ebic", verbose = FALSE),
    "ebic|glasso"
  )
})

test_that("huge.select stops cleanly for covariance input", {
  set.seed(55)
  x <- matrix(rnorm(80 * 8), nrow = 80)
  fit <- huge(cor(x), method = "glasso", nlambda = 3, verbose = FALSE)
  original.class <- class(fit)

  expect_error(
    huge.select(fit, verbose = FALSE),
    "covariance matrix"
  )
  expect_identical(class(fit), original.class)
  expect_s3_class(fit, "huge")
})

test_that("huge.select validates only active criterion parameters", {
  set.seed(56)
  x <- matrix(rnorm(80 * 8), nrow = 80)
  mb <- huge(x, method = "mb", nlambda = 3, verbose = FALSE)
  glasso <- huge(x, method = "glasso", nlambda = 3, verbose = FALSE)

  for (rep.num in list(0, -1, 1.5, NA_real_, Inf, c(2, 3), TRUE)) {
    expect_error(
      huge.select(
        mb, criterion = "ric", rep.num = rep.num, verbose = FALSE
      ),
      "rep.num"
    )
  }
  expect_error(
    huge.select(
      mb, criterion = "stars", rep.num = 2,
      num.cores = 0, verbose = FALSE
    ),
    "num.cores"
  )
  expect_error(
    huge.select(
      mb, criterion = "stars", rep.num = 2,
      stars.thresh = 0, verbose = FALSE
    ),
    "stars.thresh"
  )
  expect_error(
    huge.select(
      mb, criterion = "stars", rep.num = 2,
      stars.subsample.ratio = 0, verbose = FALSE
    ),
    "stars.subsample.ratio"
  )
  expect_error(
    huge.select(
      mb, criterion = "stars", rep.num = 2,
      stars.subsample.ratio = 1e-6, verbose = FALSE
    ),
    "stars.subsample.ratio"
  )
  expect_error(
    huge.select(
      glasso, criterion = "ebic", ebic.gamma = NA_real_,
      verbose = FALSE
    ),
    "ebic.gamma"
  )

  expect_no_error(huge.select(
    glasso, criterion = "ebic", rep.num = 0, num.cores = 0,
    stars.thresh = 0, stars.subsample.ratio = 0, verbose = FALSE
  ))
  expect_no_error(huge.select(
    mb, criterion = "ric", rep.num = 2, num.cores = 0,
    ebic.gamma = NA_real_, stars.thresh = 0,
    stars.subsample.ratio = 0, verbose = FALSE
  ))
  expect_no_error(huge.select(
    mb, criterion = "stars", rep.num = 2, num.cores = 1,
    ebic.gamma = NA_real_, verbose = FALSE
  ))
})

test_that("RIC handles an exact zero-correlation boundary", {
  x <- cbind(
    c(-1, -1, 1, 1),
    c(-1, 1, -1, 1)
  )

  for(method in c("ct", "mb", "glasso", "tiger")) {
    fit <- huge(x, method = method, nlambda = 3, verbose = FALSE)
    selected <- expect_no_warning(huge.select(
      fit, criterion = "ric", rep.num = nrow(x), verbose = FALSE
    ))

    expect_equal(selected$opt.lambda, 0, tolerance = 1e-15, info = method)
    expect_equal(sum(selected$refit), 0, info = method)
    expect_equal(selected$opt.sparsity, 0, info = method)
  }
})

test_that("RIC refits preserve explicit square-data routing", {
  x <- matrix(c(
    3, 2, 2,
    2, 4, 4,
    2, 4, 7
  ), 3, 3)
  fit <- huge(
    x, method = "tiger", lambda = 1, input.type = "data",
    verbose = FALSE
  )
  selected <- expect_no_warning(huge.select(
    fit, criterion = "ric", rep.num = nrow(x), verbose = FALSE
  ))
  data.refit <- huge.tiger(
    x, lambda = selected$opt.lambda, input.type = "data",
    verbose = FALSE
  )
  automatic.refit <- huge.tiger(
    x, lambda = selected$opt.lambda, verbose = FALSE
  )

  expect_false(selected$cov.input)
  expect_identical(
    as.matrix(selected$refit), as.matrix(data.refit$path[[1]])
  )
  expect_false(identical(
    as.matrix(selected$refit), as.matrix(automatic.refit$path[[1]])
  ))
})

test_that("RIC handles the single-variable empty-graph boundary", {
  x <- matrix(c(-1, 0, 1), ncol = 1)

  for (method in c("ct", "mb", "glasso", "tiger")) {
    fit <- huge(x, method = method, nlambda = 3, verbose = FALSE)
    selected <- huge.select(
      fit, criterion = "ric", rep.num = nrow(x), verbose = FALSE
    )

    expect_s3_class(selected, "select")
    expect_equal(selected$opt.lambda, 0, tolerance = 1e-15, info = method)
    expect_equal(dim(selected$refit), c(1L, 1L), info = method)
    expect_equal(sum(selected$refit), 0, info = method)
    expect_identical(selected$opt.sparsity, 0, info = method)

    if (method %in% c("mb", "tiger")) {
      selected_default <- huge.select(
        fit, rep.num = nrow(x), verbose = FALSE
      )
      expect_identical(selected_default$criterion, "ric", info = method)
      expect_equal(selected_default$opt.lambda, 0, tolerance = 1e-15,
                   info = method)
      expect_identical(selected_default$opt.sparsity, 0, info = method)
      expect_equal(sum(selected_default$refit), 0, info = method)
    }
  }
})

test_that("StARS rejects single-variable graph paths", {
  x <- matrix(c(-1, 0, 1), ncol = 1)

  for (method in c("ct", "mb", "glasso")) {
    fit <- huge(x, method = method, nlambda = 3, verbose = FALSE)
    expect_error(
      huge.select(
        fit, criterion = "stars", rep.num = 2, verbose = FALSE
      ),
      "StARS requires at least two variables",
      info = method
    )
  }

  tiger <- huge(x, method = "tiger", nlambda = 3, verbose = FALSE)
  expect_error(
    huge.select(tiger, criterion = "stars", rep.num = 2, verbose = FALSE),
    "TIGER.*StARS|common certified prefix"
  )
})

test_that("StARS requires a non-increasing lambda path", {
  set.seed(58)
  x <- matrix(rnorm(40 * 4), nrow = 40)

  ascending <- huge(
    x, method = "ct", lambda = c(.1, .3, .5), verbose = FALSE
  )
  expect_error(
    huge.select(ascending, rep.num = 2, verbose = FALSE),
    "non-increasing"
  )

  unordered <- huge(
    x, method = "ct", lambda = c(.5, .1, .3), verbose = FALSE
  )
  expect_error(
    huge.select(
      unordered, criterion = "stars", rep.num = 2, verbose = FALSE
    ),
    "non-increasing"
  )

  tied <- huge(
    x, method = "ct", lambda = c(.5, .5, .2), verbose = FALSE
  )
  expect_no_error(huge.select(
    tied, criterion = "stars", rep.num = 2, verbose = FALSE
  ))

  expect_no_error(huge.select(
    ascending, criterion = "ric", rep.num = 2, verbose = FALSE
  ))

  tiger <- huge(x, method = "tiger", nlambda = 3, verbose = FALSE)
  tiger$lambda <- rev(tiger$lambda)
  expect_error(
    huge.select(
      tiger, criterion = "stars", rep.num = 2, verbose = FALSE
    ),
    "TIGER.*StARS|common certified prefix"
  )
})

test_that("tiger returns valid structure", {
  set.seed(30)
  L <- huge.generator(n = 80, d = 30, graph = "hub", verbose = FALSE)
  fit <- huge(L$data, method = "tiger", verbose = FALSE)

  expect_s3_class(fit, "huge")
  expect_equal(fit$method, "tiger")
  expect_equal(length(fit$path), length(fit$lambda))
  expect_equal(length(fit$icov), length(fit$lambda))
  expect_equal(length(fit$sparsity), length(fit$lambda))
})

test_that("tiger handles single-variable raw and covariance inputs", {
  raw <- matrix(c(-1, 0, 1), ncol = 1)
  covariance <- matrix(1, nrow = 1, ncol = 1)
  cases <- list(
    raw_auto = list(x = raw, nlambda = 3, lambda.min.ratio = 0.2),
    covariance_auto = list(
      x = covariance, nlambda = 3, lambda.min.ratio = 0.2
    ),
    raw_explicit = list(x = raw, lambda = c(0.5, 0.2)),
    covariance_explicit = list(x = covariance, lambda = c(0.5, 0.2))
  )

  fits <- lapply(cases, function(args) {
    do.call(huge.tiger, c(args, list(verbose = FALSE)))
  })
  for (case in names(fits)) {
    fit <- fits[[case]]
    path_length <- if (grepl("auto$", case)) 3L else 2L
    expect_identical(
      fit$cov.input, grepl("^covariance", case), info = case
    )
    expect_length(fit$lambda, path_length)
    expect_true(all(is.finite(fit$lambda) & fit$lambda > 0), info = case)
    expect_identical(fit$sparsity, rep(0, path_length), info = case)
    expect_equal(dim(fit$df), c(1L, path_length), info = case)
    expect_identical(as.numeric(fit$df), rep(0, path_length), info = case)
    expect_length(fit$beta, path_length)
    expect_length(fit$path, path_length)
    expect_length(fit$icov, path_length)
    for (index in seq_len(path_length)) {
      expect_true(inherits(fit$beta[[index]], "sparseMatrix"), info = case)
      expect_true(inherits(fit$path[[index]], "sparseMatrix"), info = case)
      expect_silent(validObject(fit$beta[[index]]))
      expect_silent(validObject(fit$path[[index]]))
      expect_equal(dim(fit$beta[[index]]), c(1L, 1L), info = case)
      expect_equal(dim(fit$path[[index]]), c(1L, 1L), info = case)
      expect_equal(sum(fit$beta[[index]]), 0, info = case)
      expect_equal(sum(fit$path[[index]]), 0, info = case)
      expect_equal(dim(fit$icov[[index]]), c(1L, 1L), info = case)
      expect_true(all(is.finite(fit$icov[[index]])), info = case)
      expect_equal(
        fit$icov[[index]], matrix(1, 1, 1), tolerance = 1e-15,
        info = case
      )
    }
  }

  expected_auto <- exp(seq(log(1e-3), log(2e-4), length.out = 3))
  expect_identical(fits$raw_auto$lambda, expected_auto)
  expect_identical(fits$raw_auto$lambda, fits$covariance_auto$lambda)
  expect_identical(fits$raw_explicit$lambda, c(0.5, 0.2))
  expect_identical(fits$raw_explicit$lambda, fits$covariance_explicit$lambda)
})

test_that("tiger sparsity is non-decreasing", {
  set.seed(31)
  L <- huge.generator(n = 80, d = 30, graph = "hub", verbose = FALSE)
  fit <- huge(L$data, method = "tiger", verbose = FALSE)

  expect_true(all(diff(fit$lambda) < 0))
  expect_true(all(diff(fit$sparsity) >= -1e-10))
})

test_that("tiger path matrices are symmetric and binary", {
  set.seed(32)
  L <- huge.generator(n = 80, d = 30, graph = "band", verbose = FALSE)
  fit <- huge(L$data, method = "tiger", verbose = FALSE)

  for (k in seq_along(fit$path)) {
    p <- as.matrix(fit$path[[k]])
    expect_equal(p, t(p), info = paste("path asymmetric at k =", k))
    expect_true(all(p %in% c(0, 1)))
    expect_true(all(diag(p) == 0))
  }
})

test_that("tiger works across graph types", {
  set.seed(33)
  for (g in c("hub", "band", "cluster")) {
    L <- huge.generator(n = 60, d = 20, graph = g, verbose = FALSE)
    fit <- huge(L$data, method = "tiger", verbose = FALSE)
    expect_true(all(diff(fit$sparsity) >= -1e-10),
                info = paste("non-monotone sparsity for graph =", g))
  }
})

test_that("tiger raw and covariance inputs use the same correlation problem", {
  set.seed(34)
  x = matrix(rnorm(120 * 12), 120, 12)
  x = sweep(x, 2, seq_len(ncol(x)), "*")
  lambda = c(0.35, 0.2, 0.1)

  raw_fit = huge.tiger(x, lambda = lambda, verbose = FALSE)
  cov_fit = huge.tiger(cov(x), lambda = lambda, verbose = FALSE)

  expect_identical(cov_fit$lambda, lambda)
  for (i in seq_along(lambda)) {
    expect_equal(as.matrix(raw_fit$beta[[i]]),
                 as.matrix(cov_fit$beta[[i]]), tolerance = 1e-8)
    expect_identical(as.matrix(raw_fit$path[[i]]),
                     as.matrix(cov_fit$path[[i]]))
    expect_equal(raw_fit$icov[[i]], cov_fit$icov[[i]], tolerance = 1e-8)
  }
})

test_that("tiger returns the native default lambda path", {
  set.seed(35)
  x = matrix(rnorm(100 * 10), 100, 10)
  ratio = 0.2
  nlambda = 6

  raw_fit = huge.tiger(x, nlambda = nlambda,
                       lambda.min.ratio = ratio, verbose = FALSE)
  cov_fit = huge.tiger(cov(x), nlambda = nlambda,
                       lambda.min.ratio = ratio, verbose = FALSE)
  lambda_max = max(abs(cor(x)[upper.tri(cor(x))]))

  expect_length(raw_fit$lambda, nlambda)
  expect_equal(raw_fit$lambda[[1]], lambda_max, tolerance = 1e-12)
  expect_equal(raw_fit$lambda[[nlambda]], ratio * lambda_max,
               tolerance = 1e-12)
  expect_equal(raw_fit$lambda, cov_fit$lambda, tolerance = 1e-12)
})

test_that("tiger native lambda preserves weak nonzero correlations", {
  rho = 5e-4
  a = c(-1, -1, 1, 1)
  b = c(-1, 1, -1, 1)
  x = cbind(a, rho * a + sqrt(1 - rho^2) * b)
  ratio = 0.2
  expected = exp(seq(log(rho), log(rho * ratio), length.out = 3))

  raw_fit = huge.tiger(
    x, nlambda = 3, lambda.min.ratio = ratio, verbose = FALSE
  )
  covariance_fit = huge.tiger(
    matrix(c(1, rho, rho, 1), 2, 2),
    nlambda = 3, lambda.min.ratio = ratio, verbose = FALSE
  )

  expect_equal(cor(x)[1, 2], rho, tolerance = 1e-15)
  expect_equal(raw_fit$lambda, expected, tolerance = 1e-12)
  expect_equal(covariance_fit$lambda, expected, tolerance = 1e-12)
  expect_equal(raw_fit$lambda, covariance_fit$lambda, tolerance = 1e-12)
})

test_that("tiger native lambda paths stay positive at the subnormal limit", {
  minimum = .Machine$double.xmin * .Machine$double.eps
  covariance = matrix(c(1, minimum, minimum, 1), 2, 2)

  fit = huge.tiger(
    covariance, nlambda = 3, lambda.min.ratio = .1, verbose = FALSE
  )
  native = .Call(
    "_huge_SPMBgraphsqrtFit", covariance, NULL, 3L, 2L, TRUE, .1,
    PACKAGE = "huge"
  )
  expect_identical(fit$lambda, rep(minimum, 3))
  expect_identical(native$lambda, fit$lambda)
  expect_false(native$path_truncated)
  expect_true(all(vapply(fit$icov, function(x) all(is.finite(x)), logical(1))))

  identity = huge.tiger(
    diag(2), nlambda = 3, lambda.min.ratio = minimum,
    verbose = FALSE
  )
  native.identity = .Call(
    "_huge_SPMBgraphsqrtFit", diag(2), NULL, 3L, 2L, TRUE, minimum,
    PACKAGE = "huge"
  )
  expect_equal(identity$lambda[[1]], 1e-3, tolerance = 1e-15)
  expect_identical(identity$lambda[[3]], minimum)
  expect_identical(native.identity$lambda, identity$lambda)
  expect_false(native.identity$path_truncated)
  expect_true(all(is.finite(identity$lambda) & identity$lambda > 0))
  expect_true(all(diff(identity$lambda) <= 0))
  expect_equal(
    log(identity$lambda[[2]]),
    log(1e-3) + log(minimum) / 2,
    tolerance = 1e-12
  )

  singleton = huge.tiger(
    matrix(1, 1, 1), nlambda = 3, lambda.min.ratio = minimum,
    verbose = FALSE
  )
  native.singleton = .Call(
    "_huge_SPMBgraphsqrtFit", matrix(1, 1, 1), NULL, 3L, 1L, TRUE,
    minimum, PACKAGE = "huge"
  )
  expect_identical(singleton$lambda, identity$lambda)
  expect_identical(native.singleton$lambda, singleton$lambda)
  expect_false(native.singleton$path_truncated)
})

test_that("tiger explicit lambda paths allow ties and reject increases", {
  covariance = matrix(c(1, .3, .3, 1), 2, 2)
  automatic = huge.tiger(
    covariance, nlambda = 3, lambda.min.ratio = 1, verbose = FALSE
  )
  replay = huge.tiger(
    covariance, lambda = automatic$lambda, verbose = FALSE
  )
  generic.replay = huge(
    covariance, method = "tiger", lambda = automatic$lambda,
    verbose = FALSE
  )
  native.replay = .Call(
    "_huge_SPMBgraphsqrtFit", covariance, automatic$lambda, 3L, 2L,
    TRUE, .1, PACKAGE = "huge"
  )

  expect_true(all(diff(automatic$lambda) == 0))
  expect_identical(replay$lambda, automatic$lambda)
  expect_identical(generic.replay$lambda, automatic$lambda)
  expect_identical(native.replay$lambda, automatic$lambda)
  for(index in seq_along(automatic$lambda)) {
    expect_identical(
      as.matrix(replay$beta[[index]]),
      as.matrix(automatic$beta[[index]])
    )
    expect_identical(replay$icov[[index]], automatic$icov[[index]])
  }

  expect_error(
    huge.tiger(covariance, lambda = c(.1, .2), verbose = FALSE),
    "non-increasing"
  )
  expect_error(
    huge(
      covariance, method = "tiger", lambda = c(.1, .2),
      verbose = FALSE
    ),
    "non-increasing"
  )
  expect_error(
    .Call(
      "_huge_SPMBgraphsqrtFit", covariance, c(.1, .2), 2L, 2L,
      TRUE, .1, PACKAGE = "huge"
    ),
    "non-increasing"
  )
  expect_error(
    .Call(
      "_huge_SPMBgraphsqrtFit", covariance, c(.1, .2, NaN), 3L, 2L,
      TRUE, .1, PACKAGE = "huge"
    ),
    "positive and finite"
  )

  raw = cbind(c(-1, -1, 1, 1), c(-1, 1, -1, 1))
  expect_no_error(
    .Call(
      "_huge_SPMBgraphsqrt", raw, c(.5, .5), 2L, 2L,
      PACKAGE = "huge"
    )
  )
  expect_error(
    .Call(
      "_huge_SPMBgraphsqrt", raw, c(.1, .2), 2L, 2L,
      PACKAGE = "huge"
    ),
    "non-increasing"
  )
  expect_error(
    .Call(
      "_huge_SPMBgraphsqrt", raw, c(.1, .2, NaN), 3L, 2L,
      PACKAGE = "huge"
    ),
    "positive and finite"
  )
  expect_error(
    .Call(
      "_huge_SPMBgraphsqrt", raw, .5, 2L, 2L,
      PACKAGE = "huge"
    ),
    "length must match"
  )
  expect_error(
    .Call(
      "_huge_SPMBgraphsqrt", raw, c(.5, .5), 2L, 3L,
      PACKAGE = "huge"
    ),
    "dimensions"
  )
})

test_that("tiger raw correlation is stable across finite column scales", {
  x = cbind(c(-1, 0, 1), c(-1, 1, 0))
  ratio = 0.5
  reference = huge.tiger(
    x, nlambda = 3, lambda.min.ratio = ratio, verbose = FALSE
  )
  native.reference = .Call(
    "_huge_SPMBgraphsqrtFit", x, NULL, 3L, 2L, FALSE, ratio,
    PACKAGE = "huge"
  )
  explicit.reference = .Call(
    "_huge_SPMBgraphsqrtFit", x, .25, 1L, 2L, FALSE, ratio,
    PACKAGE = "huge"
  )
  expect_equal(reference$lambda[[1]], 0.5, tolerance = 1e-15)

  minimum = .Machine$double.xmin * .Machine$double.eps
  inputs = list(
    x * 1e308,
    x * 1e-200,
    x * minimum,
    sweep(x, 2, c(1e308, 1e-200), "*")
  )
  for (current in inputs) {
    expect_no_warning({
      fit = huge.tiger(
        current, nlambda = 3, lambda.min.ratio = ratio, verbose = FALSE
      )
      native = .Call(
        "_huge_SPMBgraphsqrtFit", current, NULL, 3L, 2L, FALSE, ratio,
        PACKAGE = "huge"
      )
      explicit = .Call(
        "_huge_SPMBgraphsqrtFit", current, .25, 1L, 2L, FALSE, ratio,
        PACKAGE = "huge"
      )
    })
    expect_equal(fit$lambda, reference$lambda, tolerance = 1e-14)
    for (field in c("col_cnz", "row_idx")) {
      expect_identical(native[[field]], native.reference[[field]])
      expect_identical(explicit[[field]], explicit.reference[[field]])
    }
    for (field in c("x", "icov", "lambda")) {
      expect_equal(native[[field]], native.reference[[field]], tolerance = 1e-12)
      expect_equal(
        explicit[[field]], explicit.reference[[field]], tolerance = 1e-12
      )
    }
    for (field in c("hit_max_iter", "path_truncated")) {
      expect_identical(native[[field]], native.reference[[field]])
      expect_identical(explicit[[field]], explicit.reference[[field]])
    }
    for (i in seq_along(reference$lambda)) {
      expect_equal(
        as.matrix(fit$beta[[i]]), as.matrix(reference$beta[[i]]),
        tolerance = 1e-12
      )
      expect_identical(
        as.matrix(fit$path[[i]]), as.matrix(reference$path[[i]])
      )
      expect_equal(fit$icov[[i]], reference$icov[[i]], tolerance = 1e-12)
    }
  }
})

test_that("tiger correlation solver satisfies square-root lasso KKT conditions", {
  set.seed(9)
  n = 80
  d = 15
  lambda = 0.2
  x = matrix(rnorm(n * d), n, d)
  fit = huge.tiger(x, lambda = lambda, verbose = FALSE)
  beta = as.matrix(fit$beta[[1]])
  correlation = cor(x)

  kkt_error = vapply(seq_len(d), function(response) {
    coefficient = beta[, response]
    residual_variance = correlation[response, response] -
      2 * sum(correlation[, response] * coefficient) +
      sum(coefficient * (correlation %*% coefficient))
    tau = sqrt(max(residual_variance, .Machine$double.eps))
    score = as.numeric((correlation[, response] -
                          correlation %*% coefficient) / tau)
    score[response] = 0

    active = abs(coefficient) > 1e-8
    active[response] = FALSE
    active_error = if(any(active))
      max(abs(score[active] - lambda * sign(coefficient[active]))) else 0
    inactive = !active
    inactive[response] = FALSE
    inactive_error = if(any(inactive))
      max(pmax(abs(score[inactive]) - lambda, 0)) else 0
    max(active_error, inactive_error)
  }, numeric(1))

  expect_gt(sum(abs(beta) > 1e-8), 0)
  expect_lte(max(kkt_error), 1e-6)
})

test_that("tiger matches the two-variable analytic solution", {
  rho = 0.3
  lambda = 0.2
  correlation = matrix(c(1, rho, rho, 1), 2, 2)
  expected = rho - lambda * sqrt((1 - rho^2) / (1 - lambda^2))

  fit = huge.tiger(correlation, lambda = lambda, verbose = FALSE)
  coefficient = as.matrix(fit$beta[[1]])

  expect_equal(coefficient[1, 2], expected, tolerance = 1e-7)
  expect_equal(coefficient[2, 1], expected, tolerance = 1e-7)
})

test_that("tiger handles identity, singular, and rescaled covariance inputs", {
  identity_fit = huge.tiger(diag(4), nlambda = 3, verbose = FALSE)
  expect_equal(identity_fit$lambda, c(1e-3, sqrt(1e-7), 1e-4),
               tolerance = 1e-15)
  expect_identical(identity_fit$sparsity, rep(0, 3))
  for (icov in identity_fit$icov)
    expect_true(all(is.finite(icov)))

  set.seed(36)
  x = matrix(rnorm(100 * 6), 100, 6)
  covariance = cov(x)
  scale = diag(seq_len(ncol(x)))
  fit = huge.tiger(covariance, lambda = c(0.4, 0.2), verbose = FALSE)
  rescaled = huge.tiger(scale %*% covariance %*% scale,
                        lambda = c(0.4, 0.2), verbose = FALSE)

  for (i in 1:2) {
    expect_equal(as.matrix(fit$beta[[i]]), as.matrix(rescaled$beta[[i]]),
                 tolerance = 1e-8)
    expect_equal(fit$icov[[i]], rescaled$icov[[i]], tolerance = 1e-8)
  }

  correlation = matrix(c(1, .6, .6, 1), 2, 2)
  large = .Machine$double.xmax * correlation
  large_fit = huge.tiger(large, lambda = .25, verbose = FALSE)
  correlation_fit = huge.tiger(
    correlation, lambda = .25, verbose = FALSE
  )
  expect_equal(
    as.matrix(large_fit$beta[[1]]),
    as.matrix(correlation_fit$beta[[1]]), tolerance = 1e-12
  )
  expect_equal(
    large_fit$icov[[1]], correlation_fit$icov[[1]], tolerance = 1e-12
  )

  large_auto_fit = huge.tiger(
    large, nlambda = 3, lambda.min.ratio = .5, verbose = FALSE
  )
  correlation_auto_fit = huge.tiger(
    correlation, nlambda = 3, lambda.min.ratio = .5, verbose = FALSE
  )
  expect_equal(
    large_auto_fit$lambda, correlation_auto_fit$lambda, tolerance = 1e-14
  )
  expect_equal(large_auto_fit$lambda[[1]], .6, tolerance = 1e-14)

  nearly.symmetric = large
  nearly.symmetric[2, 1] = nearly.symmetric[1, 2] *
    (1 - .Machine$double.eps)
  expect_false(identical(nearly.symmetric[1, 2], nearly.symmetric[2, 1]))
  projected = nearly.symmetric / 2 + t(nearly.symmetric) / 2
  near_fit = huge.tiger(
    nearly.symmetric, nlambda = 3, lambda.min.ratio = .5, verbose = FALSE
  )
  projected_fit = huge.tiger(
    projected, nlambda = 3, lambda.min.ratio = .5, verbose = FALSE
  )
  expect_equal(near_fit$lambda, projected_fit$lambda, tolerance = 1e-14)
  for (i in seq_along(near_fit$lambda)) {
    expect_equal(
      as.matrix(near_fit$beta[[i]]),
      as.matrix(projected_fit$beta[[i]]), tolerance = 1e-12
    )
    expect_equal(
      near_fit$icov[[i]], projected_fit$icov[[i]], tolerance = 1e-12
    )
  }

  x[, 6] = x[, 1]
  expect_warning(
    singular_fit <- huge.tiger(cov(x), nlambda = 2,
                               lambda.min.ratio = 0.9, verbose = FALSE),
    "certified prefix"
  )
  expect_length(singular_fit$lambda, 1)
  expect_true(all(is.finite(as.matrix(singular_fit$beta[[1]]))))
  expect_true(all(is.finite(singular_fit$icov[[1]])))
})

test_that("tiger preserves weak correlations across extreme variances", {
  rho = 1e-200
  high = .Machine$double.xmax
  low = 1e-300
  off.diagonal = rho * sqrt(high) * sqrt(low)
  covariance = matrix(
    c(high, off.diagonal, off.diagonal, low), nrow = 2
  )
  correlation = matrix(c(1, rho, rho, 1), nrow = 2)

  reference = huge.tiger(
    correlation, nlambda = 3, lambda.min.ratio = .5, verbose = FALSE
  )

  for (current in list(covariance, covariance[2:1, 2:1])) {
    fit = huge.tiger(
      current, nlambda = 3, lambda.min.ratio = .5, verbose = FALSE
    )
    native = .Call(
      "_huge_SPMBgraphsqrtFit", current, NULL, 3L, 2L, TRUE, .5,
      PACKAGE = "huge"
    )

    expect_equal(
      fit$lambda / rho, reference$lambda / rho, tolerance = 1e-12
    )
    expect_equal(fit$lambda[[1]] / rho, 1, tolerance = 1e-12)
    expect_equal(
      native$lambda / rho, reference$lambda / rho, tolerance = 1e-12
    )
  }
})

test_that("tiger stops generated paths before an interpolation degeneracy", {
  signal = seq_len(30)
  duplicated_data = cbind(signal, signal)

  expect_warning(
    fit <- huge.tiger(duplicated_data, nlambda = 2,
                      lambda.min.ratio = 0.99, verbose = FALSE),
    "certified prefix"
  )
  expect_equal(fit$lambda, 1, tolerance = 1e-12)
  returned_nlambda = length(fit$lambda)
  expect_length(fit$beta, returned_nlambda)
  expect_length(fit$path, returned_nlambda)
  expect_length(fit$icov, returned_nlambda)
  expect_length(fit$sparsity, returned_nlambda)
  expect_equal(ncol(fit$df), returned_nlambda)
  expect_true(all(is.finite(fit$icov[[1]])))
  expect_equal(fit$icov[[1]], diag(2), tolerance = 1e-12)

  expect_warning(
    covariance_fit <- huge.tiger(cov(duplicated_data), nlambda = 2,
                                 lambda.min.ratio = 0.99, verbose = FALSE),
    "certified prefix"
  )
  expect_identical(covariance_fit$lambda, fit$lambda)
  expect_equal(as.matrix(covariance_fit$beta[[1]]),
               as.matrix(fit$beta[[1]]), tolerance = 1e-12)
  expect_equal(covariance_fit$icov[[1]], fit$icov[[1]], tolerance = 1e-12)

  expect_error(
    huge.tiger(duplicated_data, lambda = 0.99, verbose = FALSE),
    "could not certify a supplied lambda"
  )
})

test_that("tiger rejects invalid native preprocessing inputs", {
  constant = cbind(matrix(rnorm(60 * 3), 60, 3), 1)
  expect_error(huge.tiger(constant, verbose = FALSE), "constant column")

  invalid_covariance = diag(c(1, 0, 1))
  expect_error(huge.tiger(invalid_covariance, verbose = FALSE),
               "positive finite diagonal")

  nonfinite = matrix(rnorm(60 * 4), 60, 4)
  nonfinite[1, 1] = NA_real_
  expect_error(huge.tiger(nonfinite, verbose = FALSE), "finite values")
})

test_that("tiger native covariance input validates symmetry", {
  asymmetric = matrix(c(1, 1, -1, 1), 2, 2, byrow = TRUE)
  expect_error(
    .Call(
      "_huge_SPMBgraphsqrtFit", asymmetric, .25, 1L, 2L, TRUE, .1,
      PACKAGE = "huge"
    ),
    "symmetric"
  )

  near = matrix(c(1, .5, .5 + 1e-15, 1), 2, 2, byrow = TRUE)
  expect_no_error(.Call(
    "_huge_SPMBgraphsqrtFit", near, .25, 1L, 2L, TRUE, .1,
    PACKAGE = "huge"
  ))
})

# Internal helper functions (not exported)

# Base graphics parameters have dependencies: layouts reset text scaling,
# outer regions reset figures, and margins reset plot regions. Reapply linked
# groups in dependency order so callers receive their writable state back.
.huge_graphics_state = function() {
  par(no.readonly = TRUE)
}

.huge_restore_graphics_state = function(state) {
  # `new = TRUE` warns when an error occurred before the first valid plot.
  # Restore it separately at the end so that warning cannot mask the error.
  bulk.state = state
  bulk.state$new = NULL
  par(bulk.state)

  restore.group = function(fields, setters) {
    matches = function() {
      current = par(no.readonly = TRUE)
      all(vapply(fields, function(name)
        identical(current[[name]], state[[name]]), logical(1)))
    }
    if(matches()) return(invisible(NULL))
    for(name in setters) {
      do.call(par, stats::setNames(list(state[[name]]), name))
      if(matches()) break
    }
    invisible(NULL)
  }

  multi.panel = prod(state$mfrow) > 1L
  if(multi.panel) par(mfrow = state$mfrow)
  par(pty = state$pty, cex = state$cex, mex = state$mex)
  restore.group(c("oma", "omi", "omd"), c("omi", "omd", "oma"))
  if(multi.panel) par(mfg = state$mfg)
  if(!multi.panel) restore.group(c("fig", "fin"), c("fig", "fin"))
  restore.group(c("mar", "mai"), c("mai", "mar"))
  restore.group(c("pin", "plt"), c("pin", "plt"))
  par(usr = state$usr, xaxp = state$xaxp, yaxp = state$yaxp)
  suppressWarnings(par(new = state$new))
  invisible(NULL)
}

# Unit-variance centered columns with bounded intermediate arithmetic.
# Reference shifting before max-absolute scaling preserves representable
# differences at extreme finite magnitudes.
.huge_standardize = function(x) {
  n = nrow(x)
  d = ncol(x)
  xs = matrix(0, n, d, dimnames = dimnames(x))

  for(j in seq_len(d)) {
    source = x[,j] * 1.0
    column.scale = max(abs(source))
    if(!is.finite(column.scale) || column.scale == 0)
      stop("Data must have finite, positive sample standard deviations.")

    reference = source[[1]]
    direct = source == 0 | reference == 0 |
      ((source < 0) == (reference < 0))
    delta = source / column.scale - reference / column.scale
    delta[direct] = (source[direct] - reference) / column.scale
    delta = delta - mean(delta)

    centered.scale = max(abs(delta))
    if(!is.finite(centered.scale) || centered.scale == 0)
      stop("Data must have finite, positive sample standard deviations.")
    delta = delta / centered.scale
    column.sd = sd(delta)
    if(!is.finite(column.sd) || column.sd <= 0)
      stop("Data must have finite, positive sample standard deviations.")
    xs[,j] = delta / column.sd
  }
  xs
}

# Correlation matrix via BLAS crossprod on stable standardized columns.
.huge_fast_cor = function(x) {
  xs = tryCatch(.huge_standardize(x), error = function(e) NULL)
  if(is.null(xs))
    return(cor(x))
  n = nrow(xs)
  S = crossprod(xs) / (n - 1)
  S[S > 1] = 1
  S[S < -1] = -1
  diag(S) = 1
  S
}

.huge_is_covariance_input = function(x) {
  if(nrow(x) != ncol(x)) return(FALSE)

  values = x * 1.0
  if(any(!is.finite(values))) return(FALSE)

  tolerance = 100 * .Machine$double.eps
  dimension = nrow(values)
  if(dimension <= 1) return(TRUE)

  # Compare off-diagonal entries in implied-correlation units.  This keeps
  # covariance routing invariant to a finite uniform rescaling without
  # subtracting large values before they have been normalized.
  diagonal = abs(diag(values))
  diagonal.root = sqrt(diagonal)
  for(column in 2:dimension) {
    rows = seq_len(column - 1L)
    left = values[rows, column]
    right = values[column, rows]
    different = left != right
    if(!any(different)) next

    rows = rows[different]
    left = left[different]
    right = right[different]
    covariance.scale = diagonal.root[rows] * diagonal.root[column]
    equal.diagonal = diagonal[rows] == diagonal[column]
    covariance.scale[equal.diagonal] = diagonal[column]
    reference = pmax(abs(left), abs(right), covariance.scale)
    left = left / reference
    right = right / reference
    scale = pmax(abs(left), abs(right))
    threshold = ifelse(
      scale <= tolerance, tolerance, tolerance * scale
    )
    if(any(abs(left - right) > threshold)) return(FALSE)
  }
  TRUE
}

.huge_validate_correlation_psd = function(correlation) {
  d = nrow(correlation)
  spectral.bound = max(1, max(rowSums(abs(correlation))))
  tolerance = 100 * .Machine$double.eps * max(1, d) * spectral.bound
  shifted = correlation
  diag(shifted) = diag(shifted) + tolerance
  factor = tryCatch(chol(shifted), error = function(e) NULL)
  if(is.null(factor))
    stop("Covariance input must be positive semidefinite.")
  invisible(NULL)
}

.huge_validate_input_type = function(input.type) {
  allowed = c("auto", "data", "covariance")
  if(length(input.type) != 1L || !is.character(input.type) ||
     is.na(input.type) || !(input.type %in% allowed))
    stop('input.type must be exactly one of "auto", "data", or "covariance".')
  input.type
}

.huge_validate_estimation_input = function(
  x, input.type = "auto", prepare.covariance = TRUE, require.psd = TRUE
) {
  if(!is.matrix(x) || !is.numeric(x))
    stop("x must be a numeric matrix.")

  n = nrow(x)
  d = ncol(x)
  if(n < 1 || d < 1)
    stop("x must be a non-empty numeric matrix.")
  if(any(!is.finite(x)))
    stop("x must contain only finite values.")

  input.type = .huge_validate_input_type(input.type)
  if(input.type == "covariance") {
    if(n != d)
      stop('input.type = "covariance" requires a square matrix.')
    if(!.huge_is_covariance_input(x))
      stop("Covariance input must be symmetric within numeric tolerance.")
  }
  cov.input = switch(
    input.type,
    auto = .huge_is_covariance_input(x),
    data = FALSE,
    covariance = TRUE
  )

  covariance = NULL
  correlation = NULL
  if(cov.input && prepare.covariance)
  {
    transpose = t(x)
    covariance = x * 1.0
    different = x != transpose
    covariance[different] = x[different] / 2 + transpose[different] / 2
    diagonal = diag(covariance)
    if(any(!is.finite(diagonal)) || any(diagonal <= 0))
      stop("Covariance input must have a positive finite diagonal.")

    inv.sd = 1 / sqrt(diagonal)
    correlation = diag(1, d)
    dimnames(correlation) = dimnames(covariance)
    if(d > 1) {
      for(column in 2:d) {
        for(row in seq_len(column - 1L)) {
          inv.large = max(inv.sd[[row]], inv.sd[[column]])
          inv.small = min(inv.sd[[row]], inv.sd[[column]])
          value = (covariance[[row, column]] * inv.large) * inv.small
          if(!is.finite(value))
            stop("Covariance input cannot produce a finite correlation matrix.")
          if(abs(value) > 1 + 1e-8)
            stop("Covariance input is not a valid covariance matrix.")
          value = max(-1, min(1, value))
          correlation[[row, column]] = value
          correlation[[column, row]] = value
        }
      }
    }
    if(require.psd)
      .huge_validate_correlation_psd(correlation)
  }
  else if(!cov.input)
  {
    if(n < 2)
      stop("Raw data x must contain at least two observations.")
    constant = vapply(seq_len(d), function(j) {
      all(x[,j] == x[1,j])
    }, logical(1))
    if(any(constant))
      stop("Raw data x contains a constant column.")
  }

  list(
    input = x, covariance = covariance, correlation = correlation,
    n = n, d = d, cov.input = cov.input
  )
}

.huge_preprocess = function(x, verbose = TRUE, input.type = "auto") {
  inp = .huge_validate_estimation_input(x, input.type = input.type)
  x = inp$input
  n = inp$n
  d = inp$d
  cov.input = inp$cov.input
  if(cov.input) {
    if(verbose) cat("The input is identified as the covariance matrix.\n")
    S = inp$correlation
  } else {
    S = .huge_fast_cor(x)
    if(any(!is.finite(S)))
      stop("Raw data x cannot produce a finite correlation matrix.")
  }
  list(x = x, S = S, n = n, d = d, cov.input = cov.input)
}

.huge_validate_positive_integer = function(value, name) {
  if(length(value) != 1 || !is.numeric(value) || !is.finite(value) ||
     value != floor(value) || value < 1)
    stop(sprintf("%s must be a finite positive integer.", name))
  as.integer(value)
}

.huge_validate_ratio = function(value, name = "lambda.min.ratio") {
  if(length(value) != 1 || !is.numeric(value) || !is.finite(value) ||
     value <= 0 || value > 1)
    stop(sprintf("%s must be a finite number in (0, 1].", name))
  as.numeric(value)
}

.huge_validate_lambda = function(lambda, allow.zero = FALSE) {
  if(!is.numeric(lambda))
    stop("lambda must be a numeric scalar or one-dimensional sequence.")
  lambda.dim = dim(lambda)
  if(!is.null(lambda.dim) && length(lambda.dim) > 1L)
    stop("lambda must be a numeric scalar or one-dimensional sequence.")
  if(length(lambda) < 1 || any(!is.finite(lambda)))
    stop("lambda must be a non-empty finite numeric sequence.")
  if(allow.zero)
  {
    if(any(lambda < 0))
      stop("lambda must contain non-negative values for method = \"ct\".")
  }
  else if(any(lambda <= 0))
    stop("lambda must contain strictly positive values.")
  as.numeric(lambda)
}

.huge_validate_sym = function(sym) {
  if(length(sym) != 1 || !is.character(sym) || is.na(sym) ||
     !(sym %in% c("or", "and")))
    stop("sym must be exactly one of \"or\" or \"and\".")
  sym
}

.huge_geometric_lambda = function(lambda.max, lambda.min.ratio, nlambda) {
  if(nlambda == 1L)
    return(lambda.max)

  lambda.min = lambda.max * lambda.min.ratio
  if(lambda.min > 0) {
    lambda = exp(seq(
      log(lambda.max), log(lambda.min), length.out = nlambda
    ))
    lambda[[1L]] = lambda.max
    return(cummin(lambda))
  }

  # Preserve representable interior points even when the requested endpoint
  # underflows, then saturate only the unrepresentable tail.
  fraction = (seq_len(nlambda) - 1) / (nlambda - 1)
  lambda = exp(log(lambda.max) + fraction * log(lambda.min.ratio))
  smallest.positive = .Machine$double.xmin * .Machine$double.eps
  lambda[!is.finite(lambda) | lambda <= 0] = smallest.positive
  lambda[[1L]] = lambda.max
  cummin(lambda)
}

.huge_default_lambda = function(
  S, d, nlambda = NULL, lambda.min.ratio = NULL, lambda = NULL,
  legacy.glasso.covariance = FALSE
) {
  if(!is.null(lambda)) {
    lambda = .huge_validate_lambda(lambda)
    if(length(lambda) > 1L &&
       any(lambda[-1L] > lambda[-length(lambda)]))
      stop(paste(
        "lambda must be non-increasing for method = \"mb\" or \"glasso\";",
        "tied values are allowed."
      ))
    return(list(lambda = lambda, nlambda = length(lambda)))
  }
  if(is.null(nlambda)) nlambda = 10
  else nlambda = .huge_validate_positive_integer(nlambda, "nlambda")
  if(is.null(lambda.min.ratio)) lambda.min.ratio = 0.1
  else lambda.min.ratio = .huge_validate_ratio(lambda.min.ratio)
  if(legacy.glasso.covariance)
    lambda.max = max(abs(S - diag(d)))
  else {
    offdiag = S
    diag(offdiag) = 0
    lambda.max = max(abs(offdiag))
  }
  if(lambda.max == 0) lambda.max = 1e-3
  lambda = .huge_geometric_lambda(
    lambda.max, lambda.min.ratio, nlambda
  )
  list(lambda = lambda, nlambda = nlambda)
}

test_that("ct returns valid structure", {
  set.seed(40)
  L <- huge.generator(n = 80, d = 30, graph = "hub", verbose = FALSE)
  fit <- huge(L$data, method = "ct", verbose = FALSE)

  expect_s3_class(fit, "huge")
  expect_equal(fit$method, "ct")
  expect_equal(length(fit$path), length(fit$lambda))
  expect_equal(length(fit$sparsity), length(fit$lambda))
})

test_that("ct sparsity is non-decreasing", {
  set.seed(41)
  L <- huge.generator(n = 80, d = 30, graph = "hub", verbose = FALSE)
  fit <- huge(L$data, method = "ct", verbose = FALSE)
  expect_true(all(diff(fit$sparsity) >= -1e-10))
})

test_that("ct path matrices are symmetric and binary", {
  set.seed(42)
  L <- huge.generator(n = 80, d = 30, graph = "hub", verbose = FALSE)
  fit <- huge(L$data, method = "ct", verbose = FALSE)

  for (k in seq_along(fit$path)) {
    p <- as.matrix(fit$path[[k]])
    expect_equal(p, t(p))
    expect_true(all(p %in% c(0, 1)))
    expect_true(all(diag(p) == 0))
  }
})

.ct_covariance <- function(edge.weights) {
  S <- diag(4)
  S[upper.tri(S)] <- edge.weights
  S[lower.tri(S)] <- t(S)[lower.tri(S)]
  S
}

.expect_ct_refittable <- function(S, fit) {
  refit <- huge.ct(S, lambda = fit$lambda, verbose = FALSE)
  d <- ncol(S)

  expect_equal(length(fit$path), length(fit$lambda))
  expect_equal(length(fit$sparsity), length(fit$lambda))
  for (k in seq_along(fit$path)) {
    path <- as.matrix(fit$path[[k]])
    expected.sparsity <- if (d <= 1) 0 else sum(path) / d / (d - 1)

    expect_equal(path, t(path))
    expect_true(all(diag(path) == 0))
    expect_true(all(path %in% c(0, 1)))
    expect_equal(fit$sparsity[k], expected.sparsity)
    expect_equal(path, as.matrix(refit$path[[k]]))
    expect_equal(fit$sparsity[k], refit$sparsity[k])
    if (k > 1)
      expect_true(all(as.matrix(fit$path[[k - 1]]) <= path))
  }
}

test_that("ct default path handles undirected edges and ties", {
  cases <- list(
    unique = list(
      S = .ct_covariance(c(.30, .25, .20, .15, .10, .05)),
      nlambda = 3, ratio = .5,
      lambda = c(.25, .20, .15), edges = c(1, 2, 3)
    ),
    boundary_ties = list(
      S = .ct_covariance(c(.30, .20, .20, .10, .08, .05)),
      nlambda = 3, ratio = .5,
      lambda = c(.20, .20, .10), edges = c(1, 1, 3)
    ),
    all_tied = list(
      S = .ct_covariance(rep(.10, 6)),
      nlambda = 2, ratio = 1,
      lambda = c(.10, 0), edges = c(0, 6)
    ),
    identity = list(
      S = diag(4),
      nlambda = 3, ratio = .5,
      lambda = rep(0, 3), edges = c(0, 0, 0)
    )
  )

  for (case in cases) {
    fit <- huge.ct(
      case$S, nlambda = case$nlambda,
      lambda.min.ratio = case$ratio, verbose = FALSE
    )
    edge.count <- vapply(
      fit$path,
      function(path) sum(as.matrix(path)[upper.tri(path)]),
      numeric(1)
    )

    expect_equal(fit$lambda, case$lambda)
    expect_true(all(fit$lambda >= 0))
    expect_equal(edge.count, case$edges)
    .expect_ct_refittable(case$S, fit)
  }
})

test_that("ct handles one- and two-variable default paths", {
  singleton <- huge.ct(
    matrix(1, nrow = 1, ncol = 1),
    nlambda = 3, lambda.min.ratio = .5, verbose = FALSE
  )
  expect_equal(singleton$lambda, rep(0, 3))
  expect_equal(singleton$sparsity, rep(0, 3))
  .expect_ct_refittable(matrix(1, nrow = 1, ncol = 1), singleton)

  pair.S <- matrix(c(1, .3, .3, 1), nrow = 2)
  pair <- huge.ct(
    pair.S, nlambda = 3, lambda.min.ratio = .5, verbose = FALSE
  )
  expect_equal(pair$lambda, rep(0, 3))
  expect_equal(pair$sparsity, rep(1, 3))
  .expect_ct_refittable(pair.S, pair)
})

test_that("ct default full-graph endpoint retains subnormal correlations", {
  smallest.positive = .Machine$double.xmin * .Machine$double.eps
  S = matrix(c(1, smallest.positive, smallest.positive, 1), 2, 2)

  fit = huge.ct(
    S, nlambda = 1, lambda.min.ratio = 1, verbose = FALSE
  )

  expect_identical(fit$lambda, 0)
  expect_identical(as.matrix(fit$path[[1]]), matrix(c(0, 1, 1, 0), 2, 2))
  .expect_ct_refittable(S, fit)
})

test_that("ct default intermediate thresholds retain subnormal ordering", {
  smallest.positive = .Machine$double.xmin * .Machine$double.eps
  S = diag(3)
  S[upper.tri(S)] = c(3, 2, 1) * smallest.positive
  S[lower.tri(S)] = t(S)[lower.tri(S)]

  fit = huge.ct(
    S, nlambda = 2, lambda.min.ratio = 2 / 3, verbose = FALSE
  )
  edge.count = vapply(
    fit$path,
    function(path) sum(as.matrix(path)[upper.tri(path)]),
    numeric(1)
  )

  expect_identical(
    fit$lambda, c(2, 1) * smallest.positive
  )
  expect_identical(edge.count, c(1, 2))
  .expect_ct_refittable(S, fit)
})

test_that("covariance normalization preserves symmetry across extreme scales", {
  rho = 1e-200
  variances = c(.Machine$double.xmax, 1e-300)
  covariance.value = (rho * sqrt(variances[[1]])) * sqrt(variances[[2]])
  covariance = matrix(
    c(variances[[1]], covariance.value,
      covariance.value, variances[[2]]),
    2, 2
  )

  for(S in list(covariance, covariance[2:1, 2:1])) {
    validated = huge:::.huge_validate_estimation_input(S)
    expect_identical(validated$correlation, t(validated$correlation))
    expect_equal(validated$correlation[[1, 2]], rho, tolerance = 1e-14)

    fit = huge.ct(S, lambda = rho / 2, verbose = FALSE)
    expect_identical(
      as.matrix(fit$path[[1]]), matrix(c(0, 1, 1, 0), 2, 2)
    )
  }
})

test_that("ct uses a literal strict threshold", {
  S <- diag(3)
  S[1, 2] <- S[2, 1] <- .25 + 1e-15
  S[1, 3] <- S[3, 1] <- .25

  fit <- huge.ct(S, lambda = .25, verbose = FALSE)
  path <- as.matrix(fit$path[[1]])

  expect_equal(sum(path[upper.tri(path)]), 1)
  expect_equal(path[1, 2], 1)
  expect_equal(path[1, 3], 0)
})

test_that("generated paths reject invalid nlambda", {
  S <- diag(4)
  invalid <- list(0, -1, 1.5, NA_real_, Inf, c(2, 3), TRUE, "2")

  for (method in c("mb", "glasso", "ct", "tiger")) {
    for (value in invalid) {
      expect_error(
        huge(S, method = method, nlambda = value, verbose = FALSE),
        "nlambda",
        info = paste(method, deparse(value))
      )
    }
  }
})

test_that("generated paths reject invalid lambda.min.ratio", {
  S <- diag(4)
  invalid <- list(0, -0.1, 1.1, NA_real_, Inf, c(.1, .2), TRUE, ".1")

  for (method in c("mb", "glasso", "ct", "tiger")) {
    for (value in invalid) {
      expect_error(
        huge(
          S, method = method, nlambda = 2,
          lambda.min.ratio = value, verbose = FALSE
        ),
        "lambda.min.ratio",
        info = paste(method, deparse(value))
      )
    }
    expect_no_error(huge(
      S, method = method, nlambda = 1,
      lambda.min.ratio = 1, verbose = FALSE
    ))
  }
})

test_that("MB and glasso default lambda tails retain representable values", {
  S = matrix(c(1, 1e-200, 1e-200, 1), 2, 2)
  ratio = 1e-200
  smallest.positive = .Machine$double.xmin * .Machine$double.eps

  generated = huge:::.huge_default_lambda(
    S, d = 2, nlambda = 3, lambda.min.ratio = ratio
  )$lambda

  expect_length(generated, 3)
  expect_true(all(is.finite(generated)))
  expect_true(all(generated > 0))
  expect_true(all(diff(generated) <= 0))
  expect_equal(generated[[1]], 1e-200)
  expect_equal(log(generated[[2]]), log(1e-300), tolerance = 1e-12)
  expect_identical(generated[[3]], smallest.positive)
})

test_that("explicit lambda validates only its active branch", {
  S <- diag(4)
  valid <- list(
    mb = c(.2, .1),
    glasso = c(.2, .1),
    ct = c(.2, 0),
    tiger = c(.2, .1)
  )

  for (method in names(valid)) {
    fit <- huge(
      S, method = method, lambda = valid[[method]],
      nlambda = 0, lambda.min.ratio = 2, verbose = FALSE
    )
    expect_equal(fit$lambda, valid[[method]], info = method)
  }
})

test_that("explicit lambda rejects unsafe values", {
  S <- diag(4)
  positive.methods <- c("mb", "glasso", "tiger")
  invalid.positive <- list(
    numeric(), 0, -1, NA_real_, Inf, c(.2, NA_real_), "0.2"
  )

  for (method in positive.methods) {
    for (value in invalid.positive) {
      expect_error(
        huge(S, method = method, lambda = value, verbose = FALSE),
        "lambda",
        info = paste(method, deparse(value))
      )
    }
  }

  for (value in list(numeric(), -1, NA_real_, Inf, c(.2, NA_real_), "0.2")) {
    expect_error(
      huge(S, method = "ct", lambda = value, verbose = FALSE),
      "lambda",
      info = deparse(value)
    )
  }
  expect_no_error(huge(
    S, method = "ct", lambda = c(.2, 0), verbose = FALSE
  ))
})

test_that("explicit lambda is scalar or one-dimensional", {
  S <- matrix(c(1, .2, .2, 1), nrow = 2)
  path <- c(.9, .8, .7, .6)
  valid <- list(
    scalar = .9,
    vector = path,
    one.dim.array = array(path, dim = length(path))
  )
  invalid <- list(
    row.matrix = matrix(path, nrow = 1),
    column.matrix = matrix(path, ncol = 1),
    square.matrix = matrix(path, nrow = 2),
    byrow.matrix = matrix(path, nrow = 2, byrow = TRUE),
    three.dim.array = array(path, dim = c(2, 1, 2)),
    rank.before.domain = matrix(c(.9, .8, NaN, .6), nrow = 2)
  )

  for(method in c("ct", "mb", "glasso", "tiger")) {
    direct <- get(paste0("huge.", method))
    for(name in names(valid)) {
      value <- valid[[name]]
      expect_identical(
        direct(S, lambda = value, verbose = FALSE)$lambda,
        as.numeric(value),
        info = paste(method, name, "direct")
      )
      expect_identical(
        huge(
          S, method = method, lambda = value, verbose = FALSE
        )$lambda,
        as.numeric(value),
        info = paste(method, name, "generic")
      )
    }
    for(name in names(invalid)) {
      value <- invalid[[name]]
      expect_error(
        direct(S, lambda = value, verbose = FALSE),
        "one-dimensional",
        info = paste(method, name, "direct")
      )
      expect_error(
        huge(
          S, method = method, lambda = value, verbose = FALSE
        ),
        "one-dimensional",
        info = paste(method, name, "generic")
      )
    }
  }
})

test_that("MB and glasso lambda paths allow ties and reject increases", {
  S <- matrix(c(1, .3, .3, 1), nrow = 2)
  tied <- c(.5, .5, .2)

  for(method in c("mb", "glasso")) {
    direct <- get(paste0("huge.", method))
    expect_identical(
      direct(S, lambda = tied, verbose = FALSE)$lambda,
      tied,
      info = paste(method, "direct ties")
    )
    expect_identical(
      huge(S, method = method, lambda = tied, verbose = FALSE)$lambda,
      tied,
      info = paste(method, "generic ties")
    )

    automatic <- direct(
      S, nlambda = 3, lambda.min.ratio = 1, verbose = FALSE
    )
    replay <- direct(
      S, lambda = automatic$lambda, verbose = FALSE
    )
    expect_identical(diff(automatic$lambda), c(0, 0))
    expect_identical(replay$lambda, automatic$lambda)
    expect_identical(replay$path, automatic$path)
    expect_identical(replay$sparsity, automatic$sparsity)

    expect_error(
      direct(S, lambda = c(.1, .2), verbose = FALSE),
      "non-increasing",
      info = paste(method, "direct increase")
    )
    expect_error(
      huge(
        S, method = method, lambda = c(.5, .1, .2), verbose = FALSE
      ),
      "non-increasing",
      info = paste(method, "generic unordered")
    )
    expect_error(
      huge(
        S, method = method, lambda = c(.1, .2, NA_real_),
        verbose = FALSE
      ),
      "finite",
      info = paste(method, "finite error priority")
    )
    expect_error(
      huge(
        S, method = method, lambda = c(.1, .2, -1),
        verbose = FALSE
      ),
      "strictly positive",
      info = paste(method, "sign error priority")
    )
  }
})

test_that("native MB and glasso entries fail closed on unsafe inputs", {
  S <- matrix(c(1, .3, .3, 1), nrow = 2)
  idx <- matrix(c(1L, 0L), nrow = 1)
  native.mb <- function(lambda, nlambda = length(lambda), d = 2L,
                        matrix = S) {
    .Call(
      "_huge_SPMBgraph", matrix, lambda, as.integer(nlambda),
      as.integer(d), PACKAGE = "huge"
    )
  }
  native.mb.scr <- function(lambda, nlambda = length(lambda), d = 2L,
                            matrix = S, index = idx, nscr = 1L) {
    .Call(
      "_huge_SPMBscr", matrix, lambda, as.integer(nlambda),
      as.integer(d), index, as.integer(nscr), PACKAGE = "huge"
    )
  }
  native.glasso <- function(lambda, matrix = S) {
    .Call(
      "_huge_hugeglasso", matrix, lambda, FALSE, FALSE, FALSE,
      PACKAGE = "huge"
    )
  }
  entries <- list(
    mb = native.mb,
    mb.scr = native.mb.scr,
    glasso = native.glasso
  )

  for(name in names(entries)) {
    entry <- entries[[name]]
    expect_no_error(entry(c(.2, .2, .1)))
    expect_error(
      entry(c(.1, .2)), "non-increasing",
      info = paste(name, "increase")
    )
    for(value in list(0, -.1, NA_real_, Inf)) {
      expect_error(
        entry(value), "positive and finite",
        info = paste(name, deparse(value))
      )
    }
    expect_error(
      entry(c(.1, .2, NaN)), "positive and finite",
      info = paste(name, "finite error priority")
    )
    expect_error(
      entry(c(.1, .2, -.1)), "positive and finite",
      info = paste(name, "sign error priority")
    )
    expect_error(
      entry(numeric()), "nlambda|at least one",
      info = paste(name, "empty path")
    )
  }

  expect_error(native.mb(c(.2, .1), nlambda = 1L), "length")
  expect_error(native.mb.scr(c(.2, .1), nlambda = 1L), "length")
  extra.column <- cbind(S, c(0, 0))
  expect_error(native.mb(.1, matrix = extra.column), "d by d")
  expect_error(native.mb.scr(.1, matrix = extra.column), "d by d")
  expect_error(native.mb(.1, d = 0L), "d must be positive")
  expect_error(native.mb.scr(.1, d = 0L), "d must be positive")
  expect_error(
    native.mb.scr(.1, index = matrix(integer(), 0, 2), nscr = 0L),
    "between 1 and d - 1"
  )
  expect_error(
    native.mb.scr(.1, index = matrix(integer(), 0, 3), nscr = 0L),
    "between 1 and d - 1"
  )
  expect_error(
    native.mb.scr(.1, index = cbind(idx, 0L)),
    "nscr by d"
  )
  expect_error(native.glasso(.1, matrix = extra.column), "square")
})

test_that("registered native lambda entries reject multidimensional values", {
  S <- matrix(c(1, .2, .2, 1), nrow = 2)
  raw <- cbind(c(-1, -1, 1, 1), c(-1, 1, -1, 1))
  idx <- matrix(c(1L, 0L), nrow = 1)
  path <- c(.9, .8, .7, .6)
  entries <- list(
    mb = function(lambda) .Call(
      "_huge_SPMBgraph", S, lambda, as.integer(length(lambda)), 2L,
      PACKAGE = "huge"
    ),
    mb.scr = function(lambda) .Call(
      "_huge_SPMBscr", S, lambda, as.integer(length(lambda)), 2L,
      idx, 1L, PACKAGE = "huge"
    ),
    glasso = function(lambda) .Call(
      "_huge_hugeglasso", S, lambda, FALSE, FALSE, FALSE,
      PACKAGE = "huge"
    ),
    tiger.legacy = function(lambda) .Call(
      "_huge_SPMBgraphsqrt", raw, lambda,
      as.integer(length(lambda)), 2L, PACKAGE = "huge"
    ),
    tiger.fit = function(lambda) .Call(
      "_huge_SPMBgraphsqrtFit", S, lambda,
      as.integer(length(lambda)), 2L, TRUE, .1, PACKAGE = "huge"
    )
  )
  valid <- list(.9, path, array(path, dim = length(path)))
  invalid <- list(
    matrix(path, nrow = 1),
    matrix(path, ncol = 1),
    matrix(path, nrow = 2),
    array(path, dim = c(2, 1, 2)),
    matrix(c(.9, .8, NaN, .6), nrow = 2)
  )

  for(name in names(entries)) {
    entry <- entries[[name]]
    for(value in valid)
      expect_no_error(entry(value))
    for(value in invalid) {
      expect_error(
        entry(value), "one-dimensional", info = name
      )
    }
  }

  expect_error(
    .Call(
      "_huge_SPMBgraphsqrt", raw, matrix(.9, nrow = 1),
      0L, 2L, PACKAGE = "huge"
    ),
    "one-dimensional"
  )
})

test_that("registered native RIC entry validates dimensions and rotations", {
  X <- scale(cbind(
    c(-1, -1, 1, 1),
    c(-1, 1, -1, 1)
  ))
  native.ric <- function(
      matrix = X, d = ncol(matrix), n = nrow(matrix),
      r = seq_len(n), t = length(r)) {
    .Call(
      "_huge_RIC", matrix, as.integer(d), as.integer(n), r,
      as.integer(t), PACKAGE = "huge"
    )
  }

  expect_true(is.finite(native.ric()))
  expect_true(is.finite(native.ric(r = c(0, 1, 2, 3, 4), t = 5)))
  expect_true(is.finite(native.ric(r = c(1, 2, NaN), t = 2)))

  # Both mismatches retain more physical storage than the declared problem
  # would read if validation regressed; never probe an undersized buffer.
  expect_error(
    native.ric(
      matrix = rbind(X, c(0, 0)), d = 2L, n = 4L,
      r = 1:4, t = 4L
    ),
    "dimensions.*n by d"
  )
  expect_error(
    native.ric(
      matrix = cbind(X, 0), d = 2L, n = 4L,
      r = 1:4, t = 4L
    ),
    "dimensions.*n by d"
  )
  expect_error(native.ric(d = 0L), "n and d must be positive")
  expect_error(native.ric(n = 0L), "n and d must be positive")

  for(value in c(NA_real_, NaN, Inf, -Inf)) {
    bad <- X
    bad[1, 1] <- value
    expect_error(
      native.ric(matrix = bad), "X must contain only finite",
      info = deparse(value)
    )
  }

  for(value in list(1.5, NaN, Inf, -Inf)) {
    expect_error(
      native.ric(r = value, t = 1L), "finite integers",
      info = deparse(value)
    )
  }
  for(value in c(-1, nrow(X) + 1)) {
    expect_error(
      native.ric(r = value, t = 1L), "rotation indices",
      info = deparse(value)
    )
  }
  expect_error(
    native.ric(r = .Machine$integer.max + 1, t = 1L),
    "integer range"
  )

  expect_error(native.ric(t = 0L), "t must be positive")
  expect_error(native.ric(t = -1L), "t must be positive")
  expect_error(native.ric(r = 1:2, t = 3L), "length of r")

  singleton <- matrix(seq_len(4), ncol = 1)
  expect_identical(
    native.ric(singleton, d = 1L, n = 4L, r = 1, t = 1L),
    0
  )
  singleton[1, 1] <- NA_real_
  expect_error(
    native.ric(singleton, d = 1L, n = 4L, r = 1, t = 1L),
    "X must contain only finite"
  )
})

test_that("mb and tiger require exact sym values", {
  set.seed(571)
  x <- matrix(rnorm(60 * 6), nrow = 60)
  invalid <- list(
    "bogus", "o", NULL, character(), c("or", "and"), NA_character_, 1
  )

  for (method in c("mb", "tiger")) {
    for (value in invalid) {
      expect_error(
        huge(
          x, method = method, lambda = c(.4, .2),
          sym = value, verbose = FALSE
        ),
        "sym",
        info = paste(method, deparse(value))
      )
    }
    expect_no_error(huge(
      x, method = method, lambda = c(.4, .2),
      sym = "or", verbose = FALSE
    ))
    expect_no_error(huge(
      x, method = method, lambda = c(.4, .2),
      sym = "and", verbose = FALSE
    ))
  }
})

run_estimator_entry <- function(
  method, x, direct, input.type = NULL, lambda = .2
) {
  args <- list(x = x, lambda = lambda, verbose = FALSE)
  if(!is.null(input.type))
    args$input.type <- input.type
  if (direct) {
    do.call(get(paste0("huge.", method)), args)
  } else {
    args$method <- method
    do.call(huge, args)
  }
}

test_that("input.type disambiguates square symmetric observations", {
  x <- matrix(c(
    2, 1, 0,
    1, 2, 1,
    0, 1, 2
  ), 3, 3, byrow = TRUE)

  for(method in c("ct", "mb", "glasso", "tiger")) {
    for(direct in c(FALSE, TRUE)) {
      automatic <- run_estimator_entry(method, x, direct, lambda = 1)
      data.fit <- run_estimator_entry(
        method, x, direct, input.type = "data", lambda = 1
      )
      covariance.fit <- run_estimator_entry(
        method, x, direct, input.type = "covariance", lambda = 1
      )

      expect_true(automatic$cov.input, info = paste(method, direct))
      expect_false(data.fit$cov.input, info = paste(method, direct))
      expect_true(covariance.fit$cov.input, info = paste(method, direct))
    }
  }
})

test_that("TIGER input.type keeps correlation and lambda selection native", {
  x <- matrix(c(
    2, 1, 0,
    1, 2, 1,
    0, 1, 2
  ), 3, 3, byrow = TRUE)

  raw.native <- .Call(
    "_huge_SPMBgraphsqrtFit", x, NULL, 1L, 3L, FALSE, .1,
    PACKAGE = "huge"
  )
  covariance.native <- .Call(
    "_huge_SPMBgraphsqrtFit", x, NULL, 1L, 3L, TRUE, .1,
    PACKAGE = "huge"
  )
  raw.fit <- huge.tiger(
    x, nlambda = 1, input.type = "data", verbose = FALSE
  )
  covariance.fit <- huge.tiger(
    x, nlambda = 1, input.type = "covariance", verbose = FALSE
  )
  automatic <- huge.tiger(x, nlambda = 1, verbose = FALSE)

  expect_identical(raw.fit$lambda, raw.native$lambda)
  expect_identical(covariance.fit$lambda, covariance.native$lambda)
  expect_identical(automatic$lambda, covariance.fit$lambda)
  expect_false(identical(raw.fit$lambda, covariance.fit$lambda))

  routed <- huge:::.huge_validate_estimation_input(
    x, input.type = "covariance", prepare.covariance = FALSE
  )
  expect_true(routed$cov.input)
  expect_null(routed$covariance)
  expect_null(routed$correlation)
})

test_that("input.type rejects invalid and incompatible declarations", {
  x <- matrix(seq_len(6), 3, 2)

  for(value in list("raw", NA_character_, c("data", "auto"), 1)) {
    expect_error(
      huge.ct(x, lambda = 1, input.type = value, verbose = FALSE),
      "input.type"
    )
  }
  expect_error(
    huge.ct(
      x, lambda = 1, input.type = "covariance", verbose = FALSE
    ),
    "square"
  )
  expect_error(
    huge.tiger(
      matrix(c(1, 0, 1, 1), 2, 2),
      lambda = 1, input.type = "covariance", verbose = FALSE
    ),
    "symmetric"
  )
})

test_that("estimator entries reject malformed and non-finite inputs", {
  invalid <- list(
    vector = list(value = 1:4, message = "numeric matrix"),
    nonnumeric = list(
      value = matrix(letters[1:4], 2, 2), message = "numeric matrix"
    ),
    zero.rows = list(
      value = matrix(numeric(), 0, 2), message = "non-empty"
    ),
    zero.columns = list(
      value = matrix(numeric(), 2, 0), message = "non-empty"
    ),
    missing = list(
      value = matrix(c(1, 2, 3, NA, 5, 6), 3, 2), message = "finite"
    ),
    infinite = list(
      value = matrix(c(1, 2, 3, Inf, 5, 6), 3, 2), message = "finite"
    )
  )

  for (method in c("ct", "mb", "glasso", "tiger")) {
    for (direct in c(FALSE, TRUE)) {
      for (case in invalid) {
        expect_error(
          run_estimator_entry(method, case$value, direct),
          case$message,
          info = paste(method, if (direct) "direct" else "huge")
        )
      }
    }
  }
})

test_that("estimator entries reject undefined raw correlations", {
  invalid <- list(
    one.observation = list(
      value = matrix(c(1, 2, 3), 1, 3),
      message = "at least two observations"
    ),
    constant.column = list(
      value = cbind(seq_len(4), rep(1, 4)),
      message = "constant column"
    )
  )

  for (method in c("ct", "mb", "glasso", "tiger")) {
    for (direct in c(FALSE, TRUE)) {
      for (case in invalid) {
        expect_error(
          run_estimator_entry(method, case$value, direct),
          case$message,
          info = paste(method, if (direct) "direct" else "huge")
        )
      }
    }
  }
})

test_that("estimator entries reject invalid covariance matrices", {
  invalid <- list(
    zero.diagonal = list(
      value = diag(c(0, 1)), message = "positive.*diagonal"
    ),
    negative.diagonal = list(
      value = diag(c(-1, 1)), message = "positive.*diagonal"
    ),
    cauchy.schwarz = list(
      value = matrix(c(1, 2, 2, 1), 2, 2),
      message = "valid covariance"
    )
  )

  for (method in c("ct", "mb", "glasso", "tiger")) {
    for (direct in c(FALSE, TRUE)) {
      for (case in invalid) {
        expect_error(
          run_estimator_entry(method, case$value, direct),
          case$message,
          info = paste(method, if (direct) "direct" else "huge")
        )
      }
    }
  }
})

test_that("non-PSD covariance is accepted only when glasso regularizes it", {
  S <- matrix(c(
    1, .9, .9,
    .9, 1, 0,
    .9, 0, 1
  ), 3, 3, byrow = TRUE)

  for(method in c("ct", "mb", "tiger")) {
    for(direct in c(FALSE, TRUE)) {
      expect_error(
        run_estimator_entry(method, S, direct, lambda = .5),
        "positive semidefinite",
        info = paste(method, if (direct) "direct" else "huge")
      )
    }
  }

  for(direct in c(FALSE, TRUE)) {
    args <- list(
      x = S, lambda = .5, cov.output = TRUE, verbose = FALSE
    )
    if(direct) {
      fit <- do.call(huge.glasso, args)
    } else {
      args$method <- "glasso"
      fit <- do.call(huge, args)
    }
    precision <- fit$icov[[1]]
    covariance <- fit$cov[[1]]
    expect_gt(
      min(eigen(precision, symmetric = TRUE, only.values = TRUE)$values),
      0
    )
    expect_gt(
      min(eigen(covariance, symmetric = TRUE, only.values = TRUE)$values),
      0
    )
    expect_true(
      max(rowSums(abs(covariance %*% precision - diag(3)))) <= 1e-2
    )
  }
})

test_that("covariance validation accepts PSD and spectral roundoff", {
  roundoff <- matrix(-.5 - 1e-14, 3, 3)
  diag(roundoff) <- 1
  cases <- list(singular = matrix(1, 2, 2), roundoff = roundoff)

  for (S in cases) {
    for (method in c("ct", "mb", "glasso", "tiger")) {
      expect_no_error(huge(
        S, method = method, lambda = 1, verbose = FALSE
      ))
    }
  }
})

test_that("covariance projection preserves exact subnormal diagonal", {
  smallest <- .Machine$double.xmin * .Machine$double.eps
  S <- diag(rep(smallest, 2))
  expect_gt(smallest, 0)

  for (method in c("ct", "mb", "glasso", "tiger")) {
    for (direct in c(FALSE, TRUE)) {
      fit <- run_estimator_entry(method, S, direct)
      expect_true(fit$cov.input, info = paste(method, direct))
      expect_identical(fit$sparsity, 0, info = paste(method, direct))
      expect_identical(
        as.matrix(fit$path[[1]]), matrix(0, 2, 2),
        info = paste(method, direct)
      )
      if (!is.null(fit$icov)) {
        expect_true(
          all(is.finite(fit$icov[[1]])), info = paste(method, direct)
        )
      }
    }
  }
})

test_that("covariance validation clips only correlation roundoff", {
  S <- matrix(c(1, 1 + 5e-9, 1 + 5e-9, 1), 2, 2)
  fit <- huge.ct(S, lambda = .9, verbose = FALSE)

  expect_true(fit$cov.input)
  expect_equal(sum(fit$path[[1]]), 2)
})

test_that("accepted near-symmetric covariance is projected consistently", {
  S <- diag(3)
  S[1, 2] <- .2 + 1e-15
  S[2, 1] <- .2 - 1e-15
  expect_true(isSymmetric(S))
  projected <- S / 2 + t(S) / 2

  for (method in c("ct", "mb", "glasso", "tiger")) {
    fit <- huge(S, method = method, lambda = .2, verbose = FALSE)
    reference <- huge(
      projected, method = method, lambda = .2, verbose = FALSE
    )

    expect_identical(
      as.matrix(fit$path[[1]]), as.matrix(reference$path[[1]]),
      info = method
    )
    expect_identical(
      as.matrix(fit$path[[1]]), t(as.matrix(fit$path[[1]])),
      info = method
    )
  }
})

test_that("R covariance detection uses the native per-entry tolerance", {
  aggregate.roundoff <- matrix(c(1, 1e-14, 3e-14, 1), 2, 2)
  expect_true(isSymmetric(aggregate.roundoff))

  fit <- huge.tiger(aggregate.roundoff, lambda = 1, verbose = FALSE)
  expect_false(fit$cov.input)

  near <- matrix(c(1, .5, .5 + 1e-15, 1), 2, 2)
  expect_true(huge:::.huge_is_covariance_input(near))
  expect_true(huge.tiger(near, lambda = 1, verbose = FALSE)$cov.input)

  maximum <- .Machine$double.xmax
  large.asymmetry <- matrix(
    c(maximum, maximum, maximum * (1 - 1e-13), maximum), 2, 2
  )
  expect_true(isSymmetric(large.asymmetry))
  expect_false(huge:::.huge_is_covariance_input(large.asymmetry))

  maximum.integer <- .Machine$integer.max
  integer.asymmetry <- matrix(
    c(1L, -maximum.integer, maximum.integer, 1L), 2, 2
  )
  expect_no_warning(
    expect_false(huge:::.huge_is_covariance_input(integer.asymmetry))
  )
  expect_no_warning({
    fit <- huge.tiger(integer.asymmetry, lambda = 1, verbose = FALSE)
  })
  expect_false(fit$cov.input)
})

test_that("covariance detection is invariant to finite uniform scale", {
  raw = matrix(c(1, 4, 7, 2, 5, 8, 3, 6, 10), 3, 3)
  reference = huge.ct(raw, lambda = .95, verbose = FALSE)
  tiger.reference = huge.tiger(raw, nlambda = 1, verbose = FALSE)
  minimum = .Machine$double.xmin * .Machine$double.eps

  for (factor in c(1, 1e-15, 1e-200, minimum)) {
    current = raw * factor
    expect_false(huge:::.huge_is_covariance_input(current))

    for(method in c("mb", "glasso")) {
      expect_false(
        huge(
          current, method = method, lambda = .95, verbose = FALSE
        )$cov.input,
        info = paste(method, format(factor))
      )
    }

    fit = huge.ct(current, lambda = .95, verbose = FALSE)
    expect_false(fit$cov.input)
    expect_identical(
      as.matrix(fit$path[[1]]), as.matrix(reference$path[[1]])
    )

    tiger = huge.tiger(current, nlambda = 1, verbose = FALSE)
    expect_false(tiger$cov.input)
    expect_equal(tiger$lambda, tiger.reference$lambda, tolerance = 1e-14)

    expect_error(
      .Call(
        "_huge_SPMBgraphsqrtFit", current, 1, 1L, 3L, TRUE, .1,
        PACKAGE = "huge"
      ),
      "symmetric"
    )
  }

  near = matrix(c(1, .5, .5 + 1e-15, 1), 2, 2)
  near = near * 1e-200
  expect_true(huge:::.huge_is_covariance_input(near))
  expect_true(huge.tiger(near, lambda = 1, verbose = FALSE)$cov.input)
  expect_no_error(
    .Call(
      "_huge_SPMBgraphsqrtFit", near, 1, 1L, 2L, TRUE, .1,
      PACKAGE = "huge"
    )
  )

  extreme = matrix(
    c(.Machine$double.xmax, -1, 1, 1e-300), 2, 2
  )
  expect_false(huge:::.huge_is_covariance_input(extreme))
})

test_that("covariance symmetry tolerance uses implied-correlation scale", {
  tolerance = 100 * .Machine$double.eps
  native.tiger = function(S) {
    .Call(
      "_huge_SPMBgraphsqrtFit", S, 1, 1L, 2L, TRUE, .1,
      PACKAGE = "huge"
    )
  }

  for(variance in c(.Machine$double.xmin, 1, .Machine$double.xmax)) {
    within = matrix(c(
      variance, tolerance * variance,
      0, variance
    ), 2, 2)
    outside = matrix(c(
      variance, 101 * .Machine$double.eps * variance,
      0, variance
    ), 2, 2)

    expect_true(huge:::.huge_is_covariance_input(within))
    expect_no_error(native.tiger(within))
    expect_false(huge:::.huge_is_covariance_input(outside))
    expect_error(native.tiger(outside), "symmetric")
  }

  maximum = .Machine$double.xmax
  negligible = matrix(c(maximum, -1, 1, maximum), 2, 2)
  expect_true(huge:::.huge_is_covariance_input(negligible))
  expect_true(
    huge.tiger(negligible, lambda = 1, verbose = FALSE)$cov.input
  )
  expect_no_error(native.tiger(negligible))

  extreme = matrix(c(maximum, -1, 1, 1e-300), 2, 2)
  expect_false(huge:::.huge_is_covariance_input(extreme))
  expect_error(native.tiger(extreme), "symmetric")

  invalid.diagonal = matrix(c(0, -1, 1, 1), 2, 2)
  expect_error(native.tiger(invalid.diagonal), "positive.*diagonal")
})

#-----------------------------------------------------------------------#
# Package: High-dimensional Undirected Graph Estimation                 #
# huge.inference(): graph inference                                     #
#-----------------------------------------------------------------------#

#' Graph inference
#'
#' Implements the inference for high dimensional graphical models, including Gaussian and Nonparanormal graphical models
#' We consider the problems of testing the presence of a single edge and the hypothesis is that the edge is absent.
#'
#' For Nonparanormal graphical model we provide Score test method and Wald Test. However it is really slow for inferencing on Nonparanormal model, especially for large data. Gaussian inference supports one variable, while Nonparanormal inference requires at least two. Nonparanormal score-test diagonal p-values do not represent edges and may be undefined; every tested off-diagonal p-value must be finite.
#'
#'
#' @param data A finite numeric \code{n} by \code{d} data matrix with at least two observations and no constant columns.
#' @param T A finite \code{d} by \code{d} estimate of the inverse correlation matrix with a positive diagonal.
#' @param adj A finite numeric or logical \code{d} by \code{d} adjacency matrix corresponding to the graph.
#' @param alpha The significance level in \code{(0, 1]}. The default value is \code{0.05}.
#' @param type The type of input data. There are 2 options: \code{"Gaussian"} and \code{"Nonparanormal"}. The default value is \code{"Gaussian"}.
#' @param method For a Nonparanormal model, the test method: \code{"score"} or \code{"wald"}. The default is \code{"score"}. Ignored for Gaussian inference.
#' @seealso \code{\link{huge}}, and \code{\link{huge-package}}.
#' @return
#' An object is returned:
#' \item{data}{
#'   The \code{n} by \code{d} data matrix from the input.
#' }
#' \item{p}{
#'   The \code{d} by \code{d} p-value matrix of hypothesis.
#' }
#' \item{error}{
#'   The type I error of hypothesis at alpha significance level.
#' }
#' @examples
#' #generate data
#' L = huge.generator(n = 50, d = 12, graph = "hub", g = 4)
#'
#' #graph path estimation using glasso
#' est = huge(L$data, method = "glasso")
#'
#' #inference of Gaussian graphical model at 0.05 significance level
#' T = tail(est$icov, 1)[[1]]
#' out1 = huge.inference(L$data, T, L$theta)
#'
#' #inference of Nonparanormal graphical model using score test at 0.05 significance level
#' T = tail(est$icov, 1)[[1]]
#' out2 = huge.inference(L$data, T, L$theta, type = "Nonparanormal")
#'
#' #inference of Nonparanormal graphical model using wald test at 0.05 significance level
#' T = tail(est$icov, 1)[[1]]
#' out3 = huge.inference(L$data, T, L$theta, type = "Nonparanormal", method = "wald")
#'
#' #inference of Nonparanormal graphical model using wald test at 0.1 significance level
#' T = tail(est$icov, 1)[[1]]
#' out4 = huge.inference(L$data, T, L$theta, 0.1, type = "Nonparanormal", method = "wald")
#' @references
#' 1.Q Gu, Y Cao, Y Ning, H Liu. Local and global inference for high dimensional nonparanormal graphical models.\cr
#' 2.J Jankova, S Van De Geer. Confidence intervals for high-dimensional inverse covariance estimation. \emph{Electronic Journal of Statistics}, 2015.\cr
#' @export
huge.inference = function(data, T, adj, alpha = 0.05, type = "Gaussian", method = "score"){
  if(length(type) != 1 || !is.character(type) || is.na(type) ||
     !(type %in% c("Gaussian", "Nonparanormal")))
    stop("type must be exactly one of \"Gaussian\" or \"Nonparanormal\".")
  if(type == "Nonparanormal" &&
     (length(method) != 1 || !is.character(method) || is.na(method) ||
      !(method %in% c("score", "wald"))))
    stop("method must be exactly one of \"score\" or \"wald\".")
  if(length(alpha) != 1 || !is.numeric(alpha) || !is.finite(alpha) ||
     alpha <= 0 || alpha > 1)
    stop("alpha must be a finite number in (0, 1].")

  if(!(is.matrix(data) || inherits(data, "Matrix")))
    stop("data must be a numeric matrix.")
  data.input = data
  data = as.matrix(data)
  if(!is.numeric(data))
    stop("data must be a numeric matrix.")
  d = ncol(data)
  n = nrow(data)
  if(n < 1 || d < 1)
    stop("data must be a non-empty numeric matrix.")
  if(any(!is.finite(data)))
    stop("data must contain only finite values.")
  if(n < 2)
    stop("Inference requires at least two observations.")
  if(type == "Nonparanormal" && d < 2)
    stop("Nonparanormal inference requires at least two variables.")
  constant = vapply(seq_len(d), function(j) {
    all(data[, j] == data[1, j])
  }, logical(1))
  if(any(constant))
    stop("Inference data contains a constant column.")

  if(!(is.matrix(T) || inherits(T, "Matrix")))
    stop(sprintf("T must be a numeric %d by %d matrix.", d, d))
  T = as.matrix(T)
  if(!is.numeric(T) || !identical(dim(T), c(d, d)))
    stop(sprintf("T must be a numeric %d by %d matrix.", d, d))
  if(any(!is.finite(T)))
    stop("T must contain only finite values.")

  if(!(is.matrix(adj) || inherits(adj, "Matrix")))
    stop(sprintf("adj must be a numeric or logical %d by %d matrix.", d, d))
  adj = as.matrix(adj)
  if((!is.numeric(adj) && !is.logical(adj)) ||
     !identical(dim(adj), c(d, d)))
    stop(sprintf("adj must be a numeric or logical %d by %d matrix.", d, d))
  if(any(!is.finite(adj)))
    stop("adj must contain only finite values.")
  if(any(diag(T) <= 0))
    stop("T must have a positive diagonal.")

  if(type == "Gaussian")
  {
    U = tryCatch(
      suppressWarnings(.huge_fast_cor(data)),
      error = function(e) NULL
    )
    if(is.null(U) || any(!is.finite(U)))
      stop("Gaussian inference cannot form a finite correlation matrix.")
    # De-biased estimator W = 2T - T U T with asymptotic standard deviation
    # sigma[j,k] = sqrt(T[j,j] T[k,k] + T[j,k]^2) (Jankova & van de Geer).
    variance = outer(diag(T), diag(T)) + T^2
    if(any(!is.finite(variance)) || any(variance <= 0))
      stop(paste(
        "Gaussian inference variance must be finite and positive;",
        "check the scale of T."
      ))
    W = 2*T - T%*%U%*%T
    sigma = sqrt(variance)
    p = 2*(1 - pnorm(abs(sqrt(n)*W/sigma)))
  }
  if(type == "Nonparanormal")
  {
    diag.outer = outer(diag(T), diag(T))
    if(any(!is.finite(diag.outer)) || any(diag.outer <= 0))
      stop("Products of T diagonal entries must remain finite and positive.")
    x=data
    U<-matrix(0,d,d)
    G=list()
    Test<-matrix(0,d,d)
    for(i in 1:n)
      G[[i]]<-matrix(0,d,d)
    Temp_jk<-matrix(0,n,n)
    for(j in 1:d)
    {
      for(k in 1:d)
      {
        if(j==k)
        {
          U[j, k] = 1
          next
        }

        for(i1 in 1:n)
        {
          for(i2 in 1:n)
          {
            Temp_jk[i1, i2] = sign(x[i1, j] - x[i2, j]) *
              sign(x[i1, k] - x[i2, k])
            G[[i1]][j, k] = G[[i1]][j, k] - pi/2*Temp_jk[i1, i2]

          }
        }
        U[j, k] = sin(pi/2*sum(Temp_jk)/((n-1)*n))
        for(i in 1:n)
          G[[i]][j, k] = G[[i]][j, k]/(n-1) + asin(U[j, k])
      }
    }
    #F
    F<-apply(U,1,function(x) sqrt(1-x^2))
    #R
    R<-matrix(0,d^2,d^2)
    for (i in 1:n)
      R<-R + as.matrix(as.vector(F*G[[i]]))%*%as.vector(F*G[[i]])
    R<-R/n

    #kronecker product of T
    T_k<-kronecker(T, T)

    if(method == "score")
    {
      S<-matrix(0,d,d)
      sigma<-matrix(0,d,d)
      #ST_n
      ST_n<-matrix(0,d,d)
      for(j in 1:d)
      {
        for(k in 1:d)
        {
          #S
          ej<-matrix(0,d,1)
          ek<-matrix(0,d,1)
          ej[j] = 1
          ek[k] = 1
          T_h<-T
          T_h[j, k]=0
          S[j, k] = t(ej)%*%t(T_h)%*%U%*%T_h%*%ek/(T[j, j]*T[k, k])

          idx <- (k - 1) * d + j

          #w
          temp1<-T_k[,idx]
          temp1<-temp1[-idx]
          w<-as.matrix((-temp1)/T_k[idx, idx])

          #sigma
          temp2<-R[idx,]
          temp2<-as.matrix(temp2[-idx])
          sigma[j, k] = sqrt(R[idx, idx] - 2*t(temp2)%*%w + t(w)%*%R[-idx, -idx]%*%w)

          ST_n[j, k] = S[j, k]*sqrt(n)/(2*sigma[j, k])

        }
      }

      #p-value
      p<-2*(1 - pnorm(abs(ST_n)))
      rm(temp1,temp2,ST_n,S)
    }

    if(method == "wald")
    {
      T_W<-matrix(0,d,d)
      sigma<-matrix(0,d,d)
      #W_n
      W_n<-matrix(0,d,d)
      temp1<-T%*%U
      temp2<-U%*%T
      for(j in 1:d)
      {
        for(k in 1:d)
        {
          idx <- (k - 1) * d + j

          #w
          temp3<-T_k[,idx]
          temp3<-temp3[-idx]
          w<-as.matrix((-temp3)/T_k[idx,idx])

          #sigma
          temp4<-R[idx,]
          temp4<-as.matrix(temp4[-idx])
          sigma[j, k] = sqrt(R[idx, idx] - 2*t(temp4)%*%w + t(w)%*%R[-idx, -idx]%*%w)

          #T_W
          ej<-matrix(0,d,1)
          ek<-matrix(0,d,1)
          ej[j] = 1
          ek[k] = 1
          T_W[j ,k] = (T[j,k]*t(ej)%*%temp1%*%ek + T[j, k]*t(ej)%*%temp2%*%ek - t(ej)%*%t(T)%*%temp2%*%ek)/(t(ej)%*%temp1%*%ek + t(ej)%*%temp2%*%ek - 1)
          W_n[j, k] = T_W[j, k]*sqrt(n)/(2*sigma[j, k]*T[j, j]*T[k, k])

        }
      }


      #p-value
      p<-2*(1 - pnorm(abs(W_n)))
      rm(temp1,temp2,temp3,temp4,T_W,W_n)
    }
    rm(R,F,G,w,T_k)
  }

  offdiag = row(p) != col(p)
  finite.p = if(type == "Gaussian") {
    all(is.finite(p))
  } else {
    all(is.finite(p[offdiag]))
  }
  if(!finite.p)
    stop(paste(
      "Inference produced non-finite edge p-values;",
      "the inputs are numerically degenerate."
    ))

  error=0
  for(j in 1:d)
  {
    for(k in 1:d)
    {
      if(j==k)
        next
      if(p[j, k]<alpha && adj[j, k]==0){
        error=error+1
      }
    }
  }
  error=error/d^2


  inf = list()
  inf$data = data.input
  inf$p = p
  inf$error = error

  rm(U,p)
  return(inf)
}

test_that("huge.inference Gaussian produces valid p-values", {
  set.seed(70)
  L <- huge.generator(n = 100, d = 15, graph = "hub", g = 3, verbose = FALSE)
  fit <- huge(L$data, method = "glasso", verbose = FALSE)
  T_hat <- tail(fit$icov, 1)[[1]]
  inf <- huge.inference(L$data, T_hat, L$theta)

  expect_true(!is.null(inf$p))
  expect_equal(dim(inf$p), c(15, 15))
  # p-values should be in [0, 1]
  expect_true(all(inf$p >= 0 & inf$p <= 1))
  # error rate should be in [0, 1]
  expect_true(inf$error >= 0 && inf$error <= 1)
})

test_that("huge.inference preserves the Gaussian one-variable boundary", {
  x <- matrix(0:3, ncol = 1)
  inf <- huge.inference(x, matrix(1, 1, 1), matrix(0, 1, 1))

  expect_equal(dim(inf$p), c(1L, 1L))
  expect_equal(inf$p[1, 1], 0.15729920705028505, tolerance = 1e-15)
  expect_identical(inf$error, 0)
})

test_that("huge.inference rejects degenerate data before arithmetic", {
  one.row <- matrix(c(1, 2, 3), nrow = 1)
  for (type in c("Gaussian", "Nonparanormal")) {
    expect_error(
      huge.inference(one.row, diag(3), matrix(0, 3, 3), type = type),
      "at least two observations",
      info = type
    )
  }

  one.variable <- matrix(0:3, ncol = 1)
  for (method in c("score", "wald")) {
    expect_error(
      huge.inference(
        one.variable, matrix(1, 1, 1), matrix(0, 1, 1),
        type = "Nonparanormal", method = method
      ),
      "at least two variables",
      info = method
    )
  }

  constant <- cbind(0:3, rep(1, 4))
  for (type in c("Gaussian", "Nonparanormal")) {
    expect_error(
      huge.inference(constant, diag(2), matrix(0, 2, 2), type = type),
      "constant column",
      info = type
    )
  }
})

test_that("huge.inference validates active options and matrix inputs", {
  set.seed(71)
  x <- matrix(rnorm(16), 8, 2)
  adj <- matrix(0, 2, 2)

  for (type in list("bad", c("Gaussian", "Nonparanormal"), 1, NA_character_)) {
    expect_error(huge.inference(x, diag(2), adj, type = type), "type")
  }
  expect_error(
    huge.inference(x, diag(2), adj, type = "Nonparanormal", method = "bad"),
    "method"
  )
  expect_no_error(huge.inference(x, diag(2), adj, method = "ignored"))

  for (alpha in list(0, -0.1, 1.1, NA_real_, Inf, TRUE, c(.05, .1))) {
    expect_error(huge.inference(x, diag(2), adj, alpha = alpha), "alpha")
  }
  expect_no_error(huge.inference(x, diag(2), adj, alpha = 1))

  expect_error(huge.inference(1:4, diag(2), adj), "data.*numeric matrix")
  bad.data <- x
  bad.data[1, 1] <- NA_real_
  expect_error(huge.inference(bad.data, diag(2), adj), "data.*finite")

  expect_error(huge.inference(x, diag(3), adj), "T.*2.*2")
  bad.t <- diag(2)
  bad.t[1, 2] <- Inf
  expect_error(huge.inference(x, bad.t, adj), "T.*finite")
  expect_error(
    huge.inference(x, diag(c(0, 1)), adj), "positive diagonal"
  )
  expect_error(
    huge.inference(x, diag(c(-1, 1)), adj), "positive diagonal"
  )

  expect_error(huge.inference(x, diag(2), matrix(0, 3, 3)), "adj.*2.*2")
  bad.adj <- adj
  bad.adj[1, 2] <- NaN
  expect_error(huge.inference(x, diag(2), bad.adj), "adj.*finite")

  sparse <- Matrix::Matrix(adj, sparse = TRUE)
  inf <- huge.inference(x, Matrix::Diagonal(2), sparse)
  expect_true(all(is.finite(inf$p)))

  sparse.data <- Matrix::Matrix(x, sparse = TRUE)
  sparse.inf <- huge.inference(sparse.data, diag(2), sparse)
  expect_identical(sparse.inf$data, sparse.data)
  expect_equal(sparse.inf$p, inf$p, tolerance = 0)
})

test_that("huge.inference rejects non-finite edge results", {
  set.seed(72)
  x <- matrix(rnorm(16), 8, 2)
  adj <- matrix(0, 2, 2)
  huge.t <- 1e308 * matrix(c(1, .5, .5, 1), 2, 2)
  tiny.t <- diag(2) * 1e-200

  for (T in list(huge.t, tiny.t)) {
    for (settings in list(
      list(type = "Gaussian", method = "score"),
      list(type = "Nonparanormal", method = "score"),
      list(type = "Nonparanormal", method = "wald")
    )) {
      expect_error(
        huge.inference(
          x, T, adj,
          type = settings$type, method = settings$method
        ),
        "finite|numerically",
        info = paste(settings$type, settings$method)
      )
    }
  }
})

test_that("Nonparanormal inference gates edges but keeps rank limits", {
  monotone <- matrix(c(0, 1, 0, 1), nrow = 2)
  score <- huge.inference(
    monotone, diag(2), matrix(0, 2, 2),
    type = "Nonparanormal", method = "score"
  )
  expect_true(all(is.nan(diag(score$p))))
  expect_true(all(is.finite(score$p[row(score$p) != col(score$p)])))

  extreme.ties <- cbind(c(1e308, -1e308, 0), c(0, 0, 1))
  for (method in c("score", "wald")) {
    inf <- huge.inference(
      extreme.ties, diag(2), matrix(0, 2, 2),
      type = "Nonparanormal", method = method
    )
    expect_true(
      all(is.finite(inf$p[row(inf$p) != col(inf$p)])),
      info = method
    )
  }
})

test_that("Gaussian inference is stable under extreme finite data scales", {
  data = cbind(
    c(-2, -1, 0, 1, 2, 3),
    c(3, -1, 2, -2, 1, 0)
  )
  reference = huge.inference(
    data, diag(2), matrix(0, 2, 2), type = "Gaussian"
  )

  for(factor in c(1e-300, 1e300)) {
    current = huge.inference(
      data * factor, diag(2), matrix(0, 2, 2), type = "Gaussian"
    )
    expect_equal(current$p, reference$p, tolerance = 1e-14)
    expect_equal(current$error, reference$error)
  }
})

#-----------------------------------------------------------------------#
# Package: High-dimensional Undirected Graph Estimation                 #
# huge.mb(): Meinshausen & Buhlmann graph estimation (mb)               #
#-----------------------------------------------------------------------#

#' Meinshausen & Buhlmann graph estimation
#'
#' See more details in \code{\link{huge}}
#' @param x There are 2 options: (1) \code{x} is an \code{n} by \code{d} data matrix (2) a \code{d} by \code{d} sample covariance matrix. The program automatically identifies the input matrix by checking the symmetry. (\code{n} is the sample size and \code{d} is the dimension).
#' @param lambda A numeric scalar or non-empty one-dimensional numeric input
#'   defining a finite, strictly positive, non-increasing regularization path;
#'   tied values are allowed. Leave \code{lambda = NULL}
#'   to generate a path from \code{nlambda} and \code{lambda.min.ratio}.
#' @param nlambda The number of regularization/thresholding parameters. The default value is \code{20} for \code{method = "ct"} and \code{10} for \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}.
#' @param lambda.min.ratio If \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}, it is the smallest value for \code{lambda}, as a fraction of the upperbound (\code{MAX}) of the regularization/thresholding parameter which makes all estimates equal to \code{0}. The program can automatically generate \code{lambda} as a sequence of length = \code{nlambda} starting from \code{MAX} to \code{lambda.min.ratio*MAX} in log scale. If \code{method = "ct"}, it is the largest sparsity level for estimated graphs. The program can automatically generate \code{lambda} as a sequence of length = \code{nlambda}, which makes the sparsity level of the graph path increases from \code{0} to \code{lambda.min.ratio} evenly.The default value is \code{0.1} when \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}, and 0.05 when \code{method = "ct"}.
#' @param scr If \code{scr = TRUE}, the lossy screening rule is applied to preselect the neighborhood before the graph estimation. The default value is  \code{FALSE}.
#' @param scr.num The neighborhood size after the lossy screening rule (the number of remaining neighbors per node). It must be an integer between \code{1} and \code{d - 1}. ONLY applicable when \code{scr = TRUE}. The default value is \code{n-1}. An alternative value is \code{n/log(n)}. ONLY applicable when \code{scr = TRUE} and \code{method = "mb"}.
#' @param idx.mat A \code{scr.num} by \code{d} index matrix for screening. Column \code{m} must contain distinct, zero-based indices in \code{0:(d - 1)} and must exclude the response index \code{m - 1}.
#' @param sym Symmetrize the output graphs. If \code{sym = "and"}, the edge between node \code{i} and node \code{j} is selected ONLY when both node \code{i} and node \code{j} are selected as neighbors for each other. If \code{sym = "or"}, the edge is selected when either node \code{i} or node \code{j} is selected as the neighbor for each other. The default value is \code{"or"}. ONLY applicable when \code{method = "mb"} or \code{"tiger"}.
#' @param verbose If \code{verbose = FALSE}, tracing information printing is disabled. The default value is \code{TRUE}.
#' @param input.type How to interpret \code{x}: \code{"auto"} preserves
#'   symmetry-based detection, \code{"data"} forces an observation matrix,
#'   and \code{"covariance"} requires a square covariance or correlation
#'   matrix.
#' @seealso \code{\link{huge}}, and \code{\link{huge-package}}.
#' @export
huge.mb = function(x, lambda = NULL, nlambda = NULL, lambda.min.ratio = NULL, scr = NULL, scr.num = NULL, idx.mat = NULL, sym = "or", verbose = TRUE, input.type = "auto")
{
  sym = .huge_validate_sym(sym)
  inp = .huge_preprocess(x, verbose, input.type = input.type)
  S = inp$S; n = inp$n; d = inp$d
  fit = list()
  fit$cov.input = inp$cov.input

  if(is.null(idx.mat))
  {
    if(is.null(scr))
      scr = FALSE

    if(scr)
    {
      if(is.null(scr.num))
      {
        if(n<d)
          scr.num = n-1
        if(n>=d)
        {
          if(verbose) cat("lossy screening is skipped without specifying scr.num.\n")
          scr = FALSE
        }
      }
    }
    fit$scr = scr
  }

  if(!is.null(idx.mat))
  {
    if(!is.matrix(idx.mat))
      stop("idx.mat must be a matrix.")
    if(ncol(idx.mat) != d)
      stop("idx.mat must have exactly d columns.")
    if(nrow(idx.mat) < 1 || nrow(idx.mat) >= d)
      stop("idx.mat must have between 1 and d - 1 rows.")
    if(!is.numeric(idx.mat) || any(!is.finite(idx.mat)) ||
       any(idx.mat != floor(idx.mat)))
      stop("idx.mat must contain finite integer-valued indices.")
    if(any(idx.mat < 0 | idx.mat >= d))
      stop("idx.mat must contain zero-based indices in 0:(d - 1).")

    idx.mat = matrix(as.integer(idx.mat), nrow = nrow(idx.mat), ncol = d)
    for(m in seq_len(d))
    {
      if(anyDuplicated(idx.mat[,m]))
        stop("Each column of idx.mat must contain distinct indices.")
      if((m - 1L) %in% idx.mat[,m])
        stop("Each column of idx.mat must exclude its response index.")
    }

    scr = TRUE
    fit$scr = scr
    scr.num = nrow(idx.mat)
  }

  lam = .huge_default_lambda(S, d, nlambda, lambda.min.ratio, lambda)
  lambda = lam$lambda; nlambda = lam$nlambda

  if(scr)
  {
    if(length(scr.num) != 1 || !is.numeric(scr.num) ||
       !is.finite(scr.num) || scr.num != floor(scr.num) ||
       scr.num < 1 || scr.num >= d)
      stop("scr.num must be an integer between 1 and d - 1.")
    scr.num = as.integer(scr.num)

    if(verbose)
    {
      cat("Conducting Meinshausen & Buhlmann graph estimation (mb) with lossy screening....")
      flush.console()
    }

    if(is.null(idx.mat))
    {
      idx.mat = vapply(seq_len(d), function(m) {
        candidates = seq_len(d)[-m]
        ord = order(-abs(S[candidates,m]), candidates)
        candidates[ord[seq_len(scr.num)]] - 1L
      }, integer(scr.num))
      idx.mat = matrix(idx.mat, nrow = scr.num, ncol = d)
    }

    fit$idx_mat = idx.mat
    out = .Call("_huge_SPMBscr", S, lambda, nlambda, d, idx.mat, scr.num, PACKAGE="huge")
  }
  if(!scr)
  {
    if(verbose)
    {
      cat("Conducting Meinshausen & Buhlmann graph estimation (mb)....")
      flush.console()
    }
    fit$idx_mat = NULL
    out = .Call("_huge_SPMBgraph", S, lambda, nlambda, d, PACKAGE= "huge")
  }
  if (isTRUE(out$hit_max_iter))
    warning("mb solver reached its iteration limit; estimates may not be fully converged.")

  # The core emits each column's row indices in ascending order (see
  # collect_sorted in huge_core.cpp), as dgCMatrix requires.
  nnz = out$col_cnz[d + 1]
  x_values = if(nnz > 0) out$x[seq_len(nnz)] else numeric(0)
  i_values = if(nnz > 0) out$row_idx[seq_len(nnz)] else integer(0)
  G = new("dgCMatrix", Dim = as.integer(c(d*nlambda,d)),
          x = as.vector(x_values), p = as.integer(out$col_cnz),
          i = as.integer(i_values))

  fit$beta = list()
  fit$path = list()
  fit$df = matrix(0,d,nlambda)
  fit$sparsity = rep(0,nlambda)
  for(i in 1:nlambda)
  {
    fit$beta[[i]] = G[((i-1)*d+1):(i*d),,drop = FALSE]
    fit$path[[i]] = abs(fit$beta[[i]])
    fit$df[,i] = Matrix::colSums(fit$path[[i]] != 0)

    # Matrix::t on the sparse matrix (not as.matrix) keeps the update sparse
    if(sym == "or")
      fit$path[[i]] = sign(fit$path[[i]] + Matrix::t(fit$path[[i]]))
    if(sym == "and")
      fit$path[[i]] = sign(fit$path[[i]] * Matrix::t(fit$path[[i]]))
    fit$sparsity[i] = if(d <= 1) 0 else sum(fit$path[[i]])/d/(d-1)
  }
  fit$lambda = lambda

  if(verbose)
  {
     cat("done\n")
      flush.console()
  }

  return(fit)
}

test_that("mb returns valid structure", {
  set.seed(20)
  L <- huge.generator(n = 80, d = 30, graph = "hub", verbose = FALSE)
  fit <- huge(L$data, method = "mb", verbose = FALSE)

  expect_s3_class(fit, "huge")
  expect_equal(fit$method, "mb")
  expect_equal(length(fit$path), length(fit$lambda))
  expect_equal(length(fit$sparsity), length(fit$lambda))
})

test_that("mb handles single-variable raw and covariance inputs", {
  raw <- matrix(c(-1, 0, 1), ncol = 1)
  covariance <- matrix(1, nrow = 1, ncol = 1)
  cases <- list(
    raw_auto = list(x = raw, nlambda = 3, lambda.min.ratio = 0.2),
    covariance_auto = list(
      x = covariance, nlambda = 3, lambda.min.ratio = 0.2
    ),
    raw_explicit = list(x = raw, lambda = c(0.5, 0.2)),
    covariance_explicit = list(x = covariance, lambda = c(0.5, 0.2))
  )

  fits <- lapply(cases, function(args) {
    do.call(huge.mb, c(args, list(verbose = FALSE)))
  })
  for (case in names(fits)) {
    fit <- fits[[case]]
    path_length <- if (grepl("auto$", case)) 3L else 2L
    expect_identical(
      fit$cov.input, grepl("^covariance", case), info = case
    )
    expect_length(fit$lambda, path_length)
    expect_true(all(is.finite(fit$lambda) & fit$lambda > 0), info = case)
    expect_identical(fit$sparsity, rep(0, path_length), info = case)
    expect_equal(dim(fit$df), c(1L, path_length), info = case)
    expect_identical(as.numeric(fit$df), rep(0, path_length), info = case)
    expect_length(fit$beta, path_length)
    expect_length(fit$path, path_length)
    for (index in seq_len(path_length)) {
      expect_true(inherits(fit$beta[[index]], "sparseMatrix"), info = case)
      expect_true(inherits(fit$path[[index]], "sparseMatrix"), info = case)
      expect_silent(validObject(fit$beta[[index]]))
      expect_silent(validObject(fit$path[[index]]))
      expect_equal(dim(fit$beta[[index]]), c(1L, 1L), info = case)
      expect_equal(dim(fit$path[[index]]), c(1L, 1L), info = case)
      expect_equal(sum(fit$beta[[index]]), 0, info = case)
      expect_equal(sum(fit$path[[index]]), 0, info = case)
    }
    expect_null(fit$icov, info = case)
  }

  expected_auto <- exp(seq(log(1e-3), log(2e-4), length.out = 3))
  # The generated path preserves lambda.max exactly instead of round-tripping
  # it through log()/exp(), which can move the endpoint by one ULP.
  expected_auto[[1L]] <- 1e-3
  expect_identical(fits$raw_auto$lambda, expected_auto)
  expect_identical(fits$raw_auto$lambda, fits$covariance_auto$lambda)
  expect_identical(fits$raw_explicit$lambda, c(0.5, 0.2))
  expect_identical(fits$raw_explicit$lambda, fits$covariance_explicit$lambda)
})

test_that("mb sparsity is non-decreasing", {
  set.seed(21)
  L <- huge.generator(n = 80, d = 30, graph = "hub", verbose = FALSE)
  fit <- huge(L$data, method = "mb", verbose = FALSE)

  expect_true(all(diff(fit$lambda) < 0))
  expect_true(all(diff(fit$sparsity) >= -1e-10))
})

test_that("mb path matrices are symmetric and binary", {
  set.seed(22)
  L <- huge.generator(n = 80, d = 30, graph = "band", verbose = FALSE)
  fit <- huge(L$data, method = "mb", verbose = FALSE)

  for (k in seq_along(fit$path)) {
    p <- as.matrix(fit$path[[k]])
    expect_equal(p, t(p), info = paste("path asymmetric at k =", k))
    expect_true(all(p %in% c(0, 1)))
    expect_true(all(diag(p) == 0))
  }
})

test_that("mb with scr=TRUE works", {
  set.seed(23)
  L <- huge.generator(n = 80, d = 30, graph = "hub", verbose = FALSE)
  fit <- huge(L$data, method = "mb", scr = TRUE, verbose = FALSE)

  expect_s3_class(fit, "huge")
  expect_true(all(diff(fit$sparsity) >= -1e-10))
})

test_that("mb sym='and' gives sparser graphs than sym='or'", {
  set.seed(24)
  L <- huge.generator(n = 80, d = 30, graph = "hub", verbose = FALSE)
  fit_or  <- huge(L$data, method = "mb", sym = "or",  verbose = FALSE)
  fit_and <- huge(L$data, method = "mb", sym = "and", verbose = FALSE)

  # "and" should be at least as sparse as "or" at each lambda
  for (k in seq_along(fit_or$path)) {
    expect_true(sum(fit_and$path[[k]]) <= sum(fit_or$path[[k]]) + 1e-10,
                info = paste("and not sparser than or at k =", k))
  }
})

test_that("mb works across graph types", {
  set.seed(25)
  for (g in c("hub", "band", "cluster", "random")) {
    L <- huge.generator(n = 60, d = 20, graph = g, verbose = FALSE)
    fit <- huge(L$data, method = "mb", verbose = FALSE)
    expect_true(all(diff(fit$sparsity) >= -1e-10),
                info = paste("non-monotone sparsity for graph =", g))
  }
})

test_that("mb screening excludes the response under tied correlations", {
  S <- matrix(1, nrow = 4, ncol = 4)

  fit <- huge.mb(
    S, lambda = 0.5, scr = TRUE, scr.num = 1, verbose = FALSE
  )

  expect_equal(dim(fit$idx_mat), c(1L, 4L))
  for (m in seq_len(ncol(S))) {
    expect_false((m - 1L) %in% fit$idx_mat[, m])
  }
  for (path in fit$path) {
    expect_true(all(diag(as.matrix(path)) == 0))
  }

  refit <- huge.mb(
    S, lambda = 0.5, idx.mat = fit$idx_mat, verbose = FALSE
  )
  expect_equal(refit$beta, fit$beta)
  expect_equal(refit$path, fit$path)
})

test_that("mb rejects unsafe screening index matrices", {
  S <- diag(4)
  valid <- matrix(
    c(1, 2, 0, 2, 0, 1, 0, 1), nrow = 2, ncol = 4
  )

  bad_inputs <- list(
    not_a_matrix = as.vector(valid),
    wrong_columns = valid[, 1:3, drop = FALSE],
    empty = matrix(integer(), nrow = 0, ncol = 4),
    out_of_range = replace(valid, 1, 4),
    internal_sentinel = replace(valid, 1, -1),
    negative = replace(valid, 1, -2),
    duplicate = replace(valid, 2, valid[1, 1]),
    self_index = replace(valid, 1, 0),
    missing = replace(valid, 1, NA_real_),
    infinite = replace(valid, 1, Inf),
    fractional = replace(valid, 1, 1.5)
  )

  for (case in names(bad_inputs)) {
    expect_error(
      huge.mb(
        S, lambda = 0.5, idx.mat = bad_inputs[[case]], verbose = FALSE
      ),
      "idx.mat",
      info = case
    )
  }
})

test_that("mb covariance default lambda uses correlation off-diagonals", {
  S <- matrix(
    c(100, .6, 0,
      .6, 3, -.2,
      0, -.2, .5),
    nrow = 3
  )
  correlation <- cov2cor(S)
  lambda.max <- max(abs(correlation[upper.tri(correlation)]))
  expected <- exp(seq(
    log(lambda.max), log(.25 * lambda.max), length.out = 3
  ))

  fit <- huge.mb(
    S, nlambda = 3, lambda.min.ratio = .25, verbose = FALSE
  )
  expect_equal(fit$lambda, expected, tolerance = 1e-15)
})

#-----------------------------------------------------------------------#
# Package: High-dimensional Undirected Graph Estimation                 #
# glasso(): The graphical lasso (glasso) using sparse matrix output     #
#-----------------------------------------------------------------------#

#' The graphical lasso (glasso) using sparse matrix output
#'
#' See more details in \code{\link{huge}}
#' @param x There are 2 options: (1) \code{x} is an \code{n} by \code{d} data matrix (2) a \code{d} by \code{d} sample covariance matrix. The program automatically identifies the input matrix by checking the symmetry. (\code{n} is the sample size and \code{d} is the dimension).
#' @param lambda A numeric scalar or non-empty one-dimensional numeric input
#'   defining a finite, strictly positive, non-increasing regularization path;
#'   tied values are allowed. Leave \code{lambda = NULL}
#'   to generate a path from \code{nlambda} and \code{lambda.min.ratio}.
#' @param nlambda The number of regularization/thresholding parameters. The default value is \code{20} for \code{method = "ct"} and \code{10} for \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}.
#' @param lambda.min.ratio If \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}, it is the smallest value for \code{lambda}, as a fraction of the upperbound (\code{MAX}) of the regularization/thresholding parameter which makes all estimates equal to \code{0}. The program can automatically generate \code{lambda} as a sequence of length = \code{nlambda} starting from \code{MAX} to \code{lambda.min.ratio*MAX} in log scale. If \code{method = "ct"}, it is the largest sparsity level for estimated graphs. The program can automatically generate \code{lambda} as a sequence of length = \code{nlambda}, which makes the sparsity level of the graph path increases from \code{0} to \code{lambda.min.ratio} evenly.The default value is \code{0.1} when \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}, and 0.05 when \code{method = "ct"}.
#' @param scr If \code{scr = TRUE}, the lossy screening rule is applied to preselect the neighborhood before the graph estimation. The default value is \code{FALSE}.
#' @param cov.output If \code{cov.output = TRUE}, the output will include a path of estimated covariance matrices. ONLY applicable when \code{method = "glasso"}. Since the estimated covariance matrices are generally not sparse, please use it with care, or it may take much memory under high-dimensional setting. The default value is \code{FALSE}.
#' @param verbose If \code{verbose = FALSE}, tracing information printing is disabled. The default value is \code{TRUE}.
#' @param input.type How to interpret \code{x}: \code{"auto"} preserves
#'   symmetry-based detection, \code{"data"} forces an observation matrix,
#'   and \code{"covariance"} requires a square covariance or correlation
#'   matrix. For covariance input with \code{lambda = NULL}, \code{"auto"}
#'   retains the historical diagonal-sensitive lambda scale; use
#'   \code{"covariance"} for a scale based only on off-diagonal entries.
#'   Indefinite pairwise covariance estimates are accepted only when
#'   regularization produces a certified positive-definite result.
#' @seealso \code{\link{huge}}, and \code{\link{huge-package}}.
#' @export
huge.glasso = function(x, lambda = NULL, lambda.min.ratio = NULL, nlambda = NULL, scr = NULL, cov.output = FALSE, verbose = TRUE, input.type = "auto"){

  inp = .huge_validate_estimation_input(
    x, input.type = input.type, require.psd = FALSE
  )
  x = inp$input
  n = inp$n
  d = inp$d
  cov.input = inp$cov.input
  if(cov.input)
  {
    if(verbose) cat("The input is identified as the covariance matrix.\n")
    S = inp$covariance
  }
  else
  {
    S = .huge_fast_cor(x)
    if(any(!is.finite(S)))
      stop("Raw data x cannot produce a finite correlation matrix.")
  }
  if(is.null(scr)) scr = FALSE
  lam = .huge_default_lambda(
    S, d, nlambda, lambda.min.ratio, lambda,
    legacy.glasso.covariance =
      cov.input && input.type == "auto"
  )
  lambda = lam$lambda; nlambda = lam$nlambda

  fit = .Call("_huge_hugeglasso",S,lambda,scr,verbose,cov.output,PACKAGE="huge")

  if (isTRUE(fit$hit_max_iter))
    warning("glasso solver reached its iteration limit; estimates may not be fully converged.")
  fit$hit_max_iter = NULL

  fit$scr = scr
  fit$lambda = lambda
  fit$cov.input = cov.input
  fit$cov.output = cov.output

  if(verbose){
       cat("\nConducting the graphical lasso (glasso)....done.                                          \r")
       cat("\n")
      flush.console()
  }
  return(fit)
}

test_that("glasso returns valid structure", {
  set.seed(10)
  L <- huge.generator(n = 80, d = 30, graph = "hub", verbose = FALSE)
  fit <- huge(L$data, method = "glasso", verbose = FALSE)

  expect_s3_class(fit, "huge")
  expect_equal(fit$method, "glasso")
  expect_equal(length(fit$path), length(fit$lambda))
  expect_equal(length(fit$icov), length(fit$lambda))
  expect_equal(length(fit$sparsity), length(fit$lambda))
  expect_equal(length(fit$loglik), length(fit$lambda))
  expect_equal(length(fit$df), length(fit$lambda))
})

test_that("glasso lambda path is decreasing and sparsity is non-decreasing", {
  set.seed(11)
  L <- huge.generator(n = 80, d = 30, graph = "hub", verbose = FALSE)
  fit <- huge(L$data, method = "glasso", verbose = FALSE)

  expect_true(all(diff(fit$lambda) < 0))
  expect_true(all(diff(fit$sparsity) >= -1e-10))
  expect_true(all(diff(fit$df) >= 0))
})

test_that("glasso loglik values are finite", {
  set.seed(12)
  L <- huge.generator(n = 80, d = 30, graph = "band", verbose = FALSE)
  fit <- huge(L$data, method = "glasso", verbose = FALSE)
  expect_true(all(is.finite(fit$loglik)))
})

test_that("glasso loglik matches the returned precision matrices", {
  cases <- list(
    list(S = diag(c(2, 3)), lambda = 0.5),
    list(
      S = matrix(c(2, 0.6, 0,
                   0.6, 3, 0,
                   0, 0, 4), 3, 3),
      lambda = c(0.4, 0.2)
    )
  )

  for (case in cases) {
    fit <- huge(case$S, method = "glasso", lambda = case$lambda,
                verbose = FALSE)
    for (k in seq_along(fit$icov)) {
      precision <- fit$icov[[k]]
      direct <- as.numeric(determinant(precision, logarithm = TRUE)$modulus) -
        sum(case$S * precision)
      expect_equal(fit$loglik[k], direct, tolerance = 1e-10,
                   info = paste("lambda index", k))
    }
  }
})

test_that("glasso path matrices are symmetric", {
  set.seed(13)
  L <- huge.generator(n = 80, d = 30, graph = "cluster", verbose = FALSE)
  fit <- huge(L$data, method = "glasso", verbose = FALSE)

  for (k in seq_along(fit$path)) {
    p <- as.matrix(fit$path[[k]])
    expect_equal(p, t(p), info = paste("path asymmetric at k =", k))
    expect_true(all(diag(p) == 0))
  }
})

test_that("glasso icov matrices are symmetric", {
  set.seed(14)
  L <- huge.generator(n = 80, d = 30, graph = "hub", verbose = FALSE)
  fit <- huge(L$data, method = "glasso", verbose = FALSE)

  for (k in seq_along(fit$icov)) {
    ic <- fit$icov[[k]]
    expect_equal(ic, t(ic), tolerance = 1e-4,
                 info = paste("icov asymmetric at k =", k))
  }
})

test_that("glasso precision symmetrization avoids finite overflow", {
  S <- 1e-307 * matrix(c(1, .99, .99, 1), nrow = 2)
  fit <- huge.glasso(
    S, lambda = 1e-309, cov.output = TRUE, verbose = FALSE
  )

  precision <- fit$icov[[1]]
  expect_true(all(is.finite(precision)))
  expect_identical(precision, t(precision))
  expect_true(all(is.finite(fit$cov[[1]])))
  expect_true(all(is.finite(fit$loglik)))
  direct <- as.numeric(determinant(precision, logarithm = TRUE)$modulus) -
    sum(S * precision)
  expect_equal(fit$loglik[1], direct, tolerance = 1e-10)
})

test_that("glasso log-likelihood has no absolute pivot scale", {
  S <- 1e20 * matrix(c(1, .5, .5, 1), nrow = 2)
  fit <- huge.glasso(
    S, lambda = 1e19, cov.output = TRUE, verbose = FALSE
  )

  precision <- fit$icov[[1]]
  expect_true(all(is.finite(precision)))
  expect_true(all(is.finite(fit$cov[[1]])))
  direct <- as.numeric(determinant(precision, logarithm = TRUE)$modulus) -
    sum(S * precision)
  expect_true(is.finite(direct))
  expect_equal(fit$loglik[1], direct, tolerance = 1e-10)
})

test_that("glasso rejects non-finite native results", {
  for (scale in c(1e-320, 1e308)) {
    for (cov.output in c(FALSE, TRUE)) {
      expect_error(
        huge.glasso(
          diag(scale, 2), lambda = scale,
          cov.output = cov.output, verbose = FALSE
        ),
        "non-finite"
      )
    }
  }
})

test_that("glasso with scr=TRUE produces valid results", {
  set.seed(15)
  L <- huge.generator(n = 80, d = 30, graph = "hub", verbose = FALSE)
  fit <- huge(L$data, method = "glasso", scr = TRUE, verbose = FALSE)

  expect_true(all(diff(fit$sparsity) >= -1e-10))
  expect_true(all(is.finite(fit$loglik)))
})

test_that("glasso cov.output works", {
  set.seed(16)
  L <- huge.generator(n = 80, d = 30, graph = "hub", verbose = FALSE)
  fit <- huge(L$data, method = "glasso", cov.output = TRUE, verbose = FALSE)

  expect_true(!is.null(fit$cov))
  expect_equal(length(fit$cov), length(fit$lambda))
  for (k in seq_along(fit$cov)) {
    co <- fit$cov[[k]]
    expect_equal(dim(co), c(30, 30))
    expect_equal(co, t(co), tolerance = 1e-10)
  }
})

test_that("glasso accepts covariance matrix input", {
  set.seed(17)
  L <- huge.generator(n = 80, d = 30, graph = "hub", verbose = FALSE)
  S <- cor(L$data)
  fit <- huge(S, method = "glasso", verbose = FALSE)

  expect_s3_class(fit, "huge")
  expect_true(fit$cov.input)
  expect_true(all(diff(fit$sparsity) >= -1e-10))
})

test_that("glasso default lambda ignores covariance diagonal scale", {
  expected.identity <- c(1e-3, sqrt(1e-7), 1e-4)
  diagonal <- diag(c(100, 3, .5, 8))
  diagonal.fit <- huge.glasso(
    diagonal, nlambda = 3, lambda.min.ratio = .1, verbose = FALSE,
    input.type = "covariance"
  )

  expect_equal(diagonal.fit$lambda, expected.identity, tolerance = 1e-15)
  expect_true(all(vapply(
    diagonal.fit$path, function(path) sum(path) == 0, logical(1)
  )))
  expect_true(all(vapply(
    diagonal.fit$icov, function(icov) all(is.finite(icov)), logical(1)
  )))

  S <- matrix(
    c(100, .6, 0,
      .6, 3, -.2,
      0, -.2, .5),
    nrow = 3
  )
  covariance.fit <- huge.glasso(
    S, nlambda = 3, lambda.min.ratio = .25, verbose = FALSE,
    input.type = "covariance"
  )
  expect_equal(covariance.fit$lambda, c(.6, .3, .15),
               tolerance = 1e-15)

  weak <- diag(3)
  weak[1, 2] <- weak[2, 1] <- 1e-6
  weak.fit <- huge.glasso(
    weak, nlambda = 2, lambda.min.ratio = .1, verbose = FALSE,
    input.type = "covariance"
  )
  expect_equal(weak.fit$lambda, c(1e-6, 1e-7), tolerance = 1e-20)

  explicit <- c(.55, .2)
  expect_identical(
    huge.glasso(S, lambda = explicit, verbose = FALSE)$lambda,
    explicit
  )
})

test_that("glasso auto covariance keeps its historical default lambda scale", {
  covariance <- matrix(c(.001, .002, .002, .02), nrow = 2)
  legacy.max <- max(abs(covariance - diag(2)))

  automatic <- huge.glasso(
    covariance, nlambda = 2, lambda.min.ratio = .1, verbose = FALSE
  )
  named.automatic <- huge.glasso(
    covariance, nlambda = 2, lambda.min.ratio = .1, verbose = FALSE,
    input.type = c(route = "auto")
  )
  attributed.automatic <- huge.glasso(
    covariance, nlambda = 2, lambda.min.ratio = .1, verbose = FALSE,
    input.type = structure("auto", class = "AsIs")
  )
  explicit <- huge.glasso(
    covariance, nlambda = 2, lambda.min.ratio = .1, verbose = FALSE,
    input.type = "covariance"
  )

  expect_equal(automatic$lambda, c(legacy.max, .1 * legacy.max),
               tolerance = 1e-15)
  expect_identical(named.automatic$lambda, automatic$lambda)
  expect_identical(attributed.automatic$lambda, automatic$lambda)
  expect_equal(explicit$lambda, c(.002, .0002), tolerance = 1e-15)
})

test_that("glasso raw-data default lambda remains correlation based", {
  set.seed(171)
  x <- matrix(rnorm(70 * 8), nrow = 70)
  correlation <- cor(x)
  lambda.max <- max(abs(correlation[upper.tri(correlation)]))
  expected <- exp(seq(
    log(lambda.max), log(.2 * lambda.max), length.out = 4
  ))

  fit <- huge.glasso(
    x, nlambda = 4, lambda.min.ratio = .2, verbose = FALSE
  )
  expect_equal(fit$lambda, expected, tolerance = 1e-12)
})

test_that("glasso refines symmetric precision and rejects non-SPD paths", {
  rho <- 0.9999
  covariance <- matrix(rho, 3, 3)
  diag(covariance) <- 1

  for (cov.output in c(FALSE, TRUE)) {
    expect_error(
      huge.glasso(
        covariance, lambda = c(0.1, 0.0002),
        cov.output = cov.output, verbose = FALSE
      ),
      "not positive definite"
    )
  }

  fit <- huge.glasso(
    covariance, lambda = 0.01, cov.output = TRUE, verbose = FALSE
  )
  fit.no.cov <- huge.glasso(
    covariance, lambda = 0.01, cov.output = FALSE, verbose = FALSE
  )
  precision <- fit$icov[[1]]
  expect_identical(precision, t(precision))
  expect_equal(fit.no.cov$icov[[1]], precision, tolerance = 1e-12)
  expect_gt(determinant(precision, logarithm = TRUE)$sign, 0)
  expect_true(
    max(rowSums(abs(fit$cov[[1]] %*% precision - diag(3)))) <= 1e-2
  )
})

test_that("glasso returns a coherent pair for ill-conditioned Toeplitz input", {
  d <- 20
  rho <- 0.99
  covariance <- outer(
    seq_len(d), seq_len(d), function(i, j) rho^abs(i - j)
  )
  fit <- huge.glasso(
    covariance, lambda = 0.001, cov.output = TRUE, verbose = FALSE
  )

  precision <- fit$icov[[1]]
  estimated.covariance <- fit$cov[[1]]
  expect_identical(precision, t(precision))
  expect_identical(estimated.covariance, t(estimated.covariance))
  expect_gt(determinant(precision, logarithm = TRUE)$sign, 0)
  expect_gt(determinant(estimated.covariance, logarithm = TRUE)$sign, 0)
  expect_lte(
    max(rowSums(abs(estimated.covariance %*% precision - diag(d)))),
    1e-2
  )
})

test_that("glasso refines a finite iteration-limit candidate", {
  covariance <- matrix(c(
    1.0000000000000002, .18009786022575919, -.10400558762095684,
    .76774091516199638, .18742182119843603,
    .18009786022575919, 1, .83834121674080364, .48252423066148048,
    -.066223216563695037,
    -.10400558762095684, .83834121674080364, 1, .097388083759665525,
    -.56897959252915542,
    .76774091516199638, .48252423066148048, .097388083759665525,
    1.0000000000000002, .44589188604391788,
    .18742182119843603, -.066223216563695037, -.56897959252915542,
    .44589188604391788, .99999999999999989
  ), nrow = 5, byrow = TRUE)

  expect_warning(
    fit <- huge.glasso(
      covariance, lambda = .00083834121674080362,
      cov.output = TRUE, verbose = FALSE,
      input.type = "covariance"
    ),
    "iteration limit"
  )
  expect_gt(determinant(fit$icov[[1]], logarithm = TRUE)$sign, 0)
  expect_lte(
    max(rowSums(abs(fit$cov[[1]] %*% fit$icov[[1]] - diag(5)))),
    1e-2
  )
})

test_that("glasso works across graph types", {
  set.seed(18)
  for (g in c("hub", "band", "cluster", "random")) {
    L <- huge.generator(n = 60, d = 20, graph = g, verbose = FALSE)
    fit <- huge(L$data, method = "glasso", verbose = FALSE)
    expect_true(all(diff(fit$sparsity) >= -1e-10),
                info = paste("non-monotone sparsity for graph =", g))
  }
})

#-----------------------------------------------------------------------#
# Package: High-dimensional Undirected Graph Estimation                 #
# huge.generator(): Data generator                                      #
#-----------------------------------------------------------------------#

#' Data generator
#'
#' Implements the data generation from multivariate normal distributions with different graph structures, including \code{"random"}, \code{"hub"}, \code{"cluster"}, \code{"band"} and \code{"scale-free"}.
#'
#' @param n The number of observations (sample size), as an integer of at least \code{2}. The default value is \code{200}.
#' @param d The positive integer number of variables (dimension). The default value is \code{50}.
#' @param graph The graph structure with 5 options: \code{"random"}, \code{"hub"}, \code{"cluster"}, \code{"band"} and \code{"scale-free"}.
#' @param v A finite positive number used for the off-diagonal elements of the precision matrix, controlling the magnitude of partial correlations with \code{u}. The default value is \code{0.3}.
#' @param u A finite positive number added to the diagonal elements of the precision matrix to control the magnitude of partial correlations. The default value is \code{0.1}.
#' @param g A finite positive integer. For \code{"cluster"} or \code{"hub"} graph, \code{g} is the number of clusters or hubs; values above \code{d} are treated as \code{d}. The default is about \code{d/20} if \code{d >= 40} and \code{min(2, d)} otherwise. For \code{"band"} graph, \code{g} is the bandwidth and defaults to \code{1}; values above \code{d - 1} add no edges. Not applicable to \code{"random"} or \code{"scale-free"} graph.
#' @param prob For \code{"random"} or \code{"cluster"} graph, a finite number in \code{[0, 1]} giving the probability that a pair of nodes has an edge. The default is \code{min(1, 3/d)} for \code{"random"}. For \code{"cluster"}, the default is \code{min(1, 6*g/d)} if \code{d/g <= 30} and \code{0.3} otherwise. Not applicable to \code{"hub"}, \code{"band"}, or \code{"scale-free"} graph.
#' @param vis A single non-missing logical. If \code{TRUE}, visualize the adjacency matrix, graph pattern, covariance matrix, and empirical correlation matrix. The default is \code{FALSE}.
#' @param verbose A single non-missing logical. If \code{FALSE}, tracing output is disabled. The default is \code{TRUE}.
#' @details
#' Given the adjacency matrix \code{theta}, the graph patterns are generated as below:\cr\cr
#' (I) \code{"random"}: Each pair of off-diagonal elements are randomly set \code{theta[i,j]=theta[j,i]=1} for \code{i!=j} with probability \code{prob}, and \code{0} otherwise. It results in about \code{d*(d-1)*prob/2} edges in the graph.\cr\cr
#' (II)\code{"hub"}:The row/columns are evenly partitioned into \code{g} disjoint groups. Each group is associated with a "center" row \code{i} in that group. Each pair of off-diagonal elements are set \code{theta[i,j]=theta[j,i]=1} for \code{i!=j} if \code{j} also belongs to the same group as \code{i} and \code{0} otherwise. It results in \code{d - g} edges in the graph.\cr\cr
#' (III)\code{"cluster"}:The row/columns are evenly partitioned into \code{g} disjoint groups. Each pair of off-diagonal elements are set \code{theta[i,j]=theta[j,i]=1} for \code{i!=j} with the probability \code{prob}if both \code{i} and \code{j} belong to the same group, and \code{0} otherwise. It results in about \code{g*(d/g)*(d/g-1)*prob/2} edges in the graph.\cr\cr
#' (IV)\code{"band"}: Let \code{b=min(g,d-1)}. The off-diagonal elements are set to be \code{theta[i,j]=1} if \code{1<=|i-j|<=b} and \code{0} otherwise. It results in \code{(2d-1-b)*b/2} edges in the graph.\cr\cr
#' (V) \code{"scale-free"}: The graph is generated using B-A algorithm. The initial graph has two connected nodes and each new node is connected to only one node in the existing graph with the probability proportional to the degree of the each node in the existing graph. It results in \code{d-1} edges for \code{d>=2}; for \code{d=1}, the graph is empty.
#'
#' The adjacency matrix \code{theta} has all diagonal elements equal to \code{0}. To obtain a positive definite precision matrix, the smallest eigenvalue of \code{theta*v} (denoted by \code{e}) is computed. Then we set the precision matrix equal to \code{theta*v+(|e|+0.1+u)I}. The covariance matrix is then computed to generate multivariate normal data. Every graph type returns an empty graph with zero sparsity when \code{d=1}.
#' @return
#' An object with S3 class "sim" is returned:
#' \item{data}{
#'   The \code{n} by \code{d} matrix for the generated data
#' }
#' \item{sigma}{
#'   The covariance matrix for the generated data
#' }
#' \item{omega}{
#'   The precision matrix for the generated data
#' }
#' \item{sigmahat}{
#'   The empirical correlation matrix for the generated data
#' }
#' \item{theta}{
#'   The adjacency matrix of true graph structure (in sparse matrix representation) for the generated data
#' }
#' \item{sparsity}{
#'   The proportion of possible off-diagonal entries present in the graph; zero when \code{d=1}
#' }
#' @seealso \code{\link{huge}} and \code{\link{huge-package}}
#' @examples
#' ## band graph with bandwidth 3
#' L = huge.generator(graph = "band", g = 3)
#' plot(L)
#'
#' ## random sparse graph
#' L = huge.generator(vis = TRUE)
#'
#' ## random dense graph
#' L = huge.generator(prob = 0.5, vis = TRUE)
#'
#' ## hub graph with 6 hubs
#' L = huge.generator(graph = "hub", g = 6, vis = TRUE)
#'
#' ## hub graph with 8 clusters
#' L = huge.generator(graph = "cluster", g = 8, vis = TRUE)
#'
#' ## scale-free graphs
#' L = huge.generator(graph="scale-free", vis = TRUE)
#' @export
huge.generator = function(n = 200, d = 50, graph = "random", v = NULL, u = NULL, g = NULL, prob = NULL, vis = FALSE, verbose = TRUE){
  if(length(n) == 1 && is.numeric(n) && is.finite(n) &&
     n > .Machine$integer.max)
    stop("n exceeds R's supported integer range.")
  n = .huge_validate_positive_integer(n, "n")
  if(n < 2)
    stop("n must be at least 2 so empirical correlation is defined.")
  if(length(d) == 1 && is.numeric(d) && is.finite(d) &&
     d > .Machine$integer.max)
    stop("d exceeds R's supported integer range.")
  d = .huge_validate_positive_integer(d, "d")
  graph.types = c("random", "hub", "cluster", "band", "scale-free")
  if(length(graph) != 1 || !is.character(graph) || is.na(graph) ||
     !(graph %in% graph.types))
    stop(paste(
      "graph must be exactly one of",
      "\"random\", \"hub\", \"cluster\", \"band\", or \"scale-free\"."
    ))

  .validate_flag = function(value, name){
    if(length(value) != 1 || !is.logical(value) || is.na(value))
      stop(sprintf("%s must be TRUE or FALSE.", name))
    value
  }
  vis = .validate_flag(vis, "vis")
  verbose = .validate_flag(verbose, "verbose")

  .validate_positive_number = function(value, name){
    if(length(value) != 1 || !is.numeric(value) || !is.finite(value) ||
       value <= 0)
      stop(sprintf("%s must be a finite positive number.", name))
    as.numeric(value)
  }
  v = .validate_positive_number(if(is.null(v)) 0.3 else v, "v")
  u = .validate_positive_number(if(is.null(u)) 0.1 else u, "u")

  grouped.graph = graph == "hub" || graph == "cluster"
  if(grouped.graph || graph == "band"){
    if(is.null(g)){
      g = 1L
      if(grouped.graph){
        if(d > 40)  g = ceiling(d/20)
        if(d <= 40) g = min(2, d)
      }
    } else {
      if(length(g) == 1 && is.numeric(g) && is.finite(g) &&
         g > .Machine$integer.max)
        stop("g exceeds R's supported integer range.")
      g = .huge_validate_positive_integer(g, "g")
    }
    if(grouped.graph) g = min(g, d)
  } else {
    g = 1L
  }

  .validate_probability = function(value){
    if(length(value) != 1 || !is.numeric(value) || !is.finite(value) ||
       value < 0 || value > 1)
      stop("prob must be a finite number in [0, 1].")
    as.numeric(value)
  }
  .transform_prob = function(p) sqrt(p/2)*(p<0.5)+(1-sqrt(0.5-0.5*p))*(p>=0.5)

  if(graph == "random"){
    prob = if(is.null(prob)) min(1, 3/d) else .validate_probability(prob)
    prob = .transform_prob(prob)
  }

  if(graph == "cluster"){
    if(is.null(prob)){
      if(d/g > 30)  prob = 0.3
      if(d/g <= 30) prob = min(1, 6*g/d)
    } else {
      prob = .validate_probability(prob)
    }
    prob = .transform_prob(prob)
  }

  if(verbose) cat("Generating data from the multivariate normal distribution with the", graph,"graph structure....")

  # partition variables into groups only for graph types that use them
  if(grouped.graph){
    g.large = d%%g
    g.small = g - g.large
    n.small = floor(d/g)
    n.large = n.small+1
    g.list = c(rep(n.small,g.small),rep(n.large,g.large))
    g.ind = rep(seq_len(g),g.list)
  }

  # build the graph structure
  theta = matrix(0,d,d);
  if(graph == "band"){
    for(i in seq_len(min(g, d - 1L))){
      index = seq_len(d - i)
      theta[cbind(index, index + i)] = 1
      theta[cbind(index + i, index)] = 1
    }
  }
  if(graph == "cluster"){
    for(i in 1:g){
       tmp = which(g.ind==i)
       tmp2 = matrix(runif(length(tmp)^2,0,0.5),length(tmp),length(tmp))
       tmp2 = tmp2 + t(tmp2)
       theta[tmp,tmp][tmp2<prob] = 1
    }
  }
  if(graph == "hub"){
    for(i in 1:g){
       tmp = which(g.ind==i)
       theta[tmp[1],tmp] = 1
       theta[tmp,tmp[1]] = 1
    }
  }
  if(graph == "random"){
    tmp = matrix(runif(d^2,0,0.5),d,d)
    tmp = tmp + t(tmp)
    theta[tmp < prob] = 1
  }

  if(graph == "scale-free" && d > 1L){
  out = .Call("_huge_SFGen", 2, d, PACKAGE= "huge")
  theta = matrix(as.numeric(out$G),d,d)
  }
  diag(theta) = 0
  omega = theta*v

  # make omega positive definite and standardized
  diag(omega) = abs(min(eigen(omega, symmetric = TRUE, only.values = TRUE)$values)) + 0.1 + u
  # omega is symmetric positive definite: invert via Cholesky (chol2inv),
  # ~2x faster than solve() and symmetric by construction
  sigma = cov2cor(chol2inv(chol(omega)))
  omega = chol2inv(chol(sigma))

  # generate multivariate normal data
  x = mvrnorm(n,rep(0,d),sigma)
  sigmahat = .huge_fast_cor(x)

  # graph and covariance visulization
  if(vis == TRUE){
  old.par = .huge_graphics_state()
  on.exit(.huge_restore_graphics_state(old.par), add = TRUE)
  par(mfrow = c(2, 2), pty = "s", omi=c(0.3,0.3,0.3,0.3), mai = c(0.3,0.3,0.3,0.3))
  image(theta, col = gray.colors(256),  main = "Adjacency Matrix")
  image(sigma, col = gray.colors(256), main = "Covariance Matrix")
  g = graph_from_adjacency_matrix(theta, mode="undirected", diag=FALSE)
  layout.grid = layout_with_fr(g)
  plot(g, layout=layout.grid, edge.color='gray50',vertex.color="red", vertex.size=3, vertex.label=NA,main = "Graph Pattern")
  image(sigmahat, col = gray.colors(256), main = "Empirical Correlation Matrix")
  }
  if(verbose) cat("done.\n")

  sparsity = if(d == 1L) 0 else sum(theta)/(d*(d-1))
  sim = list(data = x, sigma = sigma, sigmahat = sigmahat, omega = omega, theta = Matrix(theta,sparse = TRUE), sparsity=sparsity, graph.type=graph)
  class(sim) = "sim"
  return(sim)
}

#' Print function for S3 class "sim"
#'
#' Print the information about the sample size, the dimension, the pattern and sparsity of the true graph structure.
#'
#' @param x An object with S3 class \code{"sim"}.
#' @param \dots System reserved (No specific usage)
#' @seealso \code{\link{huge.generator}}
#' @export
print.sim = function(x, ...){
  cat("Simulated data generated by huge.generator()\n")
  cat("Sample size: n =", nrow(x$data), "\n")
  cat("Dimension: d =", ncol(x$data), "\n")
    cat("Graph type = ", x$graph.type, "\n")
    d = ncol(x$data)
    sparsity = if(d == 1L) 0 else sum(x$theta)/d/(d-1)
    cat("Sparsity level:", sparsity,"\n")
}

#' Plot function for S3 class "sim"
#'
#' Visualize the covariance matrix, the empirical correlation matrix, the adjacency matrix and the graph pattern of the true graph structure.
#'
#' @param x An object with S3 class \code{"sim"}
#' @param \dots System reserved (No specific usage)
#' @seealso \code{\link{huge.generator}} and \code{\link{huge}}
#' @export
plot.sim = function(x, ...){
     old.par = .huge_graphics_state()
     on.exit(.huge_restore_graphics_state(old.par), add = TRUE)
     par(mfrow = c(2, 2), pty = "s", omi=c(0.3,0.3,0.3,0.3), mai = c(0.3,0.3,0.3,0.3))
     image(as.matrix(x$theta), col = gray.colors(256),  main = "Adjacency Matrix")
  image(x$sigma, col = gray.colors(256), main = "Covariance Matrix")
  g = graph_from_adjacency_matrix(x$theta, mode="undirected", diag=FALSE)
  layout.grid = layout_with_fr(g)

  plot(g, layout=layout.grid, edge.color='gray50',vertex.color="red", vertex.size=3, vertex.label=NA,main = "Graph Pattern")
  image(x$sigmahat, col = gray.colors(256), main = "Empirical Correlation Matrix")
}

test_that("huge.generator produces valid output for all graph types", {
  set.seed(1)
  for (g in c("hub", "band", "cluster", "random", "scale-free")) {
    L <- huge.generator(n = 60, d = 20, graph = g, verbose = FALSE)
    expect_s3_class(L, "sim")
    expect_equal(dim(L$data), c(60, 20))
    expect_equal(dim(L$sigma), c(20, 20))
    expect_equal(dim(L$omega), c(20, 20))
    expect_equal(L$sigmahat, cor(L$data), tolerance = 1e-12)
    expect_equal(diag(L$sigmahat), rep(1, 20), tolerance = 0)

    # sigma should be positive definite
    eig <- eigen(L$sigma, symmetric = TRUE, only.values = TRUE)$values
    expect_true(all(eig > 0))

    # theta should be symmetric binary
    theta <- as.matrix(L$theta)
    expect_equal(theta, t(theta))
    expect_true(all(theta %in% c(0, 1)))
    expect_true(all(diag(theta) == 0))

    # sparsity should be in [0, 1]
    expect_true(L$sparsity >= 0 && L$sparsity <= 1)
  }
})

test_that("huge.generator respects dimension parameters", {
  set.seed(2)
  L <- huge.generator(n = 100, d = 30, graph = "hub", verbose = FALSE)
  expect_equal(nrow(L$data), 100)
  expect_equal(ncol(L$data), 30)
  expect_equal(nrow(L$sigma), 30)
})

for (graph in c("hub", "band", "cluster", "random", "scale-free")) {
  test_that(paste("huge.generator supports d = 1 for", graph), {
    set.seed(222)
    L <- huge.generator(n = 8, d = 1, graph = graph, verbose = FALSE)

    expect_equal(dim(L$data), c(8, 1))
    expect_equal(dim(L$sigma), c(1, 1))
    expect_equal(dim(L$omega), c(1, 1))
    expect_equal(dim(L$sigmahat), c(1, 1))
    expect_equal(as.matrix(L$theta), matrix(0, 1, 1))
    expect_equal(L$sigma, matrix(1, 1, 1))
    expect_equal(L$omega, matrix(1, 1, 1))
    expect_equal(L$sigmahat, matrix(1, 1, 1))
    expect_identical(L$sparsity, 0)

    printed <- paste(capture.output(print(L)), collapse = "\n")
    expect_match(printed, "Dimension: d = 1", fixed = TRUE)
    expect_match(printed, "Sparsity level: 0", fixed = TRUE)
    expect_false(grepl("NaN", printed, fixed = TRUE))
  })
}

test_that("band graphs support the maximal meaningful bandwidth", {
  for (d in c(2, 5)) {
    set.seed(223)
    L <- huge.generator(
      n = 8, d = d, graph = "band", g = d - 1,
      verbose = FALSE
    )

    expect_equal(as.matrix(L$theta), 1 - diag(d))
    expect_identical(L$sparsity, 1)
  }
})

test_that("huge.generator rejects invalid active parameters before side effects", {
  expect_clean_error <- function(args, pattern) {
    set.seed(224)
    seed.before <- .Random.seed
    warnings <- character()
    printed <- capture.output(
      result <- withCallingHandlers(
        tryCatch(
          do.call(
            huge.generator,
            c(list(n = 8, d = 5, verbose = TRUE), args)
          ),
          error = identity
        ),
        warning = function(w) {
          warnings <<- c(warnings, conditionMessage(w))
          invokeRestart("muffleWarning")
        }
      )
    )

    expect_s3_class(result, "error")
    if (inherits(result, "error")) {
      expect_match(conditionMessage(result), pattern)
    }
    expect_identical(printed, character())
    expect_identical(warnings, character())
    expect_true(identical(.Random.seed, seed.before))
  }

  for (v in list(0, -1, NA_real_, Inf, TRUE, c(0.2, 0.3), "0.3")) {
    expect_clean_error(list(graph = "random", v = v), "v")
  }
  for (u in list(0, -1, NA_real_, Inf, TRUE, c(0.1, 0.2), "0.1")) {
    expect_clean_error(list(graph = "band", u = u), "u")
  }
  for (g in list(0, -1, 1.5, 3e9, NA_real_, Inf, TRUE, c(1, 2), "2")) {
    for (graph in c("hub", "cluster", "band")) {
      expect_clean_error(list(graph = graph, g = g), "g")
    }
  }
  for (prob in list(-0.1, 1.1, NA_real_, Inf, TRUE, c(0.2, 0.3), "0.3")) {
    for (graph in c("random", "cluster")) {
      expect_clean_error(list(graph = graph, prob = prob), "prob")
    }
  }
})

test_that("huge.generator rejects invalid flags before side effects", {
  expect_clean_flag_error <- function(name, value) {
    set.seed(229)
    seed.before <- .Random.seed
    args <- list(
      n = 8, d = 5, graph = "random",
      vis = FALSE, verbose = FALSE
    )
    args[name] <- list(value)
    warnings <- character()
    printed <- capture.output(
      result <- withCallingHandlers(
        tryCatch(do.call(huge.generator, args), error = identity),
        warning = function(w) {
          warnings <<- c(warnings, conditionMessage(w))
          invokeRestart("muffleWarning")
        }
      )
    )

    expect_s3_class(result, "error")
    if (inherits(result, "error")) {
      expect_match(conditionMessage(result), name)
    }
    expect_identical(printed, character())
    expect_identical(warnings, character())
    expect_true(identical(.Random.seed, seed.before))
  }

  invalid <- list(
    NULL, NA, logical(), c(TRUE, FALSE), 0, 2, "FALSE"
  )
  for (name in c("vis", "verbose")) {
    for (value in invalid) {
      expect_clean_flag_error(name, value)
    }
  }
})

test_that("huge.generator ignores graph-inactive parameters", {
  expect_same_generation <- function(graph, name, value) {
    set.seed(225)
    expected <- huge.generator(
      n = 8, d = 5, graph = graph,
      verbose = FALSE
    )
    expected.seed <- .Random.seed

    set.seed(225)
    args <- list(n = 8, d = 5, graph = graph, verbose = FALSE)
    args[[name]] <- value
    actual <- do.call(huge.generator, args)

    expect_identical(actual, expected)
    expect_true(identical(.Random.seed, expected.seed))
  }

  for (graph in c("random", "scale-free")) {
    expect_same_generation(graph, "g", "inactive")
  }
  for (graph in c("hub", "band", "scale-free")) {
    expect_same_generation(graph, "prob", "inactive")
  }
})

test_that("huge.generator accepts parameter boundaries without changing semantics", {
  for (graph in c("random", "cluster")) {
    for (prob in c(0, 1)) {
      set.seed(226)
      L <- huge.generator(
        n = 8, d = 5, graph = graph, prob = prob,
        verbose = FALSE
      )
      expect_true(all(is.finite(L$data)))
    }
  }

  for (graph in c("hub", "cluster")) {
    set.seed(227)
    oversized <- huge.generator(
      n = 8, d = 5, graph = graph, g = 7,
      verbose = FALSE
    )
    oversized.seed <- .Random.seed
    set.seed(227)
    capped <- huge.generator(
      n = 8, d = 5, graph = graph, g = 5,
      verbose = FALSE
    )
    expect_identical(oversized, capped)
    expect_true(identical(oversized.seed, .Random.seed))
  }

  set.seed(228)
  oversized <- huge.generator(
    n = 8, d = 5, graph = "band", g = 7,
    verbose = FALSE
  )
  oversized.seed <- .Random.seed
  set.seed(228)
  capped <- huge.generator(
    n = 8, d = 5, graph = "band", g = 4,
    verbose = FALSE
  )
  expect_identical(oversized, capped)
  expect_true(identical(oversized.seed, .Random.seed))
})

test_that("huge.generator visualization restores graphics parameters", {
  plot.file <- tempfile(fileext = ".pdf")
  grDevices::pdf(plot.file, width = 6, height = 6)
  on.exit({
    grDevices::dev.off()
    unlink(plot.file)
  }, add = TRUE)
  graphics::par(
    mfrow = c(2, 3), pty = "m",
    omi = rep(0.05, 4), mai = rep(0.2, 4)
  )
  graphics::par(
    mfg = c(1, 2), cex = 1.37, mex = 0.91,
    usr = c(-2, 3, -4, 5)
  )
  graphics::par(new = FALSE)
  parameters <- c(
    "mfrow", "mfg", "pty", "omi", "mai", "cex", "mex", "usr", "new"
  )
  before <- graphics::par(parameters)

  set.seed(230)
  invisible(huge.generator(
    n = 8, d = 5, graph = "band",
    vis = TRUE, verbose = FALSE
  ))

  after <- graphics::par(parameters)
  expect_identical(after, before)
})

test_that("huge.generator rejects invalid dimensions before side effects", {
  expect_clean_error <- function(args, pattern) {
    set.seed(73)
    seed.before <- .Random.seed
    printed <- capture.output(
      result <- tryCatch(
        do.call(
          huge.generator,
          c(args, list(graph = "random", verbose = TRUE))
        ),
        error = identity
      )
    )

    expect_s3_class(result, "error")
    expect_match(conditionMessage(result), pattern)
    expect_identical(printed, character())
    expect_identical(.Random.seed, seed.before)
  }

  for (n in list(
    1, 0, -1, 1.5, 3e9, NA_real_, Inf, TRUE, c(2, 3), "2"
  )) {
    expect_clean_error(list(n = n, d = 4), "n")
  }
  for (d in list(
    0, -1, 1.5, 3e9, NA_real_, Inf, TRUE, c(2, 3), "2"
  )) {
    expect_clean_error(list(n = 8, d = d), "d")
  }
})

test_that("huge.generator rejects unknown graph types before side effects", {
  for (graph in list("unknown", c("random", "hub"), NA_character_, 1)) {
    set.seed(74)
    seed.before <- .Random.seed
    printed <- capture.output(
      result <- tryCatch(
        huge.generator(
          n = 8, d = 4, graph = graph,
          verbose = TRUE
        ),
        error = identity
      )
    )

    expect_s3_class(result, "error")
    expect_match(conditionMessage(result), "graph.*one of")
    expect_identical(printed, character())
    expect_identical(.Random.seed, seed.before)
  }
})

#-------------------------------------------------------------------------#
# Package: High-dimensional Undirected Graph Estimation                   #
# huge.select(): Model selection using:                                   #
#                1.rotation information criterion (ric)                   #
#                2.stability approach to regularization selection (stars) #
#                3.extended Bayesian information criterion (ebic)        #
#-------------------------------------------------------------------------#

#' Model selection for high-dimensional undirected graph estimation
#'
#' Implements the regularization parameter selection for high dimensional undirected graph estimation. The optional approaches are rotation information criterion (ric), stability approach to regularization selection (stars) and extended Bayesian information criterion (ebic).
#'
#' Stability approach to regularization selection (stars) selects the optimal graph by variability of subsamplings and tends to overselect edges in Gaussian graphical models. It is available for \code{"mb"}, \code{"ct"}, and \code{"glasso"}. TIGER can certify different lambda-path prefixes on different subsamples, so TIGER with stars is rejected until a common certified-prefix protocol is available; use \code{"ric"} for TIGER. Besides selecting the regularization parameters, stars can also provide an additional estimated graph by merging the corresponding subsampled graphs using the frequency counts. The subsampling procedure in stars may NOT be very efficient, we also provide the recent developed highly efficient, rotation information criterion approach (ric). Instead of tuning over a grid by cross-validation or subsampling, we directly estimate the optimal regularization parameter based on random Rotations. However, ric usually has very good empirical performances but suffers from underselections sometimes. Therefore, we suggest if user are sensitive of false negative rates, they should either consider increasing \code{rep.num} or applying the stars to model selection where supported. Extended Bayesian information criterion (ebic) is another competitive approach, but the \code{ebic.gamma} can only be tuned by experience.
#'
#' For \code{criterion = "stars"}, \code{est$lambda} must be non-increasing; tied values are allowed.
#'
#' @param est An object with S3 class \code{"huge"}.
#' @param criterion Model selection criterion. \code{"ric"} is available for all four estimation methods, \code{"stars"} for \code{"mb"}, \code{"ct"}, and \code{"glasso"}, and \code{"ebic"} only for \code{"glasso"}. Defaults are \code{"ric"} for \code{"mb"} and \code{"tiger"}, \code{"stars"} for \code{"ct"}, and \code{"ebic"} for \code{"glasso"}.
#' @param ebic.gamma The tuning parameter for ebic. The default value is 0.5. Only applicable when \code{est$method = "glasso"} and \code{criterion = "ebic"}.
#' @param stars.thresh The variability threshold in stars. The default value is \code{0.1}. An alternative value is \code{0.05}. Only applicable when \code{criterion = "stars"}.
#' @param stars.subsample.ratio The subsampling ratio. The default value is \code{10*sqrt(n)/n} when \code{n>144} and \code{0.8} when \code{n<=144}, where \code{n} is the sample size. Only applicable when \code{criterion = "stars"}.
#' @param rep.num The number of subsamplings when \code{criterion = "stars"} or rotations when \code{criterion = "ric"}. The default value is \code{20}. NOT applicable when \code{criterion = "ebic"}.
#' @param verbose If \code{verbose = FALSE}, tracing information printing is disabled. The default value is \code{TRUE}.
#' @param num.cores The number of CPU cores used to fit the stars subsamplings in parallel via \code{parallel::mclapply}. The default value is \code{1} (serial). At most two forked workers are used, and this argument is ignored on Windows. Each forked worker limits huge's native OpenMP code to one thread, but an external threaded BLAS may still create additional threads; \code{num.cores = 1} is the portable choice when a bounded thread budget or fork safety matters. Results are identical to the serial path for the same random seed. Only applicable when \code{criterion = "stars"}.
#' @return
#' An object with S3 class "select" is returned:
#'   \item{refit}{
#'     The optimal graph selected from the graph path
#'   }
#' \item{opt.icov}{
#'   The optimal precision matrix from the path only applicable when \code{method = "glasso"}
#' }
#' \item{opt.cov}{
#'   The optimal covariance matrix from the path only applicable when \code{method = "glasso"} and \code{est$cov} is available.
#' }
#' \item{merge}{
#'   The graph path estimated by merging the subsampling paths. Only applicable when the input \code{criterion = "stars"}.
#' }
#' \item{variability}{
#'   The variability along the subsampling paths. Only applicable when the input \code{criterion = "stars"}.
#' }
#' \item{ebic.score}{
#'   Extended BIC scores for regularization parameter selection. Only applicable when \code{criterion = "ebic"}.
#' }
#' \item{opt.index}{
#'   The index of the selected regularization parameter. NOT applicable when the input \code{criterion = "ric"}
#' }
#' \item{opt.lambda}{
#'   The selected regularization/thresholding parameter.
#' }
#' \item{opt.sparsity}{
#'   The sparsity level of \code{"refit"}.
#' }
#'
#' and anything else included in the input \code{est}
#'
#' @note The model selection is NOT available when the data input is the sample covariance matrix.
#' @seealso \code{\link{huge}} and \code{\link{huge-package}}.
#' @examples
#' #generate data
#' L = huge.generator(d = 20, graph="hub")
#' out.mb = huge(L$data)
#' out.ct = huge(L$data, method = "ct")
#' out.glasso = huge(L$data, method = "glasso")
#'
#' #model selection using ric
#' out.select = huge.select(out.mb)
#' plot(out.select)
#'
#' #model selection using stars
#' #out.select = huge.select(out.ct, criterion = "stars", stars.thresh = 0.05,rep.num=10)
#' #plot(out.select)
#'
#' #model selection using ebic
#' out.select = huge.select(out.glasso,criterion = "ebic")
#' plot(out.select)
#' @export
huge.select = function(est, criterion = NULL, ebic.gamma = 0.5, stars.thresh = 0.1, stars.subsample.ratio = NULL, rep.num = 20, verbose = TRUE, num.cores = 1){

  if(!is.list(est) || !inherits(est, "huge"))
    stop("est must be an object returned by huge().")

  required = c("method", "cov.input", "data", "lambda", "path", "sparsity")
  missing = required[vapply(required, function(name)
    is.null(est[[name]]), logical(1))]
  if(length(missing) > 0)
    stop(sprintf(
      "est is incomplete; missing field(s): %s.",
      paste(missing, collapse = ", ")
    ))

  if(length(est$method) != 1 || !is.character(est$method) ||
     !(est$method %in% c("mb", "ct", "glasso", "tiger")))
    stop("est has an invalid method field.")
  if(length(est$cov.input) != 1 || !is.logical(est$cov.input) ||
     is.na(est$cov.input))
    stop("est has an invalid cov.input field.")
  if(!is.matrix(est$data) || !is.numeric(est$data) ||
     nrow(est$data) < 1 || ncol(est$data) < 1)
    stop("est has an invalid data matrix.")
  if(!is.numeric(est$lambda) || length(est$lambda) < 1 ||
     any(!is.finite(est$lambda)))
    stop("est has an invalid lambda path.")

  nlambda = length(est$lambda)
  d = ncol(est$data)
  if(!is.list(est$path) || length(est$path) != nlambda ||
     any(!vapply(est$path, function(path)
       length(dim(path)) == 2 && all(dim(path) == c(d, d)), logical(1))))
    stop("est has an invalid graph path.")
  if(!is.numeric(est$sparsity) || length(est$sparsity) != nlambda ||
     any(!is.finite(est$sparsity)))
    stop("est has an invalid sparsity path.")
  if(est$method %in% c("mb", "tiger"))
  {
    if(length(est$sym) != 1 || !is.character(est$sym) || is.na(est$sym) ||
       !(est$sym %in% c("or", "and")))
      stop("est is missing valid symmetrization metadata.")
  }
  if(est$method %in% c("mb", "glasso"))
  {
    if(length(est$scr) != 1 || !is.logical(est$scr) || is.na(est$scr))
      stop("est is missing screening metadata.")
  }

  if(est$cov.input)
    stop("Model selection is not available when using the covariance matrix as input.")

  if(!est$cov.input)
  {
    if(is.null(criterion))
      criterion = switch(est$method,
        mb = "ric", ct = "stars", glasso = "ebic", tiger = "ric")
    else if(length(criterion) != 1 || !is.character(criterion) ||
            is.na(criterion) ||
            !(criterion %in% c("ric", "stars", "ebic")))
      stop("criterion must be exactly one of \"ric\", \"stars\", or \"ebic\".")

    if(criterion == "ebic" && est$method != "glasso")
      stop("criterion = \"ebic\" is available only for method = \"glasso\".")
    if(criterion == "stars" && est$method == "tiger")
      stop(paste(
        "TIGER with StARS is unavailable until subsample fits can share a",
        "common certified prefix; use criterion = \"ric\" for TIGER."
      ))
    if(criterion == "stars" && nlambda > 1 &&
       any(diff(est$lambda) > 0))
      stop(paste(
        "StARS requires est$lambda to be non-increasing;",
        "refit with lambda in decreasing order."
      ))

    if(criterion %in% c("ric", "stars"))
      rep.num = .huge_validate_positive_integer(rep.num, "rep.num")
    if(criterion == "stars")
    {
      num.cores = .huge_validate_positive_integer(num.cores, "num.cores")
      stars.thresh = .huge_validate_ratio(stars.thresh, "stars.thresh")
      if(!is.null(stars.subsample.ratio))
        stars.subsample.ratio = .huge_validate_ratio(
          stars.subsample.ratio, "stars.subsample.ratio"
        )
    }
    if(criterion == "ebic")
    {
      if(length(ebic.gamma) != 1 || !is.numeric(ebic.gamma) ||
         !is.finite(ebic.gamma))
        stop("ebic.gamma must be a finite numeric value.")
      ebic.gamma = as.numeric(ebic.gamma)
      if(!is.numeric(est$loglik) || length(est$loglik) != nlambda ||
         any(!is.finite(est$loglik)) ||
         !is.numeric(est$df) || length(est$df) != nlambda ||
         any(!is.finite(est$df)) ||
         !is.list(est$icov) || length(est$icov) != nlambda)
        stop("est is incomplete or invalid for criterion = \"ebic\".")
    }
    if(criterion == "stars" && d < 2)
      stop("StARS requires at least two variables.")
    if(criterion == "stars" && .Platform$OS.type != "windows" &&
       min(num.cores, rep.num, 2L) > 1)
      warning(paste(
        "num.cores > 1 uses at most two forked StARS workers. Each worker",
        "limits huge's native OpenMP code to one thread, but an external",
        "threaded BLAS may still create additional threads. Use",
        "num.cores = 1 for the portable single-worker path."
      ), call. = FALSE)

    n = nrow(est$data)

    # One fitting closure shared by the RIC refit and the stars subsampling:
    # refits est$method on `data` over `lambda` with the original fit's
    # screening/symmetrization settings. (This also fixes a historical
    # inconsistency where the RIC glasso refit dropped `scr` unless
    # cov.output was requested.) Selection is unavailable for covariance
    # input, so force raw-data routing when a square symmetric sample matrix
    # is refitted.
    refit.fn = function(data, lambda, cov.output = FALSE) {
      switch(est$method,
        mb     = huge.mb(data, lambda = lambda, scr = est$scr, idx.mat = est$idx_mat, sym = est$sym, verbose = FALSE, input.type = "data"),
        ct     = huge.ct(data, lambda = lambda, verbose = FALSE, input.type = "data"),
        glasso = huge.glasso(data, lambda = lambda, scr = est$scr, cov.output = cov.output, verbose = FALSE, input.type = "data"),
        tiger  = huge.tiger(data, lambda = lambda, sym = est$sym, verbose = FALSE, input.type = "data"))
    }

    if(criterion == "ric")
    {
      if(verbose)
      {
        cat("Conducting rotation information criterion (ric) selection....")
        flush.console()
      }

      if(n>rep.num){
        nr = rep.num
        r = sample(n,rep.num)
      }
      if(n<=rep.num){
        nr = n
        r = 1:n
      }

      # RIC must see standardized data: the lambda path is defined on the
      # correlation scale, and the rotated inner products are otherwise
      # scale-dependent (multiplying x by c scales opt.lambda by c^2).
      standardized.data = .huge_standardize(est$data)
      est$opt.lambda = .Call(
        "_huge_RIC", standardized.data, d, n, r, nr
      ) * 1.0 / n
      if(verbose){
        cat("done\n")
        flush.console()
      }

      if(verbose)
      {
        cat("Computing the optimal graph....")
        flush.console()
      }

      # A BLAS dot product can leave a tiny residual for mathematically
      # orthogonal columns.  Suppress only values within a conservative,
      # pair-specific dot-product roundoff bound; a fixed tolerance would
      # erase genuine weak correlations.
      denominator = max(n - 1, 1)
      correlation = crossprod(standardized.data) / denominator
      absolute.cross = crossprod(abs(standardized.data)) / denominator
      scaled.eps = n * .Machine$double.eps
      dot.gamma = if(scaled.eps < 1) scaled.eps / (1 - scaled.eps) else Inf
      cor.offdiag = abs(correlation)
      roundoff.bound = dot.gamma * absolute.cross +
        .Machine$double.eps * cor.offdiag
      cor.offdiag[cor.offdiag <= roundoff.bound] = 0
      diag(cor.offdiag) = 0
      if(est$opt.lambda>=max(cor.offdiag))
        est$refit = Matrix(0,d,d)
      else{
        # [[ ]] for exact-name access: est$cov would partial-match cov.input
        # on fits that lack a cov field (all methods except glasso+cov.output).
        refit.lambda = est$opt.lambda
        zero.proxy = est$opt.lambda == 0 && est$method != "ct"
        if(zero.proxy)
          refit.lambda = .Machine$double.xmin

        tmp = if(zero.proxy) {
          tryCatch(
            refit.fn(
              est$data, refit.lambda,
              cov.output = !is.null(est[["cov"]])
            ),
            error = function(error) NULL
          )
        } else {
          refit.fn(
            est$data, refit.lambda,
            cov.output = !is.null(est[["cov"]])
          )
        }

        if(is.null(tmp)) {
          nearest = which.min(abs(est$lambda - est$opt.lambda))
          est$refit = est$path[[nearest]]
          if(!is.null(est[["icov"]]))
            est$opt.icov = est[["icov"]][[nearest]]
          if(!is.null(est[["cov"]]))
            est$opt.cov = est[["cov"]][[nearest]]
          warning(sprintf(
            paste(
              "RIC selected lambda = 0, but the method could not certify",
              "the smallest positive proxy; the original fitted path at",
              "lambda %.17g was used."
            ),
            est$lambda[[nearest]]
          ), call. = FALSE)
        } else {
          est$refit = tmp$path[[1]]
          if(!is.null(tmp[["icov"]]))
            est$opt.icov = tmp[["icov"]][[1]]
          if(!is.null(est[["cov"]]) && !is.null(tmp[["cov"]]))
            est$opt.cov = tmp[["cov"]][[1]]
        }
      }
      est$opt.sparsity=if(d <= 1) 0 else sum(est$refit)/d/(d-1)

      if(verbose){
        cat("done\n")
        flush.console()
      }
    }

    if(criterion == "ebic"&&est$method == "glasso")
    {
      if(verbose)
      {
        cat("Conducting extended Bayesian information criterion (ebic) selection....")
        flush.console()
      }
      est$ebic.score = -n*est$loglik + log(n)*est$df + 4*ebic.gamma*log(d)*est$df
      est$opt.index = which.min(est$ebic.score)
      est$refit = est$path[[est$opt.index]]
      est$opt.icov = est$icov[[est$opt.index]]
      if(est$cov.output)
        est$opt.cov = est$cov[[est$opt.index]]
      est$opt.lambda = est$lambda[est$opt.index]
      est$opt.sparsity = est$sparsity[est$opt.index]
      if(verbose)
      {
        cat("done\n")
        flush.console()
      }
    }

    if(criterion == "stars"){
      if(is.null(stars.subsample.ratio))
      {
        if(n>144) stars.subsample.ratio = 10*sqrt(n)/n
        if(n<=144) stars.subsample.ratio = 0.8
      }
      subsample.size = floor(n * stars.subsample.ratio)
      if(subsample.size < 2)
        stop("stars.subsample.ratio must select at least two observations.")

      # Draw all subsample index sets up front: the fitting functions consume
      # no RNG, so this yields the same subsamples as drawing inside the loop
      # (bitwise-identical results), and makes the fits order-independent so
      # they can run in parallel.
      ind.list = lapply(1:rep.num, function(i)
        sample(c(1:n), subsample.size, replace=FALSE))

      subsample.fit = function(ind.sample)
        refit.fn(est$data[ind.sample,], est$lambda)$path

      # This wrapper runs only in fork children. The native R adapter reads
      # the private marker and temporarily limits huge's OpenMP regions to
      # one thread, avoiding nested package-owned parallelism.
      subsample.fit.fork = function(ind.sample) {
        previous = Sys.getenv("HUGE_R_FORK_WORKER", unset = NA_character_)
        on.exit({
          if(is.na(previous))
            Sys.unsetenv("HUGE_R_FORK_WORKER")
          else
            Sys.setenv(HUGE_R_FORK_WORKER = previous)
        }, add = TRUE)
        Sys.setenv(HUGE_R_FORK_WORKER = "1")
        subsample.fit(ind.sample)
      }

      # Fork-based parallelism (not on Windows); never start more workers
      # than there are independent subsample fits or CRAN's two-core limit.
      # num.cores = 1 keeps the serial path with per-replication progress.
      use.cores = min(num.cores, rep.num, 2L)
      if(.Platform$OS.type == "windows") use.cores = 1

      est$merge = vector("list", nlambda)
      for(lambda_idx in 1:nlambda)
        est$merge[[lambda_idx]] = Matrix(0,d,d)

      if(use.cores > 1) {
        if(verbose)
        {
          cat(sprintf("Conducting Subsampling....on %d cores", use.cores), "\r")
          flush.console()
        }
        paths = parallel::mclapply(
          ind.list, subsample.fit.fork, mc.cores = use.cores,
          mc.cleanup = TRUE, mc.allow.recursive = FALSE
        )
        failed = vapply(paths, function(p) inherits(p, "try-error") || is.null(p), logical(1))
        if(any(failed))
          stop("stars subsampling failed in a parallel worker; retry with num.cores = 1")
        for(lambda_idx in 1:nlambda)
          for(rep_idx in 1:rep.num)
            est$merge[[lambda_idx]] = est$merge[[lambda_idx]] + paths[[rep_idx]][[lambda_idx]]
      } else {
        for(rep_idx in 1:rep.num)
        {
          if(verbose)
          {
            mes <- paste(c("Conducting Subsampling....in progress:", floor(100*rep_idx/rep.num), "%"), collapse="")
            cat(mes, "\r")
            flush.console()
          }
          current.path = subsample.fit(ind.list[[rep_idx]])
          for(lambda_idx in 1:nlambda)
            est$merge[[lambda_idx]] = est$merge[[lambda_idx]] + current.path[[lambda_idx]]
          rm(current.path)
        }
      }

      if(verbose)
      {
        mes = "Conducting Subsampling....done.                 "
        cat(mes, "\r")
        cat("\n")
        flush.console()
      }

      est$variability = rep(0,nlambda)
      for(lambda_idx in 1:nlambda){
        est$merge[[lambda_idx]] = est$merge[[lambda_idx]]/rep.num
        # sum(m) - sum(m^2) == sum(m*(1-m)) but avoids densifying (1-m)
        m.tmp = est$merge[[lambda_idx]]
        est$variability[lambda_idx] = 4*(sum(m.tmp) - sum(m.tmp^2))/(d*(d-1))
      }

      stars.cross = which(est$variability >= stars.thresh)
      if(length(stars.cross) == 0)
        est$opt.index = nlambda
      else
        est$opt.index = max(stars.cross[1]-1,1)
      est$refit = est$path[[est$opt.index]]
      est$opt.lambda = est$lambda[est$opt.index]
      est$opt.sparsity = est$sparsity[est$opt.index]
      if(est$method == "glasso")
      {
        est$opt.icov = est$icov[[est$opt.index]]
        # [[ ]] avoids est$cov partial-matching cov.input/cov.output
        if(!is.null(est[["cov"]]))
          est$opt.cov = est[["cov"]][[est$opt.index]]
      }
      if(est$method == "tiger")
        est$opt.icov = est$icov[[est$opt.index]]
    }
    est$criterion = criterion
    class(est) = "select"
    return(est)
  }
}

#-----------------------------------------------------------------------#
# default printing function for class "select"                          #
#-----------------------------------------------------------------------#

#' Print function for S3 class "select"
#'
#' Print the information about the model usage, graph dimension, model selection criterion, sparsity level of the optimal graph.
#'
#' @param x An object with S3 class \code{"select"}.
#' @param \dots System reserved (No specific usage)
#' @seealso \code{\link{huge.select}}
#' @export
print.select = function(x, ...)
{
  if(x$cov.input){
    cat("Model selection is not available when using the covariance matrix as input.")
  }
  if(!x$cov.input)
  {
    if(x$method == "ct")
      cat("Model: graph estimation via correlation thresholding (ct)\n")
    if(x$method == "glasso")
      cat("Model: graphical lasso (glasso)\n")
    if(x$method == "mb")
      cat("Model: Meinshausen & Buhlmann Graph Estimation (mb)\n")
    if(x$method == "tiger")
      cat("Model: tuning-insensitive approach (tiger)\n")

    cat("selection criterion:",x$criterion,"\n")
    if((x$method != "ct")&&x$scr)
      cat("lossy screening: on\n")
    cat("Graph dimension:",ncol(x$data),"\n")
    cat("sparsity level", x$opt.sparsity,"\n")
  }
}

#' Plot function for S3 class "select"
#'
#' Plot the optimal graph by model selection.
#'
#' @param x An object with S3 class \code{"select"}
#' @param \dots System reserved (No specific usage)
#' @seealso \code{\link{huge.select}}
#' @export
plot.select = function(x, ...){
  if(x$cov.input){
    cat("Model selection is not available when using the covariance matrix as input.")
  }
  if(!x$cov.input)
  {
    old.par = .huge_graphics_state()
    on.exit(.huge_restore_graphics_state(old.par), add = TRUE)
    par(mfrow=c(1,2), pty = "s", omi=c(0.3,0.3,0.3,0.3), mai = c(0.3,0.3,0.3,0.3))

      g = graph_from_adjacency_matrix(as.matrix(x$refit), mode="undirected", diag=FALSE)
    layout.grid = layout_with_fr(g)

    plot(g, layout=layout.grid, edge.color='gray50',vertex.color="red", vertex.size=3, vertex.label=NA)
      plot(x$lambda, x$sparsity, log = "x", xlab = "Regularization Parameter", ylab = "Sparsity Level", type = "l",xlim = rev(range(x$lambda)), main = "Solution path sparsity levels")
      lines(x$opt.lambda,x$opt.sparsity,type = "p")
    }
}

#-----------------------------------------------------------------------#
# Package: High-dimensional Undirected Graph Estimation                 #
# huge(): The user interface for huge.mb(), huge.glasso(), huge.ct()    #
#         and huge.tiger()                                              #
#-----------------------------------------------------------------------#

#' High-dimensional undirected graph estimation
#'
#' The main function for high-dimensional undirected graph estimation. Four graph estimation methods, including (1) Meinshausen-Buhlmann graph estimation (\code{mb}) (2) graphical lasso (\code{glasso}) (3) correlation thresholding graph estimation (\code{ct}) and (4) tuning-insensitive graph estimation (\code{tiger}), are available for data analysis.
#'
#' The graph structure is estimated by Meinshausen-Buhlmann graph estimation or the graphical lasso, and both methods can be further accelerated via the lossy screening rule by preselecting the neighborhood of each variable by correlation thresholding. We target on high-dimensional data analysis usually d >> n, and the computation is memory-optimized using the sparse matrix output. We also provide a highly computationally efficient approaches correlation thresholding graph estimation.
#'
#' @param x There are 2 options: (1) \code{x} is an \code{n} by \code{d} data matrix (2) a \code{d} by \code{d} sample covariance matrix. The program automatically identifies the input matrix by checking the symmetry. (\code{n} is the sample size and \code{d} is the dimension).
#' @param lambda An optional numeric scalar or non-empty one-dimensional
#'   numeric path. For \code{method = "mb"}, \code{"glasso"}, or
#'   \code{"tiger"}, values must be finite, strictly
#'   positive, and non-increasing; ties are allowed. For \code{method = "ct"},
#'   values must be finite and non-negative, zero is allowed, and thresholds
#'   are applied in the supplied order. Leave \code{lambda = NULL} to generate
#'   a path from \code{nlambda} and \code{lambda.min.ratio}.
#' @param nlambda The number of regularization/thresholding parameters. The default value is \code{20} for \code{method = "ct"} and \code{10} for \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}.
#' @param lambda.min.ratio If \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}, it is the smallest value for \code{lambda}, as a fraction of the upperbound (\code{MAX}) of the regularization/thresholding parameter which makes all estimates equal to \code{0}. The program can automatically generate \code{lambda} as a sequence of length = \code{nlambda} starting from \code{MAX} to \code{lambda.min.ratio*MAX} in log scale. If \code{method = "ct"}, it is the largest sparsity level for estimated graphs. The program can automatically generate \code{lambda} as a sequence of length = \code{nlambda}, which makes the sparsity level of the graph path increases from \code{0} to \code{lambda.min.ratio} evenly.The default value is \code{0.1} when \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}, and 0.05 when \code{method = "ct"}.
#' @param method Graph estimation methods with 4 options: \code{"mb"}, \code{"ct"}, \code{"glasso"} and \code{"tiger"}. The default value is \code{"mb"}.
#' @param scr If \code{scr = TRUE}, the lossy screening rule is applied to preselect the neighborhood before the graph estimation. The default value is  \code{FALSE}. NOT applicable when \code{method = "ct"} or \code{"tiger"}.
#' @param scr.num The neighborhood size after the lossy screening rule (the number of remaining neighbors per node). ONLY applicable when \code{scr = TRUE}. The default value is \code{n-1}. An alternative value is \code{n/log(n)}. ONLY applicable when \code{scr = TRUE} and \code{method = "mb"}.
#' @param cov.output If \code{cov.output = TRUE}, the output will include a path of estimated covariance matrices. ONLY applicable when \code{method = "glasso"}. Since the estimated covariance matrices are generally not sparse, please use it with care, or it may take much memory under high-dimensional setting. The default value is \code{FALSE}.
#' @param sym Symmetrize the output graphs. If \code{sym = "and"}, the edge between node \code{i} and node \code{j} is selected ONLY when both node \code{i} and node \code{j} are selected as neighbors for each other. If \code{sym = "or"}, the edge is selected when either node \code{i} or node \code{j} is selected as the neighbor for each other. The default value is \code{"or"}. ONLY applicable when \code{method = "mb"} or \code{"tiger"}.
#' @param verbose If \code{verbose = FALSE}, tracing information printing is disabled. The default value is \code{TRUE}.
#' @param input.type How to interpret \code{x}: \code{"auto"} preserves
#'   symmetry-based detection, \code{"data"} forces an observation matrix,
#'   and \code{"covariance"} requires a square covariance or correlation
#'   matrix. Use \code{"data"} for square symmetric observation matrices.
#'   For glasso covariance input, \code{"auto"} retains the historical
#'   diagonal-sensitive default lambda scale; use \code{"covariance"} for a
#'   scale based only on off-diagonal entries. Glasso may accept an indefinite
#'   pairwise covariance estimate when regularization produces a certified
#'   positive-definite result; CT, MB, and TIGER require positive
#'   semidefiniteness.
#' @return
#' An object with S3 class \code{"huge"} is returned:
#' \item{data}{
#'   The \code{n} by \code{d} data matrix or \code{d} by \code{d} sample covariance matrix from the input
#' }
#' \item{cov.input}{
#'   An indicator of the sample covariance.
#' }
#' \item{ind.mat}{
#'   The \code{scr.num} by \code{d} matrix with each column containing the indices of the remaining neighbors of the corresponding variable after the lossy screening. ONLY applicable when \code{scr = TRUE}.
#' }
#' \item{lambda}{
#'   The sequence of regularization parameters used in mb or thresholding parameters in ct.
#' }
#' \item{sym}{
#'   The \code{sym} from the input. ONLY applicable when \code{method = "mb"} or \code{"tiger"}.
#' }
#' \item{scr}{
#'   The \code{scr} from the input. ONLY applicable when \code{method = "mb"} or \code{"glasso"}.
#' }
#' \item{path}{
#'   A list of \code{d} by \code{d} adjacency matrices of estimated graphs as a graph path corresponding to \code{lambda}.
#' }
#' \item{sparsity}{
#'   The sparsity levels of the graph path.
#' }
#' \item{icov}{
#'   A list of \code{d} by \code{d} precision matrices as an alternative graph path (numerical path) corresponding to \code{lambda}. ONLY applicable when \code{method = "glasso"} or \code{"tiger"}.
#' }
#' \item{cov}{
#'   A list of \code{d} by \code{d} estimated covariance matrices corresponding to \code{lambda}. ONLY applicable when \code{cov.output = TRUE} and \code{method = "glasso"}
#' }
#' \item{method}{
#'   The method used in the graph estimation stage.
#' }
#' \item{df}{
#'   If \code{method = "mb"} or \code{"tiger"}, it is a \code{d} by \code{nlambda} matrix. Each row contains the number of nonzero coefficients along the lasso solution path. If \code{method = "glasso"}, it is a \code{nlambda} dimensional vector containing the number of nonzero coefficients along the graph path \code{icov}.
#' }
#' \item{loglik}{
#'   A \code{nlambda} dimensional vector containing the likelihood scores along the graph path (\code{icov}). ONLY applicable when \code{method = "glasso"}. For an estimated inverse covariance Z, the program only calculates log(det(Z)) - trace(SZ) where S is the empirical covariance matrix. For the likelihood for n observations, please multiply by n/2.
#' }
#' @note This function ONLY estimates the graph path. For more information about the optimal graph selection, please refer to \code{\link{huge.select}}.\cr
#' @seealso \code{\link{huge.generator}}, \code{\link{huge.select}}, \code{\link{huge.plot}}, \code{\link{huge.roc}}, and \code{\link{huge-package}}.
#' @examples
#' #generate data
#' L = huge.generator(n = 50, d = 12, graph = "hub", g = 4)
#'
#' #graph path estimation using mb
#' out1 = huge(L$data)
#' out1
#' plot(out1)         #Not aligned
#' plot(out1, align = TRUE) #Aligned
#' huge.plot(out1$path[[3]])
#' @export
huge = function(x, lambda = NULL, nlambda = NULL, lambda.min.ratio = NULL, method = "mb", scr = NULL, scr.num = NULL, cov.output = FALSE, sym = "or", verbose = TRUE, input.type = "auto")
{

	# Each method function returns a list carrying all of its output fields;
	# adopt it wholesale rather than copying fields one by one (a historical
	# source of missed-field bugs), then attach the shared metadata.
	if(method == "ct")
		est = huge.ct(x, nlambda = nlambda, lambda.min.ratio = lambda.min.ratio, lambda = lambda, verbose = verbose, input.type = input.type)
	else if(method == "mb")
		est = huge.mb(x, lambda = lambda, nlambda = nlambda, lambda.min.ratio = lambda.min.ratio, scr = scr, scr.num = scr.num, sym = sym, verbose = verbose, input.type = input.type)
	else if(method == "glasso")
		est = huge.glasso(x, nlambda = nlambda, lambda.min.ratio = lambda.min.ratio, lambda = lambda, scr = scr, cov.output = cov.output, verbose = verbose, input.type = input.type)
	else if(method == "tiger")
		est = huge.tiger(x, lambda = lambda, nlambda = nlambda, lambda.min.ratio = lambda.min.ratio, sym = sym, verbose = verbose, input.type = input.type)
	else
		stop("method must be one of \"mb\", \"glasso\", \"ct\", \"tiger\".")

	if(method %in% c("mb", "tiger")) est$sym = sym

	est$method = method
	est$data = x

	class(est) = "huge"
	return(est)
}

#' Print function for S3 class "huge"
#'
#' Print the information about the model usage, the graph path length, graph dimension, sparsity level.
#'
#' @param x An object with S3 class \code{"huge"}.
#' @param \dots System reserved (No specific usage)
#' @seealso \code{\link{huge}}
#' @export
print.huge = function(x, ...)
{
	if(x$method == "ct")
		cat("Model: graph estimation via correlation thresholding (ct)\n")
	else if(x$method == "glasso")
		cat("Model: graphical lasso (glasso)\n")
	else if(x$method == "mb")
		cat("Model: Meinshausen & Buhlmann graph estimation (mb)\n")
	else if(x$method == "tiger")
    cat("Model: tuning-insensitive approach\n")

	if((x$method == "glasso")&&(x$scr)) cat("lossy screening: on\n")

	if(x$cov.input) cat("Input: The Covariance Matrix\n")
	if(!x$cov.input) cat("Input: The Data Matrix\n")

	cat("Path length:",length(x$lambda),"\n")
	cat("Graph dimension:",ncol(x$data),"\n")
	cat("Sparsity level:",min(x$sparsity),"----->",max(x$sparsity),"\n")
}

#' Plot function for S3 class "huge"
#'
#' Plot sparsity level information and 3 typical sparse graphs from the graph path.
#'
#' @param x An object with S3 class \code{"huge"}
#' @param align If \code{align = TRUE}, 3 plotted graphs are aligned to the same sparsity levels. The default value is \code{FALSE}.
#' @param \dots System reserved (No specific usage)
#' @seealso \code{\link{huge}}
#' @export
plot.huge = function(x, align = FALSE, ...){
	old.par = .huge_graphics_state()
	on.exit(.huge_restore_graphics_state(old.par), add = TRUE)
	if(length(x$lambda) == 1)	par(mfrow = c(1, 2), pty = "s", omi=c(0.3,0.3,0.3,0.3), mai = c(0.3,0.3,0.3,0.3))
	if(length(x$lambda) == 2)	par(mfrow = c(1, 3), pty = "s", omi=c(0.3,0.3,0.3,0.3), mai = c(0.3,0.3,0.3,0.3))
	if(length(x$lambda) >= 3)	par(mfrow = c(1, 4), pty = "s", omi=c(0.3,0.3,0.3,0.3), mai = c(0.3,0.3,0.3,0.3))

	if(length(x$lambda) <= 3)	z.final = 1:length(x$lambda)

	if(length(x$lambda) >=4){
		z.max = max(x$sparsity)
		z.min = min(x$sparsity)
		z = z.max - z.min
		z.unique = unique(c(which(x$sparsity>=(z.min + 0.03*z))[1],which(x$sparsity>=(z.min + 0.07*z))[1],which(x$sparsity>=(z.min + 0.15*z))[1]))


		if(length(z.unique) == 1){
			if(z.unique<(length(x$lambda)-1))	z.final = c(z.unique,z.unique+1,z.unique+2)
			if(z.unique==(length(x$lambda)-1)) z.final = c(z.unique-1,z.unique,z.unique+1)
			if(z.unique==length(x$lambda)) 	z.final = c(z.unique-2,z.unique-1,z.unique)
		}

		if(length(z.unique) == 2){
			if(diff(z.unique)==1){
				if(z.unique[2]<length(x$lambda)) z.final = c(z.unique,z.unique[2]+1)
				if(z.unique[2]==length(x$lambda)) z.final = c(z.unique[1]-1,z.unique)
			}
			if(diff(z.unique)>1) z.final = c(z.unique[1],z.unique[1]+1,z.unique[2])
		}

		if(length(z.unique) == 3) z.final = z.unique

	

	}
	plot(x$lambda, x$sparsity, log = "x", xlab = "Regularization Parameter", ylab = "Sparsity Level", type = "l",xlim = rev(range(x$lambda)), main = "Sparsity vs. Regularization")

	lines(x$lambda[z.final],x$sparsity[z.final],type = "p")

	if(align){
		layout.grid = layout_with_fr(graph_from_adjacency_matrix(as.matrix(x$path[[z.final[length(z.final)]]]), mode="undirected", diag=FALSE))
		for(i in z.final){
			g = graph_from_adjacency_matrix(as.matrix(x$path[[i]]), mode="undirected", diag=FALSE)
			plot(g, layout=layout.grid, edge.color='gray50',vertex.color="red", vertex.size=3, vertex.label=NA, main = paste("lambda = ",as.character(round(x$lambda[i],3)),sep = ""))
		}
	}
	if(!align){
		for(i in z.final){
			g = graph_from_adjacency_matrix(as.matrix(x$path[[i]]), mode="undirected", diag=FALSE)
			layout.grid = layout_with_fr(g)
			plot(g, layout=layout.grid, edge.color='gray50',vertex.color="red", vertex.size=3, vertex.label=NA, main = paste("lambda = ",as.character(round(x$lambda[i],3)),sep = ""))
		}
	}
	if(align) cat("Three plotted graphs are aligned according to the third graph\n")
}

# Property-based invariant tests: random inputs, universal properties.
# Complements the fixed-fixture tests — these hunt for unknown bugs rather
# than guarding known fixes. Seeded loops keep runtime bounded (~seconds).

random_shape = function(seed) {
  set.seed(seed)
  list(n = sample(15:120, 1), d = sample(4:50, 1), nlambda = sample(1:8, 1))
}

test_that("all methods: outputs symmetric, sparsity in [0,1] and monotone", {
  for (seed in 1:6) {
    s = random_shape(seed)
    x = matrix(rnorm(s$n * s$d), s$n, s$d)
    for (m in c("mb", "glasso", "ct", "tiger")) {
      fit = if (m == "tiger")
        suppressWarnings(huge(x, method = m, nlambda = s$nlambda,
                              verbose = FALSE)) else
        huge(x, method = m, nlambda = s$nlambda, verbose = FALSE)
      if (m == "tiger") {
        expect_gte(length(fit$path), 1)
        expect_lte(length(fit$path), s$nlambda)
      } else {
        expect_length(fit$path, s$nlambda)
      }
      for (p in fit$path) {
        pm = as.matrix(p)
        expect_identical(max(abs(pm - t(pm))), 0)
        expect_identical(max(abs(diag(pm))), 0)
      }
      expect_true(all(fit$sparsity >= 0 & fit$sparsity <= 1))
      # lambda is decreasing along the path, so sparsity is nondecreasing
      if (s$nlambda > 1) expect_true(all(diff(fit$sparsity) >= -1e-12))
    }
  }
})

test_that("ct path is nested: larger lambda support is subset of smaller", {
  for (seed in 7:9) {
    s = random_shape(seed)
    if (s$nlambda < 2) s$nlambda = 4
    x = matrix(rnorm(s$n * s$d), s$n, s$d)
    fit = huge(x, method = "ct", nlambda = s$nlambda, verbose = FALSE)
    for (i in 1:(s$nlambda - 1)) {
      a = as.matrix(fit$path[[i]]) != 0
      b = as.matrix(fit$path[[i + 1]]) != 0
      expect_true(all(b[a]))  # every edge at larger lambda persists
    }
  }
})

test_that("glasso icov positive definite and consistent with path", {
  for (seed in 10:12) {
    s = random_shape(seed)
    x = matrix(rnorm(s$n * s$d), s$n, s$d)
    fit = huge(x, method = "glasso", nlambda = 4, verbose = FALSE)
    for (i in seq_along(fit$icov)) {
      ic = as.matrix(fit$icov[[i]])
      expect_gt(min(eigen(ic, symmetric = TRUE, only.values = TRUE)$values), 0)
      # path pattern == off-diagonal support of icov
      expect_identical(as.matrix(fit$path[[i]]) != 0,
                       (ic != 0) & (row(ic) != col(ic)))
    }
  }
})

test_that("extreme column scales do not break estimation (internal standardization)", {
  set.seed(31)
  n = 60; d = 12
  x = matrix(rnorm(n * d), n, d)
  x_scaled = sweep(x, 2, 10^seq(-6, 6, length.out = d), "*")
  f1 = huge(x, method = "mb", nlambda = 4, verbose = FALSE)
  f2 = huge(x_scaled, method = "mb", nlambda = 4, verbose = FALSE)
  # correlation-based: identical graphs regardless of column scales
  for (i in 1:4)
    expect_identical(as.matrix(f1$path[[i]]), as.matrix(f2$path[[i]]))
})

test_that("near-singular correlation input stays finite", {
  set.seed(32)
  n = 80; d = 10
  base = rnorm(n)
  x = sapply(1:d, function(j) base + rnorm(n, sd = 0.01))  # corr ~ 0.9999
  for (m in c("mb", "glasso", "tiger")) {
    if (m == "tiger") {
      expect_warning(
        fit <- huge(x, method = m, nlambda = 4, verbose = FALSE),
        "certified prefix"
      )
    } else {
      fit = huge(x, method = m, nlambda = 4, verbose = FALSE)
    }
    expect_true(all(is.finite(fit$sparsity)))
    if (!is.null(fit[["icov"]]))
      for (ic in fit$icov) expect_true(all(is.finite(as.matrix(ic))))
  }
})

test_that("nlambda = 1 works everywhere (degenerate path)", {
  set.seed(33)
  x = matrix(rnorm(50 * 8), 50, 8)
  for (m in c("mb", "glasso", "ct", "tiger")) {
    fit = huge(x, method = m, nlambda = 1, verbose = FALSE)
    expect_length(fit$path, 1)
  }
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

RIC <- function(X, d, n, r, t) {
    .Call(`_huge_RIC`, X, d, n, r, t)
}

SFGen <- function(d0, d) {
    .Call(`_huge_SFGen`, d0, d)
}

SPMBscr <- function(S, lambda, nlambda, d, idx_scr, nscr) {
    .Call(`_huge_SPMBscr`, S, lambda, nlambda, d, idx_scr, nscr)
}

SPMBgraph <- function(S, lambda, nlambda, d) {
    .Call(`_huge_SPMBgraph`, S, lambda, nlambda, d)
}

SPMBgraphsqrt <- function(data, lambda, nlambda, d) {
    .Call(`_huge_SPMBgraphsqrt`, data, lambda, nlambda, d)
}

SPMBgraphsqrtFit <- function(input, lambda, nlambda, d, covariance_input, lambda_min_ratio) {
    .Call(`_huge_SPMBgraphsqrtFit`, input, lambda, nlambda, d, covariance_input, lambda_min_ratio)
}

hugeglasso <- function(S, lambda, scr, verbose, cov_output) {
    .Call(`_huge_hugeglasso`, S, lambda, scr, verbose, cov_output)
}

#' @useDynLib huge, .registration = TRUE
#' @importFrom Rcpp sourceCpp
#' @importFrom Rcpp evalCpp
#' @importFrom stats cov2cor cor runif qnorm sd pnorm
#' @importFrom graphics par image plot lines
#' @importFrom grDevices postscript gray.colors dev.off
#' @importFrom Matrix Matrix
#' @importFrom parallel mclapply
#' @importFrom utils flush.console
#' @importFrom MASS mvrnorm
#' @importFrom igraph graph_from_adjacency_matrix layout_with_fr
#' @importFrom methods new
NULL

#' High-Dimensional Undirected Graph Estimation
#'
#' A package for high-dimensional undirected graph estimation
#'
#' The package "huge" provides 9 main functions:\cr
#' (1) the data generator creates random samples from multivariate normal distributions with different graph structures. Please refer to \code{\link{huge.generator}}.\cr
#' (2) the nonparanormal (npn) transformation helps relax the normality assumption. Please refer to \code{\link{huge.npn}}.\cr
#' (3) The correlation thresholding graph estimation. Please refer to \code{\link{huge}}.\cr
#' (4) The Meinshausen-Buhlmann graph estimation. Please refer to \code{\link{huge}}.\cr
#' (5) The graphical Lasso algorithm using lossless screening rule. Please refer to \code{\link{huge}}.\cr
#' (6) The tuning-insensitive graph estimation (tiger). Please refer to \code{\link{huge.tiger}}.\cr
#' \cr
#' Both (4) and (5) can be further accelerated by the lossy screening rule preselecting the neighborhood of each node via thresholding sample correlation.
#' \cr
#' (7) The model selection using the stability approach to regularization selection. Please refer to \code{\link{huge.select}}.\cr
#' (8) The model selection using the rotation information criterion. Please refer to \code{\link{huge.select}}.\cr
#' (9) The model selection using the extended Bayesian information criterion. Please refer to \code{\link{huge.select}}.\cr
#' @docType package
#' @aliases huge-package
#' @author Tuo Zhao, Han Liu, Haoming Jiang, Kathryn Roeder, John Lafferty, and Larry Wasserman \cr
#' Maintainer: Tuo Zhao <tourzhao@gatech.edu>
#' @references
#' 1.  T. Zhao and H. Liu. The huge Package for High-dimensional Undirected Graph Estimation in R. \emph{Journal of Machine Learning Research}, 2012\cr
#' 2.  H. Liu, F. Han, M. Yuan, J. Lafferty and L. Wasserman. High Dimensional Semiparametric Gaussian Copula Graphical Models. \emph{Annals of Statistics},2012 \cr
#' 3.  D. Witten and J. Friedman. New insights and faster computations for the graphical lasso. \emph{Journal of Computational and Graphical Statistics}, to appear, 2011.
#' 4.  Han Liu, Kathryn Roeder and Larry Wasserman. Stability Approach to Regularization Selection (StARS) for High Dimensional Graphical Models. \emph{Advances in Neural Information Processing Systems}, 2010.\cr
#' 5.  R. Foygel and M. Drton. Extended bayesian information criteria for gaussian graphical models. \emph{Advances in Neural Information Processing Systems}, 2010.\cr
#' 6.  H. Liu, J. Lafferty and L. Wasserman. The Nonparanormal: Semiparametric Estimation of High Dimensional Undirected Graphs. \emph{Journal of Machine Learning Research}, 2009 \cr
#' 7.  J. Fan and J. Lv. Sure independence screening for ultra-high dimensional feature space (with discussion). \emph{Journal of Royal Statistical Society B}, 2008.\cr
#' 8.  O. Banerjee, L. E. Ghaoui, A. d'Aspremont: Model Selection Through Sparse Maximum Likelihood Estimation for Multivariate Gaussian or Binary Data. \emph{Journal of Machine Learning Research}, 2008.\cr
#' 9.  J. Friedman, T. Hastie and R. Tibshirani. Regularization Paths for Generalized Linear Models via Coordinate Descent. \emph{Journal of Statistical Software}, 2008. \cr
#' 10. J. Friedman, T. Hastie and R. Tibshirani. Sparse inverse covariance estimation with the lasso, \emph{Biostatistics}, 2007.\cr
#' 11. N. Meinshausen and P. Buhlmann. High-dimensional Graphs and Variable Selection with the Lasso. \emph{The Annals of Statistics}, 2006.\cr
#' @seealso \code{\link{huge.generator}}, \code{\link{huge.npn}}, \code{\link{huge}}, \code{\link{huge.tiger}}, \code{\link{huge.select}}, \code{\link{huge.plot}} and \code{\link{huge.roc}}
"_PACKAGE"
#> [1] "_PACKAGE"

#' Stock price of S&P 500 companies from 2003 to 2008
#'
#' This data set consists of stock price and company information.
#'
#' This data set can be used to perform high-dimensional graph estimation to analyze the relationships between S&P 500 companies.
#'
#' @usage data(stockdata)
#' @format
#' The format is a list containing contains two matrices.
#' 1. data - 1258x452, represents the 452 stocks' close prices for 1258 trading days.
#'   2. info - 452x3:
#'   The 1st column: the query symbol for each company.
#'   The 2nd column: the category for each company.
#'   The 3rd column: the full name of each company.
#' @source It was publicly available at finance.yahoo, which is now out of date
#' @examples
#' data(stockdata)
#' image(stockdata$data)
#' stockdata$info
#' @keywords datasets
#' @docType data
"stockdata"
#> [1] "stockdata"

#-----------------------------------------------------------------------#
# Package: High-dimensional Undirected Graph Estimation                 #
# huge.gect(): graph estimation via correlation thresholding (ct)       #
#-----------------------------------------------------------------------#

#' Graph estimation via correlation thresholding (ct) 
#' 
#' See more details in \code{\link{huge}}
#' @details The default path targets increasing numbers of undirected edges
#'   and defines each graph by the strict rule
#'   \code{abs(correlation) > lambda}. Equal-weight edges are never split, so
#'   ties can make the realized sparsity smaller than the nominal target.
#'   Reusing the returned \code{lambda} values therefore reconstructs the
#'   same path. When \code{lambda = NULL}, \code{nlambda} must be a positive
#'   integer and \code{lambda.min.ratio} must lie in \code{(0, 1]}. Supplying
#'   \code{lambda} overrides those two arguments.
#' @param x There are 2 options: (1) \code{x} is an \code{n} by \code{d} data matrix (2) a \code{d} by \code{d} sample covariance matrix. The program automatically identifies the input matrix by checking the symmetry. (\code{n} is the sample size and \code{d} is the dimension).
#' @param lambda A numeric scalar or non-empty one-dimensional numeric input
#'   of finite, non-negative thresholds. Values are applied in the supplied
#'   order, and zero is allowed. Leave
#'   \code{lambda = NULL} to generate a path from \code{nlambda} and
#'   \code{lambda.min.ratio}.
#' @param nlambda The number of regularization/thresholding parameters. The default value is \code{20} for \code{method = "ct"} and \code{10} for \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}.
#' @param lambda.min.ratio If \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}, it is the smallest value for \code{lambda}, as a fraction of the upperbound (\code{MAX}) of the regularization/thresholding parameter which makes all estimates equal to \code{0}. The program can automatically generate \code{lambda} as a sequence of length = \code{nlambda} starting from \code{MAX} to \code{lambda.min.ratio*MAX} in log scale. If \code{method = "ct"}, it is the largest sparsity level for estimated graphs. The program can automatically generate \code{lambda} as a sequence of length = \code{nlambda}, which makes the sparsity level of the graph path increases from \code{0} to \code{lambda.min.ratio} evenly.The default value is \code{0.1} when \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}, and 0.05 when \code{method = "ct"}.
#' @param verbose If \code{verbose = FALSE}, tracing information printing is disabled. The default value is \code{TRUE}.
#' @param input.type How to interpret \code{x}: \code{"auto"} preserves
#'   symmetry-based detection, \code{"data"} forces an observation matrix,
#'   and \code{"covariance"} requires a square covariance or correlation
#'   matrix.
#' @seealso \code{\link{huge}}, and \code{\link{huge-package}}.
#' @export
huge.ct = function(x, nlambda = NULL, lambda.min.ratio = NULL, lambda = NULL, verbose = TRUE, input.type = "auto")
{
  inp = .huge_preprocess(x, verbose, input.type = input.type)
  S = inp$S; d = inp$d
  fit = list()
  fit$cov.input = inp$cov.input

  diag(S) = 0
  S = abs(S)

  ct.progress = function(i)
  {
    cat(paste(c("Conducting the graph estimation via correlation thresholding (ct)....in progress:", floor(100*i/nlambda), "%"), collapse=""), "\r")
    flush.console()
  }

  if(is.null(lambda))
  {
    if(is.null(nlambda))
      nlambda = 20
    else
      nlambda = .huge_validate_positive_integer(nlambda, "nlambda")
    if(is.null(lambda.min.ratio))
      lambda.min.ratio = 0.05
    else
      lambda.min.ratio = .huge_validate_ratio(lambda.min.ratio)

    edge.weights = sort(S[upper.tri(S)], decreasing = TRUE)
    edge.total = length(edge.weights)

    if(edge.total == 0)
      lambda = rep(0, nlambda)
    else
    {
      target.edges = ceiling(seq(
        1, lambda.min.ratio * edge.total, length.out = nlambda
      ))
      target.edges = pmax(0L, pmin(edge.total, as.integer(target.edges)))
      lambda = vapply(target.edges, function(target) {
        if(target < edge.total)
        {
          next.edge = max(target + 1L, 1L)
          return(edge.weights[next.edge])
        }
        0
      }, numeric(1))
    }
  }
  else
    lambda = .huge_validate_lambda(lambda, allow.zero = TRUE)

  nlambda = length(lambda)
  fit$path = list()
  fit$sparsity = rep(0,nlambda)
  for(i in seq_len(nlambda))
  {
    fit$path[[i]] = Matrix(0, d, d)
    fit$path[[i]][S > lambda[i]] = 1
    if(d <= 1)
      fit$sparsity[i] = 0
    else
      fit$sparsity[i] = sum(fit$path[[i]])/d/(d-1)
    if(verbose) ct.progress(i)
  }
  fit$lambda = lambda

  if(verbose)
  {
        cat("Conducting the graph estimation via correlation thresholding (ct)....done.             \r\n")
        flush.console()
    }
  return(fit)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huge.roc.R
\name{plot.roc}
\alias{plot.roc}
\title{Plot function for S3 class "roc"}
\usage{
\method{plot}{roc}(x, ...)
}
\arguments{
\item{x}{An object with S3 class \code{"roc"}}

\item{\dots}{System reserved (No specific usage)}
}
\description{
Plot the ROC curve for an object with S3 class \code{"roc"}.
}
\seealso{
\code{\link{huge.roc}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huge.R
\name{print.huge}
\alias{print.huge}
\title{Print function for S3 class "huge"}
\usage{
\method{print}{huge}(x, ...)
}
\arguments{
\item{x}{An object with S3 class \code{"huge"}.}

\item{\dots}{System reserved (No specific usage)}
}
\description{
Print the information about the model usage, the graph path length, graph dimension, sparsity level.
}
\seealso{
\code{\link{huge}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huge.roc.R
\name{print.roc}
\alias{print.roc}
\title{Print function for S3 class "roc"}
\usage{
\method{print}{roc}(x, ...)
}
\arguments{
\item{x}{An object with S3 class \code{"roc"}.}

\item{\dots}{System reserved (No specific usage)}
}
\description{
Print the information about true positive rates, false positive rates, the area under curve and maximum F1 score.
}
\seealso{
\code{\link{huge.roc}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huge.plot.R
\name{huge.plot}
\alias{huge.plot}
\title{Graph visualization}
\usage{
huge.plot(
  G,
  epsflag = FALSE,
  graph.name = "default",
  cur.num = 1,
  location = NULL
)
}
\arguments{
\item{G}{The adjacency matrix corresponding to the graph.}

\item{epsflag}{If \code{epsflag = TRUE}, save the plot as an eps file in the target directory. The default value is \code{FALSE}.}

\item{graph.name}{The name of the output eps files. The default value is "default".}

\item{cur.num}{The number of plots saved as eps files. Only applicale when \code{epsflag = TRUE}. The default value is 1.}

\item{location}{Target directory. The default value is the current working directory.}
}
\description{
Implements the graph visualization using adjacency matrix. It can automatic organize 2D embedding layout.
}
\details{
The user can change \code{cur.num} to plot several figures and select the best one. The implementation is based on the popular package "igraph".
}
\examples{
## visualize the hub graph
L = huge.generator(graph = "hub")
huge.plot(L$theta)

## visualize the band graph
L = huge.generator(graph = "band",g=5)
huge.plot(L$theta)

## visualize the cluster graph
L = huge.generator(graph = "cluster")
huge.plot(L$theta)

## plot 5 graphs and save the plots as eps files in the tempdir()
huge.plot(L$theta, epsflag = TRUE, cur.num = 5, location = tempdir())
}
\seealso{
\code{\link{huge}} and \code{\link{huge-package}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huge.mb.R
\name{huge.mb}
\alias{huge.mb}
\title{Meinshausen & Buhlmann graph estimation}
\usage{
huge.mb(
  x,
  lambda = NULL,
  nlambda = NULL,
  lambda.min.ratio = NULL,
  scr = NULL,
  scr.num = NULL,
  idx.mat = NULL,
  sym = "or",
  verbose = TRUE,
  input.type = "auto"
)
}
\arguments{
\item{x}{There are 2 options: (1) \code{x} is an \code{n} by \code{d} data matrix (2) a \code{d} by \code{d} sample covariance matrix. The program automatically identifies the input matrix by checking the symmetry. (\code{n} is the sample size and \code{d} is the dimension).}

\item{lambda}{A numeric scalar or non-empty one-dimensional numeric input
defining a finite, strictly positive, non-increasing regularization path;
tied values are allowed. Leave \code{lambda = NULL}
to generate a path from \code{nlambda} and \code{lambda.min.ratio}.}

\item{nlambda}{The number of regularization/thresholding parameters. The default value is \code{20} for \code{method = "ct"} and \code{10} for \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}.}

\item{lambda.min.ratio}{If \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}, it is the smallest value for \code{lambda}, as a fraction of the upperbound (\code{MAX}) of the regularization/thresholding parameter which makes all estimates equal to \code{0}. The program can automatically generate \code{lambda} as a sequence of length = \code{nlambda} starting from \code{MAX} to \code{lambda.min.ratio*MAX} in log scale. If \code{method = "ct"}, it is the largest sparsity level for estimated graphs. The program can automatically generate \code{lambda} as a sequence of length = \code{nlambda}, which makes the sparsity level of the graph path increases from \code{0} to \code{lambda.min.ratio} evenly.The default value is \code{0.1} when \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}, and 0.05 when \code{method = "ct"}.}

\item{scr}{If \code{scr = TRUE}, the lossy screening rule is applied to preselect the neighborhood before the graph estimation. The default value is  \code{FALSE}.}

\item{scr.num}{The neighborhood size after the lossy screening rule (the number of remaining neighbors per node). It must be an integer between \code{1} and \code{d - 1}. ONLY applicable when \code{scr = TRUE}. The default value is \code{n-1}. An alternative value is \code{n/log(n)}. ONLY applicable when \code{scr = TRUE} and \code{method = "mb"}.}

\item{idx.mat}{A \code{scr.num} by \code{d} index matrix for screening. Column \code{m} must contain distinct, zero-based indices in \code{0:(d - 1)} and must exclude the response index \code{m - 1}.}

\item{sym}{Symmetrize the output graphs. If \code{sym = "and"}, the edge between node \code{i} and node \code{j} is selected ONLY when both node \code{i} and node \code{j} are selected as neighbors for each other. If \code{sym = "or"}, the edge is selected when either node \code{i} or node \code{j} is selected as the neighbor for each other. The default value is \code{"or"}. ONLY applicable when \code{method = "mb"} or \code{"tiger"}.}

\item{verbose}{If \code{verbose = FALSE}, tracing information printing is disabled. The default value is \code{TRUE}.}

\item{input.type}{How to interpret \code{x}: \code{"auto"} preserves
symmetry-based detection, \code{"data"} forces an observation matrix,
and \code{"covariance"} requires a square covariance or correlation
matrix.}
}
\description{
See more details in \code{\link{huge}}
}
\seealso{
\code{\link{huge}}, and \code{\link{huge-package}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huge.select.R
\name{plot.select}
\alias{plot.select}
\title{Plot function for S3 class "select"}
\usage{
\method{plot}{select}(x, ...)
}
\arguments{
\item{x}{An object with S3 class \code{"select"}}

\item{\dots}{System reserved (No specific usage)}
}
\description{
Plot the optimal graph by model selection.
}
\seealso{
\code{\link{huge.select}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huge.select.R
\name{print.select}
\alias{print.select}
\title{Print function for S3 class "select"}
\usage{
\method{print}{select}(x, ...)
}
\arguments{
\item{x}{An object with S3 class \code{"select"}.}

\item{\dots}{System reserved (No specific usage)}
}
\description{
Print the information about the model usage, graph dimension, model selection criterion, sparsity level of the optimal graph.
}
\seealso{
\code{\link{huge.select}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huge.generator.R
\name{print.sim}
\alias{print.sim}
\title{Print function for S3 class "sim"}
\usage{
\method{print}{sim}(x, ...)
}
\arguments{
\item{x}{An object with S3 class \code{"sim"}.}

\item{\dots}{System reserved (No specific usage)}
}
\description{
Print the information about the sample size, the dimension, the pattern and sparsity of the true graph structure.
}
\seealso{
\code{\link{huge.generator}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huge-package.R
\docType{package}
\name{huge-package}
\alias{huge-package}
\title{High-Dimensional Undirected Graph Estimation}
\description{
A package for high-dimensional undirected graph estimation
}
\details{
The package "huge" provides 9 main functions:\cr
(1) the data generator creates random samples from multivariate normal distributions with different graph structures. Please refer to \code{\link{huge.generator}}.\cr
(2) the nonparanormal (npn) transformation helps relax the normality assumption. Please refer to \code{\link{huge.npn}}.\cr
(3) The correlation thresholding graph estimation. Please refer to \code{\link{huge}}.\cr
(4) The Meinshausen-Buhlmann graph estimation. Please refer to \code{\link{huge}}.\cr
(5) The graphical Lasso algorithm using lossless screening rule. Please refer to \code{\link{huge}}.\cr
(6) The tuning-insensitive graph estimation (tiger). Please refer to \code{\link{huge.tiger}}.\cr
\cr
Both (4) and (5) can be further accelerated by the lossy screening rule preselecting the neighborhood of each node via thresholding sample correlation.
\cr
(7) The model selection using the stability approach to regularization selection. Please refer to \code{\link{huge.select}}.\cr
(8) The model selection using the rotation information criterion. Please refer to \code{\link{huge.select}}.\cr
(9) The model selection using the extended Bayesian information criterion. Please refer to \code{\link{huge.select}}.\cr
}
\references{
1.  T. Zhao and H. Liu. The huge Package for High-dimensional Undirected Graph Estimation in R. \emph{Journal of Machine Learning Research}, 2012\cr
2.  H. Liu, F. Han, M. Yuan, J. Lafferty and L. Wasserman. High Dimensional Semiparametric Gaussian Copula Graphical Models. \emph{Annals of Statistics},2012 \cr
3.  D. Witten and J. Friedman. New insights and faster computations for the graphical lasso. \emph{Journal of Computational and Graphical Statistics}, to appear, 2011.
4.  Han Liu, Kathryn Roeder and Larry Wasserman. Stability Approach to Regularization Selection (StARS) for High Dimensional Graphical Models. \emph{Advances in Neural Information Processing Systems}, 2010.\cr
5.  R. Foygel and M. Drton. Extended bayesian information criteria for gaussian graphical models. \emph{Advances in Neural Information Processing Systems}, 2010.\cr
6.  H. Liu, J. Lafferty and L. Wasserman. The Nonparanormal: Semiparametric Estimation of High Dimensional Undirected Graphs. \emph{Journal of Machine Learning Research}, 2009 \cr
7.  J. Fan and J. Lv. Sure independence screening for ultra-high dimensional feature space (with discussion). \emph{Journal of Royal Statistical Society B}, 2008.\cr
8.  O. Banerjee, L. E. Ghaoui, A. d'Aspremont: Model Selection Through Sparse Maximum Likelihood Estimation for Multivariate Gaussian or Binary Data. \emph{Journal of Machine Learning Research}, 2008.\cr
9.  J. Friedman, T. Hastie and R. Tibshirani. Regularization Paths for Generalized Linear Models via Coordinate Descent. \emph{Journal of Statistical Software}, 2008. \cr
10. J. Friedman, T. Hastie and R. Tibshirani. Sparse inverse covariance estimation with the lasso, \emph{Biostatistics}, 2007.\cr
11. N. Meinshausen and P. Buhlmann. High-dimensional Graphs and Variable Selection with the Lasso. \emph{The Annals of Statistics}, 2006.\cr
}
\seealso{
\code{\link{huge.generator}}, \code{\link{huge.npn}}, \code{\link{huge}}, \code{\link{huge.tiger}}, \code{\link{huge.select}}, \code{\link{huge.plot}} and \code{\link{huge.roc}}
}
\author{
Tuo Zhao, Han Liu, Haoming Jiang, Kathryn Roeder, John Lafferty, and Larry Wasserman \cr
Maintainer: Tuo Zhao <tourzhao@gatech.edu>
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huge-package.R
\docType{data}
\name{stockdata}
\alias{stockdata}
\title{Stock price of S&P 500 companies from 2003 to 2008}
\format{
The format is a list containing contains two matrices.
1. data - 1258x452, represents the 452 stocks' close prices for 1258 trading days.
  2. info - 452x3:
  The 1st column: the query symbol for each company.
  The 2nd column: the category for each company.
  The 3rd column: the full name of each company.
}
\source{
It was publicly available at finance.yahoo, which is now out of date
}
\usage{
data(stockdata)
}
\description{
This data set consists of stock price and company information.
}
\details{
This data set can be used to perform high-dimensional graph estimation to analyze the relationships between S&P 500 companies.
}
\examples{
data(stockdata)
image(stockdata$data)
stockdata$info
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huge.inference.R
\name{huge.inference}
\alias{huge.inference}
\title{Graph inference}
\usage{
huge.inference(data, T, adj, alpha = 0.05, type = "Gaussian", method = "score")
}
\arguments{
\item{data}{A finite numeric \code{n} by \code{d} data matrix with at least two observations and no constant columns.}

\item{T}{A finite \code{d} by \code{d} estimate of the inverse correlation matrix with a positive diagonal.}

\item{adj}{A finite numeric or logical \code{d} by \code{d} adjacency matrix corresponding to the graph.}

\item{alpha}{The significance level in \code{(0, 1]}. The default value is \code{0.05}.}

\item{type}{The type of input data. There are 2 options: \code{"Gaussian"} and \code{"Nonparanormal"}. The default value is \code{"Gaussian"}.}

\item{method}{For a Nonparanormal model, the test method: \code{"score"} or \code{"wald"}. The default is \code{"score"}. Ignored for Gaussian inference.}
}
\value{
An object is returned:
\item{data}{
  The \code{n} by \code{d} data matrix from the input.
}
\item{p}{
  The \code{d} by \code{d} p-value matrix of hypothesis.
}
\item{error}{
  The type I error of hypothesis at alpha significance level.
}
}
\description{
Implements the inference for high dimensional graphical models, including Gaussian and Nonparanormal graphical models
We consider the problems of testing the presence of a single edge and the hypothesis is that the edge is absent.
}
\details{
For Nonparanormal graphical model we provide Score test method and Wald Test. However it is really slow for inferencing on Nonparanormal model, especially for large data. Gaussian inference supports one variable, while Nonparanormal inference requires at least two. Nonparanormal score-test diagonal p-values do not represent edges and may be undefined; every tested off-diagonal p-value must be finite.
}
\examples{
#generate data
L = huge.generator(n = 50, d = 12, graph = "hub", g = 4)

#graph path estimation using glasso
est = huge(L$data, method = "glasso")

#inference of Gaussian graphical model at 0.05 significance level
T = tail(est$icov, 1)[[1]]
out1 = huge.inference(L$data, T, L$theta)

#inference of Nonparanormal graphical model using score test at 0.05 significance level
T = tail(est$icov, 1)[[1]]
out2 = huge.inference(L$data, T, L$theta, type = "Nonparanormal")

#inference of Nonparanormal graphical model using wald test at 0.05 significance level
T = tail(est$icov, 1)[[1]]
out3 = huge.inference(L$data, T, L$theta, type = "Nonparanormal", method = "wald")

#inference of Nonparanormal graphical model using wald test at 0.1 significance level
T = tail(est$icov, 1)[[1]]
out4 = huge.inference(L$data, T, L$theta, 0.1, type = "Nonparanormal", method = "wald")
}
\references{
1.Q Gu, Y Cao, Y Ning, H Liu. Local and global inference for high dimensional nonparanormal graphical models.\cr
2.J Jankova, S Van De Geer. Confidence intervals for high-dimensional inverse covariance estimation. \emph{Electronic Journal of Statistics}, 2015.\cr
}
\seealso{
\code{\link{huge}}, and \code{\link{huge-package}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huge.roc.R
\name{huge.roc}
\alias{huge.roc}
\title{Draw ROC Curve for a graph path}
\usage{
huge.roc(path, theta, verbose = TRUE)
}
\arguments{
\item{path}{A graph path.}

\item{theta}{The true graph structure, containing at least one edge and one absent off-diagonal edge.}

\item{verbose}{If \code{verbose = FALSE}, tracing information printing is disabled. The default value is \code{TRUE}.}
}
\value{
An object with S3 class "roc" is returned:
  \item{F1}{
    The F1 scores along the graph path.
  }
\item{tp}{
  The true positive rates along the graph path
}
\item{fp}{
  The false positive rates along the graph paths
}
\item{AUC}{
  Area under the ROC curve
}
}
\description{
Draws ROC curve for a graph path according to the true graph structure.
}
\details{
To avoid the horizontal oscillation, false positive rates is automatically sorted in the ascent order and true positive rates also follow the same order.
}
\note{
ROC/AUC is undefined when \code{theta} contains only edges or only non-edges, so those one-class truth matrices are rejected. For a lasso regression, the number of nonzero coefficients is at most \code{n-1}. If \code{d>>n}, even when regularization parameter is very small, the estimated graph may still be sparse. In this case, the AUC may not be a good choice to evaluate the performance.
}
\examples{
#generate data
L = huge.generator(d = 200, graph = "cluster", prob = 0.3)
out1 = huge(L$data)

#draw ROC curve
Z1 = huge.roc(out1$path,L$theta)

#Maximum F1 score
max(Z1$F1)
}
\seealso{
\code{\link{huge}} and \code{\link{huge-package}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huge.npn.R
\name{huge.npn}
\alias{huge.npn}
\title{Nonparanormal(npn) transformation}
\usage{
huge.npn(
  x,
  npn.func = "shrinkage",
  npn.thresh = NULL,
  verbose = TRUE,
  na.last = "keep"
)
}
\arguments{
\item{x}{The \code{n} by \code{d} data matrix representing \code{n} observations in \code{d} dimensions}

\item{npn.func}{The transformation function used in the npn transformation. If \code{npn.func = "truncation"}, the truncated ECDF is applied. If \code{npn.func = "shrinkage"}, the shrunken ECDF is applied. The default is \code{"shrinkage"}. If \code{npn.func = "skeptic"}, the nonparanormal skeptic is applied.}

\item{npn.thresh}{The truncation threshold used in nonparanormal transformation, ONLY applicable when \code{npn.func = "truncation"}. The default value is \code{1/(4*(n^0.25)*} \code{sqrt(pi*log(n)))}.}

\item{verbose}{If \code{verbose = FALSE}, tracing information printing is disabled. The default value is \code{TRUE}.}

\item{na.last}{for controlling the treatment of NAs. If TRUE, missing values in the data are put last; if FALSE, they are put first; if NA, they are removed; if "keep" they are kept with rank NA. See also \code{\link{rank}}.}
}
\value{
\item{data}{
A \code{d} by \code{d} nonparanormal correlation matrix if \code{npn.func = "skeptic"}, and A \code{n} by \code{d} data matrix representing \code{n} observations in \code{d} transformed dimensions otherwise.
}
}
\description{
Implements the Gaussianization to help relax the assumption of normality.
}
\details{
The nonparanormal extends Gaussian graphical models to semiparametric Gaussian copula models.Motivated by sparse additive models, the nonparanormal method estimates the Gaussian copula by marginally transforming the variables using smooth functions.Computationally, the estimation of a nonparanormal transformation is very efficient and only requires one pass of the data matrix.
}
\examples{
# generate nonparanormal data
L = huge.generator(graph = "cluster", g = 5)
L$data = L$data^5

# transform the data using the shrunken ECDF
Q = huge.npn(L$data)

# transform the non-Gaussian data using the truncated ECDF
Q = huge.npn(L$data, npn.func = "truncation")

# estimate the nonparanormal correlation matrix using the skeptic estimator
Q = huge.npn(L$data, npn.func = "skeptic")
}
\seealso{
\code{\link{huge}} and \code{\link{huge-package}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huge.R
\name{huge}
\alias{huge}
\title{High-dimensional undirected graph estimation}
\usage{
huge(
  x,
  lambda = NULL,
  nlambda = NULL,
  lambda.min.ratio = NULL,
  method = "mb",
  scr = NULL,
  scr.num = NULL,
  cov.output = FALSE,
  sym = "or",
  verbose = TRUE,
  input.type = "auto"
)
}
\arguments{
\item{x}{There are 2 options: (1) \code{x} is an \code{n} by \code{d} data matrix (2) a \code{d} by \code{d} sample covariance matrix. The program automatically identifies the input matrix by checking the symmetry. (\code{n} is the sample size and \code{d} is the dimension).}

\item{lambda}{An optional numeric scalar or non-empty one-dimensional
numeric path. For \code{method = "mb"}, \code{"glasso"}, or
\code{"tiger"}, values must be finite, strictly
positive, and non-increasing; ties are allowed. For \code{method = "ct"},
values must be finite and non-negative, zero is allowed, and thresholds
are applied in the supplied order. Leave \code{lambda = NULL} to generate
a path from \code{nlambda} and \code{lambda.min.ratio}.}

\item{nlambda}{The number of regularization/thresholding parameters. The default value is \code{20} for \code{method = "ct"} and \code{10} for \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}.}

\item{lambda.min.ratio}{If \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}, it is the smallest value for \code{lambda}, as a fraction of the upperbound (\code{MAX}) of the regularization/thresholding parameter which makes all estimates equal to \code{0}. The program can automatically generate \code{lambda} as a sequence of length = \code{nlambda} starting from \code{MAX} to \code{lambda.min.ratio*MAX} in log scale. If \code{method = "ct"}, it is the largest sparsity level for estimated graphs. The program can automatically generate \code{lambda} as a sequence of length = \code{nlambda}, which makes the sparsity level of the graph path increases from \code{0} to \code{lambda.min.ratio} evenly.The default value is \code{0.1} when \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}, and 0.05 when \code{method = "ct"}.}

\item{method}{Graph estimation methods with 4 options: \code{"mb"}, \code{"ct"}, \code{"glasso"} and \code{"tiger"}. The default value is \code{"mb"}.}

\item{scr}{If \code{scr = TRUE}, the lossy screening rule is applied to preselect the neighborhood before the graph estimation. The default value is  \code{FALSE}. NOT applicable when \code{method = "ct"} or \code{"tiger"}.}

\item{scr.num}{The neighborhood size after the lossy screening rule (the number of remaining neighbors per node). ONLY applicable when \code{scr = TRUE}. The default value is \code{n-1}. An alternative value is \code{n/log(n)}. ONLY applicable when \code{scr = TRUE} and \code{method = "mb"}.}

\item{cov.output}{If \code{cov.output = TRUE}, the output will include a path of estimated covariance matrices. ONLY applicable when \code{method = "glasso"}. Since the estimated covariance matrices are generally not sparse, please use it with care, or it may take much memory under high-dimensional setting. The default value is \code{FALSE}.}

\item{sym}{Symmetrize the output graphs. If \code{sym = "and"}, the edge between node \code{i} and node \code{j} is selected ONLY when both node \code{i} and node \code{j} are selected as neighbors for each other. If \code{sym = "or"}, the edge is selected when either node \code{i} or node \code{j} is selected as the neighbor for each other. The default value is \code{"or"}. ONLY applicable when \code{method = "mb"} or \code{"tiger"}.}

\item{verbose}{If \code{verbose = FALSE}, tracing information printing is disabled. The default value is \code{TRUE}.}

\item{input.type}{How to interpret \code{x}: \code{"auto"} preserves
symmetry-based detection, \code{"data"} forces an observation matrix,
and \code{"covariance"} requires a square covariance or correlation
matrix. Use \code{"data"} for square symmetric observation matrices.
For glasso covariance input, \code{"auto"} retains the historical
diagonal-sensitive default lambda scale; use \code{"covariance"} for a
scale based only on off-diagonal entries. Glasso may accept an indefinite
pairwise covariance estimate when regularization produces a certified
positive-definite result; CT, MB, and TIGER require positive
semidefiniteness.}
}
\value{
An object with S3 class \code{"huge"} is returned:
\item{data}{
  The \code{n} by \code{d} data matrix or \code{d} by \code{d} sample covariance matrix from the input
}
\item{cov.input}{
  An indicator of the sample covariance.
}
\item{ind.mat}{
  The \code{scr.num} by \code{d} matrix with each column containing the indices of the remaining neighbors of the corresponding variable after the lossy screening. ONLY applicable when \code{scr = TRUE}.
}
\item{lambda}{
  The sequence of regularization parameters used in mb or thresholding parameters in ct.
}
\item{sym}{
  The \code{sym} from the input. ONLY applicable when \code{method = "mb"} or \code{"tiger"}.
}
\item{scr}{
  The \code{scr} from the input. ONLY applicable when \code{method = "mb"} or \code{"glasso"}.
}
\item{path}{
  A list of \code{d} by \code{d} adjacency matrices of estimated graphs as a graph path corresponding to \code{lambda}.
}
\item{sparsity}{
  The sparsity levels of the graph path.
}
\item{icov}{
  A list of \code{d} by \code{d} precision matrices as an alternative graph path (numerical path) corresponding to \code{lambda}. ONLY applicable when \code{method = "glasso"} or \code{"tiger"}.
}
\item{cov}{
  A list of \code{d} by \code{d} estimated covariance matrices corresponding to \code{lambda}. ONLY applicable when \code{cov.output = TRUE} and \code{method = "glasso"}
}
\item{method}{
  The method used in the graph estimation stage.
}
\item{df}{
  If \code{method = "mb"} or \code{"tiger"}, it is a \code{d} by \code{nlambda} matrix. Each row contains the number of nonzero coefficients along the lasso solution path. If \code{method = "glasso"}, it is a \code{nlambda} dimensional vector containing the number of nonzero coefficients along the graph path \code{icov}.
}
\item{loglik}{
  A \code{nlambda} dimensional vector containing the likelihood scores along the graph path (\code{icov}). ONLY applicable when \code{method = "glasso"}. For an estimated inverse covariance Z, the program only calculates log(det(Z)) - trace(SZ) where S is the empirical covariance matrix. For the likelihood for n observations, please multiply by n/2.
}
}
\description{
The main function for high-dimensional undirected graph estimation. Four graph estimation methods, including (1) Meinshausen-Buhlmann graph estimation (\code{mb}) (2) graphical lasso (\code{glasso}) (3) correlation thresholding graph estimation (\code{ct}) and (4) tuning-insensitive graph estimation (\code{tiger}), are available for data analysis.
}
\details{
The graph structure is estimated by Meinshausen-Buhlmann graph estimation or the graphical lasso, and both methods can be further accelerated via the lossy screening rule by preselecting the neighborhood of each variable by correlation thresholding. We target on high-dimensional data analysis usually d >> n, and the computation is memory-optimized using the sparse matrix output. We also provide a highly computationally efficient approaches correlation thresholding graph estimation.
}
\note{
This function ONLY estimates the graph path. For more information about the optimal graph selection, please refer to \code{\link{huge.select}}.\cr
}
\examples{
#generate data
L = huge.generator(n = 50, d = 12, graph = "hub", g = 4)

#graph path estimation using mb
out1 = huge(L$data)
out1
plot(out1)         #Not aligned
plot(out1, align = TRUE) #Aligned
huge.plot(out1$path[[3]])
}
\seealso{
\code{\link{huge.generator}}, \code{\link{huge.select}}, \code{\link{huge.plot}}, \code{\link{huge.roc}}, and \code{\link{huge-package}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huge.R
\name{plot.huge}
\alias{plot.huge}
\title{Plot function for S3 class "huge"}
\usage{
\method{plot}{huge}(x, align = FALSE, ...)
}
\arguments{
\item{x}{An object with S3 class \code{"huge"}}

\item{align}{If \code{align = TRUE}, 3 plotted graphs are aligned to the same sparsity levels. The default value is \code{FALSE}.}

\item{\dots}{System reserved (No specific usage)}
}
\description{
Plot sparsity level information and 3 typical sparse graphs from the graph path.
}
\seealso{
\code{\link{huge}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huge.generator.R
\name{plot.sim}
\alias{plot.sim}
\title{Plot function for S3 class "sim"}
\usage{
\method{plot}{sim}(x, ...)
}
\arguments{
\item{x}{An object with S3 class \code{"sim"}}

\item{\dots}{System reserved (No specific usage)}
}
\description{
Visualize the covariance matrix, the empirical correlation matrix, the adjacency matrix and the graph pattern of the true graph structure.
}
\seealso{
\code{\link{huge.generator}} and \code{\link{huge}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huge.generator.R
\name{huge.generator}
\alias{huge.generator}
\title{Data generator}
\usage{
huge.generator(
  n = 200,
  d = 50,
  graph = "random",
  v = NULL,
  u = NULL,
  g = NULL,
  prob = NULL,
  vis = FALSE,
  verbose = TRUE
)
}
\arguments{
\item{n}{The number of observations (sample size), as an integer of at least \code{2}. The default value is \code{200}.}

\item{d}{The positive integer number of variables (dimension). The default value is \code{50}.}

\item{graph}{The graph structure with 5 options: \code{"random"}, \code{"hub"}, \code{"cluster"}, \code{"band"} and \code{"scale-free"}.}

\item{v}{A finite positive number used for the off-diagonal elements of the precision matrix, controlling the magnitude of partial correlations with \code{u}. The default value is \code{0.3}.}

\item{u}{A finite positive number added to the diagonal elements of the precision matrix to control the magnitude of partial correlations. The default value is \code{0.1}.}

\item{g}{A finite positive integer. For \code{"cluster"} or \code{"hub"} graph, \code{g} is the number of clusters or hubs; values above \code{d} are treated as \code{d}. The default is about \code{d/20} if \code{d >= 40} and \code{min(2, d)} otherwise. For \code{"band"} graph, \code{g} is the bandwidth and defaults to \code{1}; values above \code{d - 1} add no edges. Not applicable to \code{"random"} or \code{"scale-free"} graph.}

\item{prob}{For \code{"random"} or \code{"cluster"} graph, a finite number in \code{[0, 1]} giving the probability that a pair of nodes has an edge. The default is \code{min(1, 3/d)} for \code{"random"}. For \code{"cluster"}, the default is \code{min(1, 6*g/d)} if \code{d/g <= 30} and \code{0.3} otherwise. Not applicable to \code{"hub"}, \code{"band"}, or \code{"scale-free"} graph.}

\item{vis}{A single non-missing logical. If \code{TRUE}, visualize the adjacency matrix, graph pattern, covariance matrix, and empirical correlation matrix. The default is \code{FALSE}.}

\item{verbose}{A single non-missing logical. If \code{FALSE}, tracing output is disabled. The default is \code{TRUE}.}
}
\value{
An object with S3 class "sim" is returned:
\item{data}{
  The \code{n} by \code{d} matrix for the generated data
}
\item{sigma}{
  The covariance matrix for the generated data
}
\item{omega}{
  The precision matrix for the generated data
}
\item{sigmahat}{
  The empirical correlation matrix for the generated data
}
\item{theta}{
  The adjacency matrix of true graph structure (in sparse matrix representation) for the generated data
}
\item{sparsity}{
  The proportion of possible off-diagonal entries present in the graph; zero when \code{d=1}
}
}
\description{
Implements the data generation from multivariate normal distributions with different graph structures, including \code{"random"}, \code{"hub"}, \code{"cluster"}, \code{"band"} and \code{"scale-free"}.
}
\details{
Given the adjacency matrix \code{theta}, the graph patterns are generated as below:\cr\cr
(I) \code{"random"}: Each pair of off-diagonal elements are randomly set \code{theta[i,j]=theta[j,i]=1} for \code{i!=j} with probability \code{prob}, and \code{0} otherwise. It results in about \code{d*(d-1)*prob/2} edges in the graph.\cr\cr
(II)\code{"hub"}:The row/columns are evenly partitioned into \code{g} disjoint groups. Each group is associated with a "center" row \code{i} in that group. Each pair of off-diagonal elements are set \code{theta[i,j]=theta[j,i]=1} for \code{i!=j} if \code{j} also belongs to the same group as \code{i} and \code{0} otherwise. It results in \code{d - g} edges in the graph.\cr\cr
(III)\code{"cluster"}:The row/columns are evenly partitioned into \code{g} disjoint groups. Each pair of off-diagonal elements are set \code{theta[i,j]=theta[j,i]=1} for \code{i!=j} with the probability \code{prob}if both \code{i} and \code{j} belong to the same group, and \code{0} otherwise. It results in about \code{g*(d/g)*(d/g-1)*prob/2} edges in the graph.\cr\cr
(IV)\code{"band"}: Let \code{b=min(g,d-1)}. The off-diagonal elements are set to be \code{theta[i,j]=1} if \code{1<=|i-j|<=b} and \code{0} otherwise. It results in \code{(2d-1-b)*b/2} edges in the graph.\cr\cr
(V) \code{"scale-free"}: The graph is generated using B-A algorithm. The initial graph has two connected nodes and each new node is connected to only one node in the existing graph with the probability proportional to the degree of the each node in the existing graph. It results in \code{d-1} edges for \code{d>=2}; for \code{d=1}, the graph is empty.

The adjacency matrix \code{theta} has all diagonal elements equal to \code{0}. To obtain a positive definite precision matrix, the smallest eigenvalue of \code{theta*v} (denoted by \code{e}) is computed. Then we set the precision matrix equal to \code{theta*v+(|e|+0.1+u)I}. The covariance matrix is then computed to generate multivariate normal data. Every graph type returns an empty graph with zero sparsity when \code{d=1}.
}
\examples{
## band graph with bandwidth 3
L = huge.generator(graph = "band", g = 3)
plot(L)

## random sparse graph
L = huge.generator(vis = TRUE)

## random dense graph
L = huge.generator(prob = 0.5, vis = TRUE)

## hub graph with 6 hubs
L = huge.generator(graph = "hub", g = 6, vis = TRUE)

## hub graph with 8 clusters
L = huge.generator(graph = "cluster", g = 8, vis = TRUE)

## scale-free graphs
L = huge.generator(graph="scale-free", vis = TRUE)
}
\seealso{
\code{\link{huge}} and \code{\link{huge-package}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huge.ct.R
\name{huge.ct}
\alias{huge.ct}
\title{Graph estimation via correlation thresholding (ct)}
\usage{
huge.ct(
  x,
  nlambda = NULL,
  lambda.min.ratio = NULL,
  lambda = NULL,
  verbose = TRUE,
  input.type = "auto"
)
}
\arguments{
\item{x}{There are 2 options: (1) \code{x} is an \code{n} by \code{d} data matrix (2) a \code{d} by \code{d} sample covariance matrix. The program automatically identifies the input matrix by checking the symmetry. (\code{n} is the sample size and \code{d} is the dimension).}

\item{nlambda}{The number of regularization/thresholding parameters. The default value is \code{20} for \code{method = "ct"} and \code{10} for \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}.}

\item{lambda.min.ratio}{If \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}, it is the smallest value for \code{lambda}, as a fraction of the upperbound (\code{MAX}) of the regularization/thresholding parameter which makes all estimates equal to \code{0}. The program can automatically generate \code{lambda} as a sequence of length = \code{nlambda} starting from \code{MAX} to \code{lambda.min.ratio*MAX} in log scale. If \code{method = "ct"}, it is the largest sparsity level for estimated graphs. The program can automatically generate \code{lambda} as a sequence of length = \code{nlambda}, which makes the sparsity level of the graph path increases from \code{0} to \code{lambda.min.ratio} evenly.The default value is \code{0.1} when \code{method = "mb"}, \code{"glasso"} or \code{"tiger"}, and 0.05 when \code{method = "ct"}.}

\item{lambda}{A numeric scalar or non-empty one-dimensional numeric input
of finite, non-negative thresholds. Values are applied in the supplied
order, and zero is allowed. Leave
\code{lambda = NULL} to generate a path from \code{nlambda} and
\code{lambda.min.ratio}.}

\item{verbose}{If \code{verbose = FALSE}, tracing information printing is disabled. The default value is \code{TRUE}.}

\item{input.type}{How to interpret \code{x}: \code{"auto"} preserves
symmetry-based detection, \code{"data"} forces an observation matrix,
and \code{"covariance"} requires a square covariance or correlation
matrix.}
}
\description{
See more details in \code{\link{huge}}
}
\details{
The default path targets increasing numbers of undirected edges
  and defines each graph by the strict rule
  \code{abs(correlation) > lambda}. Equal-weight edges are never split, so
  ties can make the realized sparsity smaller than the nominal target.
  Reusing the returned \code{lambda} values therefore reconstructs the
  same path. When \code{lambda = NULL}, \code{nlambda} must be a positive
  integer and \code{lambda.min.ratio} must lie in \code{(0, 1]}. Supplying
  \code{lambda} overrides those two arguments.
}
\seealso{
\code{\link{huge}}, and \code{\link{huge-package}}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/huge.select.R
\name{huge.select}
\alias{huge.select}
\title{Model selection for high-dimensional undirected graph estimation}
\usage{
huge.select(
  est,
  criterion = NULL,
  ebic.gamma = 0.5,
  stars.thresh = 0.1,
  stars.subsample.ratio = NULL,
  rep.num = 20,
  verbose = TRUE,
  num.cores = 1
)
}
\arguments{
\item{est}{An object with S3 class \code{"huge"}.}

\item{criterion}{Model selection criterion. \code{"ric"} is available for all four estimation methods, \code{"stars"} for \code{"mb"}, \code{"ct"}, and \code{"glasso"}, and \code{"ebic"} only for \code{"glasso"}. Defaults are \code{"ric"} for \code{"mb"} and \code{"tiger"}, \code{"stars"} for \code{"ct"}, and \code{"ebic"} for \code{"glasso"}.}

\item{ebic.gamma}{The tuning parameter for ebic. The default value is 0.5. Only applicable when \code{est$method = "glasso"} and \code{criterion = "ebic"}.}

\item{stars.thresh}{The variability threshold in stars. The default value is \code{0.1}. An alternative value is \code{0.05}. Only applicable when \code{criterion = "stars"}.}

\item{stars.subsample.ratio}{The subsampling ratio. The default value is \code{10*sqrt(n)/n} when \code{n>144} and \code{0.8} when \code{n<=144}, where \code{n} is the sample size. Only applicable when \code{criterion = "stars"}.}

\item{rep.num}{The number of subsamplings when \code{criterion = "stars"} or rotations when \code{criterion = "ric"}. The default value is \code{20}. NOT applicable when \code{criterion = "ebic"}.}

\item{verbose}{If \code{verbose = FALSE}, tracing information printing is disabled. The default value is \code{TRUE}.}

\item{num.cores}{The number of CPU cores used to fit the stars subsamplings in parallel via \code{parallel::mclapply}. The default value is \code{1} (serial). At most two forked workers are used, and this argument is ignored on Windows. Each forked worker limits huge's native OpenMP code to one thread, but an external threaded BLAS may still create additional threads; \code{num.cores = 1} is the portable choice when a bounded thread budget or fork safety matters. Results are identical to the serial path for the same random seed. Only applicable when \code{criterion = "stars"}.}
}
\value{
An object with S3 class "select" is returned:
  \item{refit}{
    The optimal graph selected from the graph path
  }
\item{opt.icov}{
  The optimal precision matrix from the path only applicable when \code{method = "glasso"}
}
\item{opt.cov}{
  The optimal covariance matrix from the path only applicable when \code{method = "glasso"} and \code{est$cov} is available.
}
\item{merge}{
  The graph path estimated by merging the subsampling paths. Only applicable when the input \code{criterion = "stars"}.
}
\item{variability}{
  The variability along the subsampling paths. Only applicable when the input \code{criterion = "stars"}.
}
\item{ebic.score}{
  Extended BIC scores for regularization parameter selection. Only applicable when \code{criterion = "ebic"}.
}
\item{opt.index}{
  The index of the selected regularization parameter. NOT applicable when the input \code{criterion = "ric"}
}
\item{opt.lambda}{
  The selected regularization/thresholding parameter.
}
\item{opt.sparsity}{
  The sparsity level of \code{"refit"}.
}

and anything else included in the input \code{est}
}
\description{
Implements the regularization parameter selection for high dimensional undirected graph estimation. The optional approaches are rotation information criterion (ric), stability approach to regularization selection (stars) and extended Bayesian information criterion (ebic).
}
\details{
Stability approach to regularization selection (stars) selects the optimal graph by variability of subsamplings and tends to overselect edges in Gaussian graphical models. It is available for \code{"mb"}, \code{"ct"}, and \code{"glasso"}. TIGER can certify different lambda-path prefixes on different subsamples, so TIGER with stars is rejected until a common certified-prefix protocol is available; use \code{"ric"} for TIGER. Besides selecting the regularization parameters, stars can also provide an additional estimated graph by merging the corresponding subsampled graphs using the frequency counts. The subsampling procedure in stars may NOT be very efficient, we also provide the recent developed highly efficient, rotation information criterion approach (ric). Instead of tuning over a grid by cross-validation or subsampling, we directly estimate the optimal regularization parameter based on random Rotations. However, ric usually has very good empirical performances but suffers from underselections sometimes. Therefore, we suggest if user are sensitive of false negative rates, they should either consider increasing \code{rep.num} or applying the stars to model selection where supported. Extended Bayesian information criterion (ebic) is another competitive approach, but the \code{ebic.gamma} can only be tuned by experience.

For \code{criterion = "stars"}, \code{est$lambda} must be non-increasing; tied values are allowed.
}
\note{
The model selection is NOT available when the data input is the sample covariance matrix.
}
\examples{
#generate data
L = huge.generator(d = 20, graph="hub")
out.mb = huge(L$data)
out.ct = huge(L$data, method = "ct")
out.glasso = huge(L$data, method = "glasso")

#model selection using ric
out.select = huge.select(out.mb)
plot(out.select)

#model selection using stars
#out.select = huge.select(out.ct, criterion = "stars", stars.thresh = 0.05,rep.num=10)
#plot(out.select)

#model selection using ebic
out.select = huge.select(out.glasso,criterion = "ebic")
plot(out.select)
}
\seealso{
\code{\link{huge}} and \code{\link{huge-package}}.
}
